#include "StdAfx.h"

#include "BigFile.h"
#include "xarastruc.h"

CBigFile::CBigFile()
{
	for (int i=0;i<MF;i++) 
		m_f[i]=NULL;
	m_krel=-1;
}

CBigFile::~CBigFile(void)
{
}

bool CBigFile::open(const xairo_uchar* path,const xairo_uchar* mode)
{
	m_path=_wcsdup(path);
	m_access=_wcsdup(mode);
	m_f[0]=_wfopen(path,mode);
	m_n=m_p=0;
	return m_f[0] != NULL;
}

void CBigFile::reopen(const xairo_uchar* mode)
{
	fclose(m_f[0]);
	m_f[0]=_wfopen(m_path,mode);
	for (int i=1;i<MF;i++) {
		if (m_f[i]) {
			fclose(m_f[i]);
			xairo_uchar p[_MAX_PATH];
			swprintf(p,L("%s%d"),m_path,i);
			m_f[i]=_wfopen(p,mode);
		}
	}
	m_n=0;
	m_p=0;
	m_access=_wcsdup(mode);
}



void CBigFile::close()
{
	for (int i=0;i<MF;i++)
		if (m_f[i]) fclose(m_f[i]);
}


// Allow premature deletion of files that are no longer required

void CBigFile::release(__int64 i)
{
	int k=(int) (i/FS);
	if (k-1<=m_krel) return;
	if (k-1==0)
		_wremove(m_path);
	else {
			xairo_uchar p[_MAX_PATH];
			swprintf(p,L("%s%d"),m_path,k-1);
			_wremove(p);
		}
	
	m_krel=k-1;
	printf("BigFile: Released file %d\n",k-1);
}

// Don't call this with too many items - you may activate a non-existent page


int CBigFile::read(void* v,size_t x,int n)
{
	int t=n*x; int s; long db;
	if (t+m_p < FS) {
		db=ftell(m_f[m_n]);
		s=fread(v,x,n,m_f[m_n]);
		db=ftell(m_f[m_n]);
		m_p+=t;
	} else {
		int t0=FS-m_p;
		s=fread(v,1,t0,m_f[m_n]);
		m_n++;
		Activate();
		s+=fread(((char*) v)+t0,1,t-t0,m_f[m_n]);
		m_p=t-t0;
		s /=x;
	}
	Dump();
	return s;
}

long CBigFile::Dump()
{
	long db=ftell(m_f[m_n]);
	return db;
}

int CBigFile::write(void* v,size_t x,int n)
{
	int t=n*x; int s;
	if (t+m_p < FS) {
		s=fwrite(v,x,n,m_f[m_n]);
		m_p+=t;
	} else {
		int t0=FS-m_p;
		s=fwrite(v,1,t0,m_f[m_n]);
		m_n++;
		Activate();
		s+=fwrite(((char*) v)+t0,1,t-t0,m_f[m_n]);
		m_p=t-t0;
		s /= x;
	}
	return s;
}

void CBigFile::seek(__int64 i)
{
	m_n=(int) (i/FS);
	Activate();
	m_p=(unsigned) (i%FS);
	fseek(m_f[m_n],m_p,SEEK_SET);
}


void CBigFile::remove()
{
	_wremove(m_path);
	for (int i=1;i<MF;i++) {
		if (m_f[i]) {
			xairo_uchar p[_MAX_PATH];
			swprintf(p,L("%s%d"),m_path,i);
			_wremove(p);
		}
	}

}

void CBigFile::Activate()
{
	if (m_f[m_n]==NULL) {
		xairo_uchar p[_MAX_PATH];
		swprintf(p,L("%s%d"),m_path,m_n);
		m_f[m_n]=_wfopen(p,m_access);
	}
	fseek(m_f[m_n],0,SEEK_SET);
	m_p=0;
}

void CPagedFile::SetPageCount(int np,__int64 x)
{
	seek(sizeof(xara_loc_hdr)+np*sizeof (xara_loc_page));
	m_pages=new xara_loc_page [np];
	m_pn=0;
	m_maxpage=0;
	m_maxseq=x;
}

void CPagedFile::header(int ks)
{
	fseek(m_f[0],0,SEEK_SET);
	xara_loc_hdr h;
	h.ps=PS;
	h.strength=ks;
	h.np=m_pn;
	h.nf=m_n+1;
	h.maxpage=m_maxpage;
	fwrite(&h,sizeof(h),1,m_f[0]);
	fwrite(&m_pages[0],sizeof (xara_loc_page),m_pn,m_f[0]);
}

CPagedFile::CPagedFile()
{
}

CPagedFile::~CPagedFile()
{
}

int CPagedFile::writepage(void* v,int d)
{
	if (d>=FS) return 0; // avoid loop
	if (m_p+d < FS) {
		m_pages[m_pn].fi=m_n;
		m_pages[m_pn].off=m_p;
		m_pages[m_pn].len=d;
		m_pn++;
		int k=fwrite(v,1,d,m_f[m_n]);
		m_p+=d;
		if (d > m_maxpage) m_maxpage=d;
		return k;
	}
	m_n++;
	Activate();
	return writepage(v,d);
}
