// Dict.cpp: implementation of the CDict class.
//
//////////////////////////////////////////////////////////////////////

#define _CRTDBG_MAP_ALLOC
#include "StdAfx.h"
#include "Dict.h"
#include "WordList.h"
#include "Corpus.h"
#include "Indexer.h"
#include "MemoryErr.h"

#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDict::CDict(CCorpus* c):m_loc(c)
{
	m_wl=mynew CWordList;
	m_corpus=c;
	m_tree=mynew CDAVLTree;
	m_band=MINLOCS;
	m_nodecount=0;
	m_nnlu=0;
	m_b=0;
	m_maxd0=0;
	m_cont=NULL;
}

CAVLNode* CDAVLTree::NewNode(const xairo_uchar* w)
{
	_dnode* d= mynew _dnode;
	if (d==NULL)  throw CMemoryErr("dictionary",1,sizeof (_dnode));
//	d->val=wcsdup(w);
	d->m_factory=this;
	return d;
}

CDict::~CDict()
{
	mydelete m_wl;
	if (m_cont) mydelete [] m_cont;
	DelNode((_dnode*) m_tree->m_root);
	m_tree->m_root=NULL;
	mydelete m_tree;
}

void CDict::DelNode(_dnode* n)
{
	if (n==NULL) return;
	DelNode((_dnode*) n->left);
	n->left=NULL;
	DelNode((_dnode*) n->right);
	n->right=NULL;
//	mydelete [] n->x.con;
//	if (n->val) mydelete [] n->val;
	mydelete n;
}

#define MAXD 100

//#define NLU

// cacheing is not very effective

int CDict::compare(xara_word* x1,xara_word* x2)
{
//	return memcmp(x1,x2,sizeof(xara_word));
	int k=x1->val-x2->val;
	if (k==0) {
		k=x1->t-x2->t;
		if (k==0) {
			k=x1->uri-x2->uri;
			if (k==0) {
				k=x1->name-x2->name;
				if (k==0) {
					k=x1->euri-x2->euri;
					if (k==0) {
						k=x1->element-x2->element;
						for (int i=0;i<x1->nco;i++) {
							if (k==0) k=x1->con[i]-x2->con[i];
						}
						if (k==0) k=x1->xval-x2->xval;
					}
				}
			}
		}
	}
	return k;
}

/*
_dnode* CDict::Add(xara_word w)
{
//	xairo_uchar* t=wcsdup(thingy(w).getTerminatedBuffer());
	const xairo_uchar* t=thingy(w);
#ifdef NLU
	if (w.t==it_start || w.t==it_end) {
		for (int i=m_nnlu-1;i>=0;i--) {
			if (wcscmp(t,m_nlu[i]->val)==0) 
			{
				int kk=0;
				_dnode* n=m_nlu[i];
//				mydelete [] w.con;
//				delete [] t;
				n->countx++;
				return n;
			}
		}
	}
#endif
	bool bNew;
	_dnode* n=(_dnode*) m_tree->Insert(t,bNew);
	if (bNew) {
		n->x=w; // copy and keep con array
		n->x.con=mynew int[w.nco];
		for (int i=0;i<w.nco;i++)
			n->x.con[i]=w.con[i];
		n->countx=1;
		n->prev=n->first=-1;
		m_nodecount++;
	} else {
//		mydelete [] w.con;
//		delete [] t;
		n->countx++;
	}		
#ifdef NLU
	if (w.t==it_start || w.t==it_end) {
		if (m_nnlu==NLUMAX) {
			for (int j=0;j<m_nnlu-1;j++) {
				m_nlu[j]=m_nlu[j+1];
			}
			m_nnlu--;
		}
		m_nlu[m_nnlu++]=n;
	}
#endif
	return n;
}
*/

_dnode* CDict::AddNew(xara_word& w)
{
//	xairo_uchar* t=wcsdup(thingy(w).getTerminatedBuffer());
//	const xairo_uchar* t=thingy(w);
#ifdef NLU
	if (w.t==it_start || w.t==it_end) {
		for (int i=m_nnlu-1;i>=0;i--) {
			if (w.val==m_nlu[i]->x.val) // ok? need uri too?
			{
				int kk=0;
				_dnode* n=m_nlu[i];
//				mydelete [] w.con;
//				delete [] t;
				n->countx++;
				return n;
			}
		}
	}
#endif
	bool bNew; int d=0;
	_dnode* n=(_dnode*) m_tree->Insert(w,bNew,d);
	if (d > m_maxd0) m_maxd0=d;
	if (bNew) {
		n->x.SetWord(w); // copy and keep con array
		n->x.w.con=mynew int[w.nco];
		for (int i=0;i<w.nco;i++)
			n->x.w.con[i]=w.con[i];
		n->countx=1;
		n->prev=n->first=-1;
		m_nodecount++;
		m_nn[w.t]++;
	} else {
//		mydelete [] w.con;
//		delete [] t;
		n->countx++;
	}		
#ifdef NLU
	if (w.t==it_start || w.t==it_end) {
		if (m_nnlu==NLUMAX) {
			for (int j=0;j<m_nnlu-1;j++) {
				m_nlu[j]=m_nlu[j+1];
			}
			m_nnlu--;
		}
		m_nlu[m_nnlu++]=n;
	}
#endif
	return n;
}
/*
xairo_uchar thbuff [10000]; // allocate properly but only once

static UChar s_tab[]={'\t',0};

xairo_uchar * CDict::thingy(xara_word w)
{
	wcscpy(thbuff,m_wl->Get(w.val));
	wcscat(thbuff,s_tab);
	int k=wcslen(thbuff);
	thbuff[k]=w.t+'0';
	thbuff[k+1]=0;
	wcscat(thbuff,s_tab);
	wcscat(thbuff,m_wl->Get(w.uri));
	wcscat(thbuff,s_tab);
	wcscat(thbuff,m_wl->Get(w.name));
	wcscat(thbuff,s_tab);
	wcscat(thbuff,m_wl->Get(w.euri));
	wcscat(thbuff,s_tab);
	wcscat(thbuff,m_wl->Get(w.element));
	wcscat(thbuff,s_tab);
	for (int i=0;i<w.nco;i++) {
		wcscat(thbuff,m_wl->Get(w.con[i]));
		wcscat(thbuff,s_tab);
	}
	wcscat(thbuff,m_wl->Get(w.xval));
	wcscat(thbuff,s_tab);
	assert(wcslen(thbuff) < 10000);
	return thbuff;
}
*/
/*
int CDict::compare(xara_word w1, xara_word w2)
{
	if (w1.val != w2.val) return m_wl->Get(w1.val).compare(m_wl->Get(w2.val));
	if (w1.t != w2.t) return w1.t-w2.t;
	if (w1.uri != w2.uri) return m_wl->Get(w1.uri).compare(m_wl->Get(w2.uri));
	if (w1.name != w2.name) return m_wl->Get(w1.name).compare(m_wl->Get(w2.name));
	if (w1.euri != w2.euri) return m_wl->Get(w1.euri).compare(m_wl->Get(w2.euri));
	if (w1.element != w2.element) return m_wl->Get(w1.element).compare(m_wl->Get(w2.element));
	assert(w1.nco==w2.nco);
	for (int i=0;i<w1.nco;i++) {
		if (w1.con[i] != w2.con[i]) return m_wl->Get(w1.con[i]).compare(m_wl->Get(w2.con[i]));
	}
	if (w1.xval != w2.xval) return m_wl->Get(w1.xval)==m_wl->Get(w2.xval);
	return 0;
}
*/
/*
#pragma optimize("at",on)

_dnode* CDict::Add(xara_word w, _dnode **n,xairo_uchar* t,int& kk)
{
	/*
	if (*n==NULL) {
		*n=mynew _dnode;
		(*n)->x=w; // copy and keep con array
		(*n)->pre=(*n)->post=NULL;
		(*n)->prev=(*n)->first=-1;
		(*n)->countx=1;
		(*n)->thing=t;
#ifdef TROPT
		(*n)->maxnode=NULL;
#endif
		m_nodecount++;
		return *n;
	}
//	int k=compare(w,(*n)->x);
//	int k=wcscmp(t,(*n)->thing);
#ifdef TROPT
	if ((*n)->maxnode && wcscmp(t,(*n)->maxt) > 0) {
		_dnode* nn=Add(w,&(*n)->maxnode,t,kk);
		(*n)->maxnode=nn;
		(*n)->maxt=t;
		return nn; }
#endif TROPT
	const xairo_uchar* w1=t;
	const xairo_uchar* w2=(*n)->thing;
	int k;
	while (*w1 && (k=(*w1++-*w2++))==0);
	if (k==0) {
		mydelete [] w.con;
		delete [] t;
		(*n)->countx++;
		return *n;
	}
	else if (k < 0) 
	{ kk++; return Add(w,&((*n)->pre),t,kk);}
	else 
	{ 
		kk++; 
		_dnode* nn=Add(w,&((*n)->post),t,kk);
#ifdef TROPT	
		if ((*n)->maxnode==NULL) {
			(*n)->maxnode=nn;
			(*n)->maxt=t;
		}
#endif
		return nn;
	}

}

#pragma optimize("",on)
*/
void CDict::Open(CIndexer* ind,int k,int ks,const char* loc)
{
	m_ind=ind;
	if (loc[0]) m_locale=mynew Locale(loc); else m_locale=new Locale;
	UnicodeString u;
	m_locale->getDisplayLanguage(u);
	m_ind->Log(L("note"),L("Default language is %s"),u.getTerminatedBuffer());
	if (k>m_band) m_band=k;
	m_loc.Open(m_band);
	m_ncont=m_corpus->m_r.m_naddkeys-1;
	m_co=mynew int[m_ncont];
	m_cont=mynew _ca [m_ncont];
	for (int i=0;i<m_ncont;i++) {
		m_cont[i].name=m_corpus->m_r.m_addkeys[i+1].name;
		m_cont[i].index=m_corpus->m_r.m_addkeys[i+1].isLemma;
		m_cont[i].def=m_corpus->m_r.m_addkeys[i+1].defval;
	}
	m_lemma=m_corpus->m_r.m_lemma;
	for (int j=0;j<it_max;j++)
		m_nk[j]=m_nn[j]=0;
	if (ks < 0 || ks > 9) ks=5;
	m_ks=ks;
}


void CDict::Index(typ t,UnicodeString uri,UnicodeString name,UnicodeString euri,UnicodeString elt,UnicodeString val,UnicodeString xval, int nc, UnicodeString* co, int fi, int nw, int off,int len,int par)
{
//	int* con = mynew int[nc];
	for (int i=0;i<nc;i++) {
		m_co[i]=m_wl->Add(co[i]);
	}
	xara_word w(t,m_wl->Add(uri),m_wl->Add(name),m_wl->Add(euri),m_wl->Add(elt),m_wl->Add(val),m_wl->Add(xval),nc,m_co);
	xara_loc l(fi,nw,off,len,par);
	IndexWord(w,l);
}

void CDict::Index(typ t,UnicodeString uri,UnicodeString name,UnicodeString euri,UnicodeString elt,UnicodeString val,UnicodeString xval, int nc, UnicodeString* co, int fi, UnicodeString nw, int off,int len,int par)
{
//	int* con = mynew int[nc];
	for (int i=0;i<nc;i++) {
		m_co[i]=m_wl->Add(co[i]);
	}
	xara_word w(t,m_wl->Add(uri),m_wl->Add(name),m_wl->Add(euri),m_wl->Add(elt),m_wl->Add(val),m_wl->Add(xval),nc,m_co);
	xara_loc l(fi,m_wl->Add(nw),off,len,par);
	IndexWord(w,l);
}

void CDict::IndexWord(xara_word w,xara_loc l)
{
//	assert(wcscmp(m_wl->m_tab[w.val].word,L"encoding")!=0);
	m_nk[w.t]++;
	_dnode* d=AddNew(w);
	int x=m_loc.Add(l,d->prev,d,d->countx-1);
	assert(d->prev==-1 || x > d->prev);
	d->prev=x;
	if (d->first==-1) d->first=d->prev;
}

void CDict::Close()
{
	m_ind->Log(L("process"),L("Sorting patch"));
	m_loc.Close();
}

void CDict::Report()
{
//	static int nr=1;
//	if (nr++ % 100 == 0) m_wl->Dump();
//	Typestats();
}

#define PRNTFREQ false

struct _sl {
		int val;
		int loc;
};

int CDict::comp(const void* e1,const void* e2)
{
	_sl* s1=(_sl*) e1;
	_sl* s2=(_sl*) e2;
	return s1->val-s2->val;
}

static CWordList* g_twl;
Collator* g_Collator;
int g_nf;

int CDict::slcomp(const void* e1,const void* e2)
{
	xara_lemma_0* s1=(xara_lemma_0*) e1;
	xara_lemma_0* s2=(xara_lemma_0*) e2;
//	const xairo_uchar* w1=g_twl->Get(s1->word);
//	const xairo_uchar* w2=g_twl->Get(s2->word);
//	return wcscmp(w1,w2);
	for (int i=0;i<g_nf;i++) {
		int k=g_Collator->compare((const UChar*) g_twl->Get(s1->word[i]),(const UChar*) g_twl->Get(s2->word[i])); // use collator here
		if (k != 0) return k;
	}
	return 0;
}


int CDict::hs(const void* e1,const void* e2)
{
	xara_lemma_0* s1=(xara_lemma_0*) e1;
	xara_lemma_0* s2=(xara_lemma_0*) e2;
	for (int i=0;i<g_nf;i++) {
		const xairo_uchar* w1=g_twl->Get(s1->word[i]);
		const xairo_uchar* w2=g_twl->Get(s2->word[i]);
		int k=wcscmp(w1,w2);
//	int k=g_Collator->compare(g_twl->Get(s1->word),g_twl->Get(s2->word)); // use collator here
		if (k != 0) return k;
	}
	return wcscmp(g_twl->Get((int)(long) s1->pt),g_twl->Get((int)(long) s2->pt));

}

// Now make the lemma table. 

void CDict::LemmaTable()
{
	UErrorCode success = U_ZERO_ERROR;
	g_Collator = Collator::createInstance(success);
	g_Collator->setStrength((Collator::ECollationStrength) 2);
	int m=-1;
	LemmaTable(-1,1,&m);
	int i=0;
	for (i=0;i<m_ncont;i++)
		LemmaTable(i,1,&i);
// and now add the explicitly defined ones
	for (int ii=0;ii<m_corpus->m_r.m_nls;ii++) {
		_ls* l=&m_corpus->m_r.m_ls[ii];
		int* x=new int[l->n];
		for (int j=0;j<l->n;j++) {
			int c;
			for (c=0;c<m_ncont;c++) {
				if (m_cont[c].name==l->a[j]) break;
			}
			x[j]=c;
		}
		LemmaTable(i++,l->n,x);
	}
}

bool diff(int a[MAXF],int b[MAXF])
{
	for (int i=0;i<g_nf;i++)
		if (a[i] != b[i]) return true;
	return false;
}

void assign(int a[MAXF],int b[MAXF])
{
	for (int i=0;i<MAXF;i++)
		a[i] = b[i];
}

void init(int a[MAXF])
{
	for (int i=0;i<MAXF;i++)
		a[i] = -1;
}

void CDict::LemmaTable(int lem,int nf,int* fac)
{
	g_nf=nf;
	int c=-1; int k=0;
	DP(c,(_dnode*) m_tree->m_root,k);
	xara_word cx;
	cx.val=-1;
	m_nd=0; 
	LT0(cx,(_dnode*) m_tree->m_root,k); 
	xara_lemma_0* ltab = mynew xara_lemma_0[m_nd];
	if (ltab==NULL) throw(CMemoryErr("ltab",m_nd,sizeof(xara_lemma_0))); 
	xara_lemma_0* ltab0 = ltab;
	m_nd=m_nd0=0;
	ltab--;
// make the unsorted lemma table
	cx.val=-1;
	LT1(cx,ltab,(_dnode*) m_tree->m_root,nf,fac);
	ltab++;
	int nfo=ltab-ltab0;
// now we want to sort the table by headword
	g_twl=m_wl;
	qsort(ltab0,ltab-ltab0,sizeof (xara_lemma_0),hs);
// now count the number of headwords
	int w[MAXF]; int nhw=0;
	init(w);
	xara_lemma_0* l;
	for (l=ltab0;l<ltab;l++) {
		if (diff(l->word,w)) {
			assign(w,l->word);
			nhw++;
		}
	}
	xara_lemma_0* lltab0=mynew xara_lemma_0[nhw];
	if (lltab0==NULL) throw CMemoryErr("lltab0",nhw,sizeof (xara_lemma)); 
	xara_lemma_0* lltab=lltab0;
	lltab--;
	init(w); 
	for (l=ltab0;l<ltab;l++) {
		if (diff(l->word,w)) {
			assign(w,l->word);
			lltab++;
			lltab->forms=1;
			lltab->freq=l->freq;
		} else {
			lltab->forms++;
			lltab->freq+=l->freq;
		}
	}
	init(w); 
	lltab=lltab0;
	lltab--; int kl;
	for (l=ltab0;l<ltab;l++) {
		if (diff(l->word,w)) {
			assign(w,l->word);
			lltab++;
			lltab->pt=mynew xara_form[lltab->forms];
			assign(lltab->word,w);
			lltab->forms=1;
			kl=0;
		} else lltab->forms++;
		lltab->pt[kl].ptr=l->nxt;
		lltab->pt[kl].count=l->forms;
		lltab->pt[kl].freq=l->freq;
		kl++;
	}
	lltab++;
// there it is!
	mydelete [] ltab0;
	UErrorCode success = U_ZERO_ERROR;
	Locale loc;
	g_Collator = Collator::createInstance(success);
	g_Collator->setStrength((Collator::ECollationStrength) 2);
//	qsort(lltab0,nhw,sizeof lltab0[0],slcomp);
	if (m_corpus->m_bFreq) {
		xara_lemma_0* sltab=mynew xara_lemma_0 [nhw];
		int i;
		for (i=0;i<nhw;i++) {
			assign(sltab[i].word,lltab0[i].word);
			sltab[i].freq=lltab0[i].freq;
			sltab[i].forms=lltab0[i].forms;
		}

		qsort(sltab,nhw,sizeof sltab[0],slcomp);
		xairo_uchar fname[20];
		swprintf(fname,L("freq%d.txt"),lem+1);
		FILE* fi=_wfopen((const xairo_uchar*) m_corpus->exppath(CCorpus::PARAM_ETCPATH,(const UChar*) fname).getTerminatedBuffer(),L("wb"));
		fprintf(fi,"\0xfe\0xff"); // hmm
		for (i=0;i<nhw;i++) {
			xara_lemma_0* l=&sltab[i];
			for (int j=0;j<nf;j++) {
				fwprintf(fi,L("%s "),m_wl->Get(l->word[j]));
			}
			fwprintf(fi,L("%d %d\r\n"),l->freq,l->forms);

		}
		fclose(fi);
	}
	_sl* s=mynew _sl [nhw];
	int nu=0;
	int i;
	for (i=0;i<nhw;i++) {
		nu+=s[i].val=lltab0[i].freq;
		s[i].loc=i;
	}
	qsort(s,nhw,sizeof(_sl),comp);
	int maxloc=s[nhw-1].loc;
	for (int j=nhw-2;j>=0;j--) {
		lltab0[s[j+1].loc].nxt=s[j].loc;
	}
	if (nhw) lltab0[s[0].loc].nxt=-1;
	mydelete [] s;
	UChar sn[20];
	swprintf((xairo_uchar*) sn,L("xdblemma%d"),lem+1);
	FILE* f=iopen(m_corpus->exppath(CCorpus::PARAM_INDEXPATH,sn),(const UChar*) L("wb"));
	xara_lemma_header h;
	h.maxpos=maxloc;
	h.ndix=nhw;
	h.nu=nu;
	h.nfacets=nf;
	m_ind->Log(L("stat"),L("<stat name=\"nu\" lemma=\"%d\">%d</stat>"),lem+1,nu);
	m_ind->Log(L("stat"),L("<stat name=\"nw\" lemma=\"%d\">%d</stat>"),lem+1,nhw);
	m_ind->Log(L("stat"),L("<stat name=\"nfo\" lemma=\"%d\">%d</stat>"),lem+1,nfo);
//	m_ind->Log(L"stat",L"<stat name=\"nt\" lemma=\"%s\">%d</stat>",IntToWide(lem+1).getTerminatedBuffer(),IntToWide(k).getTerminatedBuffer());

	fwrite(&h,sizeof h,1,f);
	for (i=0;i<nhw;i++) {
		fwrite(&lltab0[i],3*sizeof(int),1,f);
		fwrite(&lltab0[i].word[0],h.nfacets*sizeof(int),1,f);
		fwrite(lltab0[i].pt,sizeof(xara_form),lltab0[i].forms,f);
	}
	fclose(f);
	for (i=0;i<nhw;i++) {
		mydelete [] lltab0[i].pt;
	}
	mydelete [] lltab0;
}

void CDict::DP(int& w,_dnode* d,int& k)
{
	if (d==NULL) return;
	DP(w,(_dnode*) d->left,k);
	int v=d->x.GetWord()->val;
	if (w != v) {
		m_wl->m_tab[v].dicptr=k;
		w=v;
	}
	k++;
	DP(w,(_dnode*) d->right,k);
}

void CDict::LT0(xara_word& w,_dnode* d,int& k)
{
	if (d==NULL) return;
	LT0(w,(_dnode*) d->left,k);
	xara_word* wd=d->x.GetWord();
	if (distinct(&w,wd)) {
//		m_wl->m_tab[d->x.val].dicptr=k;
		if (wd->t == it_content) {
			m_nd++;
		}
		w=*wd;
	}
	k++;
	LT0(w,(_dnode*) d->right,k);
}

// This isn't right - should fold as it indexes so as to use 
// correct language
int CDict::HW(xara_word w,int k)
{
		if (k >= 0) {
			bool bFold=!m_corpus->m_r.m_addkeys[k+1].bCaseSensitive;
			CTokeniser* t=m_ind->GetTokeniser();
			int x1=w.con[k];
			const xairo_uchar* ex=m_wl->Get(x1);
			UnicodeString f=t->FoldCase((const UChar*) ex);
			if (bFold) ex=(const xairo_uchar*) f.getTerminatedBuffer();
			int x2=m_wl->Add(ex);
			return x2;
		}
		else
			return w.val;
}

bool CDict::distinct(xara_word* w1,xara_word* w2)
{
	if (w1->val != w2->val) return true;
	for (int i=0;i<m_ncont;i++) {
		if (w1->con[i] != w2->con[i]) return true;
	}
	return false;
}


void CDict::LT1(xara_word& w,xara_lemma_0 *&lt,_dnode* d,int nf,int* f)
{
	if (d==NULL) return;
	LT1(w,lt,(_dnode*) d->left,nf,f);
	xara_word* wd=d->x.GetWord();
	if (wd->t == it_content && !distinct(&w, wd)) {
		lt->freq+=d->count;
		lt->forms++;
	}
	else 
	{
		if (wd->t==it_content) {
			lt++;
			w=*wd;
			for (int i=0;i<nf;i++)
				lt->word[i]=HW(*wd,f[i]); // make this the lemma
//			CTokeniser t;

//			lt->word=m_wl->Add(t.FoldCase(m_wl->Get(lt->word)));
			lt->freq=d->count;
			lt->forms=1; 
			lt->nxt=m_nd;  // shameful reuse of this field
			lt->pt=(xara_form*) (long) wd->val; // even worse, but it keeps the forms in order
		} else w.val=-1;
	}
	m_nd++;
	LT1(w,lt,(_dnode*) d->right,nf,f);
}

void CDict::FUBand()
{
	m_loc.FUBand();
	m_loc.FULocsBand();
	m_loc.FUEndBand();
}




void CDict::SaveLocs2()
{
	m_ind->Log(L("process"),L("Merging"));
	m_loc.SaveLoc3(m_ks);

}


void CDict::Typestats()
{
	for (int i=0;i<it_max;i++) {
		m_ind->Log(L("stat"),L("<stat name=\"nty\" ty=\"%s\">%d/%d</stat>"),CIndexer::Lname((typ) i).getTerminatedBuffer(),m_nk[i],m_nn[i]);
	}
	m_ind->Log(L("stat"),L("<stat name=\"wt\">%d/%d</stat>"),m_wl->m_next,m_wl->m_nchars);
}

void CDict::Save(int nt)
{
//	m_ind=ind;
//	if (nLoc > m_band) m_band=nLoc;
	UnicodeString dict=m_corpus->exppath(CCorpus::PARAM_INDEXPATH,(const UChar*) L("xdbdict"));
	m_df=iopen(dict,(const UChar*) L("wb"));
	m_dlf=NULL;
	if (m_corpus->m_bFreq) {
		m_dlf=_wfopen((const xairo_uchar*) m_corpus->exppath(CCorpus::PARAM_ETCPATH,L("freqa.txt")).getTerminatedBuffer(),L("wb"));
		fprintf(m_dlf,"\0xfe\0xff"); // hmm
	}
	m_loc.Open2(nt,m_ind);
//	m_loc.m_band=m_band;
	int sbn=0;
	m_p=0;
	m_ind->Log(L("process"),L("Save dictionary"));
	m_ind->Log(L("stat"),L("<stat name=\"kloc\">%d</stat>"),m_kloc);
	m_ind->Log(L("stat"),L("<stat name=\"klocs\">%d</stat>"),m_kloc*sizeof(locplus));
	Typestats();
	m_kloc=0; m_maxd=0;
	SaveNode((_dnode*) m_tree->m_root,sbn,0);
	printf("                                               \r");
	m_loc.m_kloc=m_kloc;
	int nb=(m_kloc+m_band-1)/m_band;
	m_ind->Log(L("stat"),L("<stat name=\"nband\">%d</stat>"),nb);
	m_ind->Log(L("stat"),L("<stat name=\"mband\">%d</stat>"),m_band);
	m_loc.Fixup0();
	for (int i=0;i<nb;i++) {
		m_ind->Log(L("process"),L("Band %d of %d"),i+1,nb);
		FUBand(); // was SaveLocsBand
	}
	m_ind->Log(L("temp"),L("Finished band sorts"));
	m_loc.Fixup1(); 
	SaveLocs2(); // compress - must free up loc buffer
	m_ind->Log(L("stat"),L("<stat name=\"nd\">%d</stat>"),sbn);
	int sd=sizeof(_dnode)+m_ncont*4;
	m_ind->Log(L("stat"),L("<stat name=\"sd\">%d</stat>"),sd);
	m_ind->Log(L("stat"),L("<stat exp=\"nd*sd\">%d</stat>"),sbn*sd);
	m_ind->Log(L("stat"),L("<stat name=\"maxd\">%d</stat>"),m_maxd);
//	m_ind->Log(L("stat"),L("<stat exp=\"2^maxd-1\">%.f</stat>"),pow(2.0,m_maxd)-1);
	m_ind->Log(L("process"),L("Accelerators"));
	int na=m_loc.Accelerate();
	m_ind->Log(L("stat"),L("<stat name=\"naccel\">%d</stat>"),na);
//	m_ind->Log(L"process",L"Thread locations");
//	m_loc.Thread();
//	m_ind->Log(L"process",L"Compressing");
	m_loc.Close2();
	fclose(m_df);
	if (m_dlf) fclose(m_dlf);
	m_ind->Log(L("process"),L("Lemma table"));
	LemmaTable();
	m_ind->Log(L("process"),L("Typelink"));
	TypeLink();
	xara_wl_header_addkey* k = mynew xara_wl_header_addkey [m_ncont];
	for (int i=0;i<m_ncont;i++) {
		k[i].name=m_wl->Add(m_cont[i].name);
		k[i].bLemma=m_cont[i].index;
		k[i].def=m_wl->Add(m_cont[i].def);
	}
	int ntax=m_corpus->m_r.m_ntax;
	xara_wl_header_tax* kk=mynew xara_wl_header_tax [ntax];
	for (int i=0;i<ntax;i++) {
		kk[i].name=m_wl->Add(m_corpus->m_r.m_taxonomies[i].id);
		kk[i].nc=m_corpus->m_r.m_taxonomies[i].n;
	}
	m_ind->Log(L("process"),L("Wordlist"));
	m_wl->Save(m_corpus->exppath(CCorpus::PARAM_INDEXPATH,"xdbwl"),m_ncont,k,ntax,m_corpus->m_r.m_nls,kk,m_lemma);
	m_ind->Log(L("stat"),L("<stat name=\"nsym\">%d</stat>"),m_wl->GetSymbols());
	m_ind->Log(L("stat"),L("<stat name=\"nht\">%d</stat>"),m_wl->GetHashSize());
	m_ind->Log(L("stat"),L("<stat name=\"nlhc\">%d</stat>"),m_wl->GetLHC());
	m_ind->Log(L("stat"),L("<stat name=\"nsb\">%d</stat>"),m_wl->GetChars()*2);
	m_ind->Log(L("stat"),L("<stat name=\"maxwl\">%d</stat>"),m_wl->GetMaxWL());
	m_ind->Log(L("stat"),L("<stat name=\"maxw\">%.500s</stat>"),m_wl->Get(m_wl->GetMaxW()));

	mydelete [] k;
	mydelete [] kk;
}

// We are going to have a single header file from now on.



void CDict::SaveNode(_dnode *n,int& sbn,int d)
{
tr:
	if (n==NULL) return;
	if (d > m_maxd) m_maxd=d;
	printf("%d/%d\r",m_p++,m_nodecount);
	SaveNode((_dnode*) n->left,sbn,d+1);
	xara_streambase sb;
	sb.w=*n->x.GetWord();
	sb.nxt=-1;
	sb.first=n->ptr=n->ptr0=m_kloc;
	n->count=sb.count=n->countx;
	bool bAcc=HasAcc(&sb);
	n->sbn=sbn++;
//	m_loc.SaveLocs(n->first,sbn++,bAcc,a);
	n->acc=sb.acc=bAcc?m_loc.StartAcc()+1:0;
//	n->count=sb.count=m_loc.Count()-sb.first;
	int delta=sizeof (void*) - sizeof (int);
	fwrite(&sb.w,sizeof(sb.w)-delta,1,m_df);
	fwrite(&sb.first,sizeof(sb)-sizeof(sb.w),1,m_df);
//	fwrite(&sb,sizeof sb-delta,1,m_df);
	if (m_dlf && sb.w.t==it_content) {
		fwprintf(m_dlf,L("%s"),m_wl->Get(sb.w.val));
		for (int i=0;i<m_ncont;i++) {
			fwprintf(m_dlf,L(",%s"),m_wl->Get(sb.w.con[i]));
		}
		fwprintf(m_dlf,L(",%d\r\n"),sb.count);
	}
	m_kloc+=n->countx;
	int* con = mynew int [m_ncont];
	int i;
	for (i=0;i<m_ncont;i++)
		con[i]=0;
	for (i=0;i<sb.w.nco;i++) {
		con[i]=sb.w.con[i];
	}
	fwrite(con,sizeof (int),m_ncont,m_df);
	mydelete [] con;
//	delete n->val;
	n=(_dnode*) n->right;
	d++;
	goto tr;
//	SaveNode(n->post,sbn);
}



void CDict::TypeLink()
{
	UnicodeString dict=m_corpus->exppath(CCorpus::PARAM_INDEXPATH,"xdbdict");
	int delta=sizeof(void*)-sizeof(int);
	FILE* sbf=iopen(dict,(const UChar*) L("r+b"));
	int x=sizeof(struct xara_streambase)+m_ncont*sizeof(int)-delta;
	m_troot=0;
	xara_streambase sb, sb1;
	int * con=mynew int [m_ncont];
	for (int i=0;;i++) {
		if (fread(&sb,sizeof sb-delta,1,sbf)==0) break;
		sb.w.con=con;
		fread(sb.w.con,sizeof (int),m_ncont,sbf);
		int k=AddWord(sb.w,i);
		if (k==-1) continue;
		int p=ftell(sbf);
		fseek(sbf,k*x,SEEK_SET);
		fread(&sb1,sizeof sb1.w-delta,1,sbf); // no this will move nxt 
		fread(&sb1.first,sizeof sb1-sizeof sb1.w,1,sbf);
		sb1.nxt=i;
		fseek(sbf,k*x,SEEK_SET);
		fwrite(&sb1,sizeof sb1.w-delta,1,sbf);
		fwrite(&sb1.first,sizeof sb1-sizeof sb1.w,1,sbf);
		fseek(sbf,p,SEEK_SET);
	}
	fclose(sbf);
	mydelete [] con;
	TypeList();
	_tnode* t=m_troot;
	while (t) {
		_tnode* tt=t->next;
		mydelete t;
		t=tt;
	}
}

int CDict::AddWord(xara_word w,int k0)
{
	_tnode** t=&m_troot;
	while (*t) {
		if (MatchType(w,(*t)->ty)) break;
		t=&(*t)->next;
	}
	_tnode* tt; int k;
	if (*t==NULL) {
		tt=mynew _tnode;
		tt->ty=w;
		tt->first=k0;
		tt->next=NULL;
		*t=tt;
		k=-1;
	} else {
		tt=*t;
		k=tt->last;
	}
	tt->last=k0;
	return k;
}

bool CDict::MatchType(xara_word w,xara_word ww) 
{
	if (w.t != ww.t) return false;
	switch (w.t) {
		case it_local_attribute:
			return w.element==ww.element &&
				w.name==ww.name;
		case it_global_attribute:
			return w.name==ww.name;
		default:
			return true;
	}
}

void CDict::TypeList()
{
	UnicodeString dict=m_corpus->exppath(CCorpus::PARAM_INDEXPATH,"xdbtype");
	FILE* f=_wfopen((const xairo_uchar*) dict.getTerminatedBuffer(),L("wb"));
	_tnode *t=m_troot;
	while (t) {
		xara_type tp;
		tp.t=t->ty.t;
		tp.att=t->ty.name;
		tp.element=t->ty.element;
		tp.first=t->first;
		fwrite(&tp,sizeof(tp),1,f);
		t=t->next;
	}
	fclose(f);
}

static UChar s_unit[]={'u','n','i','t',0};
static UChar s_scope[]={'s','c','o','p','e',0};

bool CDict::HasAcc(xara_streambase* sb)
{
	if (sb->count>10000) return true;
	if (sb->w.t != it_global_attribute_value) return false;
	if (wcscmp(m_wl->Get(sb->w.name),s_unit)==0) return true;
	if (wcscmp(m_wl->Get(sb->w.name),s_scope)==0) return true;
	return false;
}


/*
void xwcomp::SetWord(xara_word w)
{
	if (w.t==it_content)
	{
		val=w.val;
		t=w.t;
		xval=w.xval;
		nco=w.nco;
		ptr.con=w.con;
//		ptr.con=mynew int[nco];
//		for (int i=0;i<nco;i++)
//			ptr.con[i]=w.con[i];
	}
	else 
	{
		t=w.t;
		ptr.m_w=new xara_word;
		*ptr.m_w=w;
	}
}

xara_word xwcomp::GetWord()
{
	if (t==it_content) {
		xara_word w;
		memchr(&w,0,sizeof(w));
		w.val=val;
		w.t=t;
		w.xval=xval;
		w.nco=nco;
		w.con=ptr.con;
		return w;
	} else return *ptr.m_w;
}

xwcomp::~xwcomp()
{
	if (t==it_content)
		delete [] ptr.con;
	else 
		delete ptr.m_w;
}

int xwcomp::Compare(xara_word w)
{
	int k=0;
	if (val != w.val) k=w.val-val;
	else if (t != w.t) k=w.t-t;
	else if (t==it_content) {
		for (int i=0;i<nco;i++) {
			if (k==0) k=w.con[i]-ptr.con[i];
		}
		if (k==0)
			k=w.xval-xval;
	}
	else {
        k=memcmp(&w,ptr.m_w,(size_t) 6*sizeof(int));
		if (k==0) {
			k=w.element-ptr.m_w->element;
//			for (int i=0;i<ptr.m_w->nco;i++) {
//				if (k==0) k=w.con[i]-ptr.m_w->con[i];
			if (k==0)
				k=w.xval-ptr.m_w->xval;
		}
	}
	return k;
}
*/

void xwcomp::SetWord(xara_word w)
{
	this->w=w;
}



xwcomp::~xwcomp()
{
	delete [] w.con;
}

int xwcomp::Compare(xara_word w)
{
    int k=memcmp(&w,&this->w,(size_t) 6*sizeof(int));
	if (k==0) {
//			k=w.element-ptr.m_w->element;
		for (int i=0;i<this->w.nco;i++) {
			if (k==0) k=w.con[i]-this->w.con[i];
		}
		if (k==0) k=w.xval-this->w.xval;
	}
	return k;
}

