// Indexer.h: interface for the CIndexer class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_INDEXER_H__8E895388_52B8_4170_A015_17BC0E78847D__INCLUDED_)
#define AFX_INDEXER_H__8E895388_52B8_4170_A015_17BC0E78847D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000



using namespace xercesc;


#include "Logger.h"
#include "Tokeniser.h"
#include "xarastruc.h"
#include "Dict.h"
#include "DeclarationManager.h"
//class Locator;
class CCorpus;

#define FILEN "file"
#define ENAME "ename"
#define ENCODING "encoding"
#define PATH "path"
#define ATTACH "attach"
#define LINE "line"
#define OFFSET "offset"
#define CORPUS "corpus"
#define NAME "name"
#define TEXTS "texts"


struct _ctxt {
	class CTextReader* reader;
	int iif, nw, line, lll, col, oldo;
	_ctxt* next;
};


class CIndexer  
{
	friend class SaraIndexContentHandler;
public:
	void Region(int,bool);
	void EndCorpus();
	void StartCorpus(UnicodeString,int,_lang*);
	void IndexPI(UnicodeString,UnicodeString,int,int,int);
	void IndexSPM(UnicodeString,UnicodeString,int,int,int);
	void IndexEPM(UnicodeString,int,int,int);
	CIndexer(CCorpus*,int,int);
	virtual ~CIndexer();
	int Index(typ,UnicodeString,int,int,int,int,UnicodeString,int);
	int Index(typ,UnicodeString,int,int,UnicodeString,int);
	int IndexStartTag(UnicodeString,int,int,int,int,UnicodeString,bool&,bool&,int);
	void IndexEndTag(UnicodeString,int,int,int,int,UnicodeString,bool,int);

	void IndexGlobalAttr(UnicodeString,UnicodeString,int,int,UnicodeString,int);
	void IndexGlobalAttrVal(UnicodeString,UnicodeString,int,int,UnicodeString,int);
	void IndexLocalAttr(UnicodeString,UnicodeString,UnicodeString,int,int,UnicodeString,int);
	void IndexAddkey(UnicodeString&,UnicodeString&,int,int,int);
	void IndexNullAddkey(UnicodeString&,UnicodeString&);
	void IndexInternalEntity(UnicodeString&,UnicodeString&);
	void Premap(int,int);
	void Premap2(int,int);
	void Error(const SAXParseException& e);
	void Error(int l,int c,const xairo_uchar* msg);
	void HeaderError(const xairo_uchar* msg);
	void Note(XMLCh*);
	void Log(const XMLCh*,const XMLCh*,...);
	XMLFormatter m_f;
	XMLFormatter m_fe;
	CLogger m_dump;
	CLogger m_clog;
	void NewFile(int);
	void StartText(UnicodeString,UnicodeString);
	void SkipText(UnicodeString);
	void EndText(UnicodeString,int);
	UnicodeString m_name;
	UnicodeString m_relpath;
	class CTextReader* m_reader;
	CDict m_dict;
	void Phase2();
	void EndLog();
	void IndexLabel(UnicodeString,UnicodeString,int,int,int);
	void IndexScope(int,int,int);
	void Bib(int,int,int,UnicodeString);
	bool m_bDump;
	class CSaraHeaderReader* GetReader(); 
	void IndexWord(const xairo_uchar*,int,int,int,int,int);
	bool m_bDoc;
	void dummyDocument(UnicodeString,int);
	int m_if;
	int m_nw;
	void Push();
	void Pop();
	int ErrCount() {return m_errs;}
	void IncErrs() {m_errs++;}
	int m_nt;
	void Stats();
	bool m_bMem;
	void SetBreak(char*);
	static UnicodeString Lname(typ);
	void DSC();
	int GetSeq() {return m_dict.GetSeq();}
	void SetLang(const char* = "-");
	const char* GetLang() {return m_lang;}
	CTokeniser* GetTokeniser() {return m_tok;}
	void Regions(int,UnicodeString);
private:
	void InitWC();
	int* NewWC();
	bool* m_rmask;
	const char* m_lang;
	void Break(int,int);
//	void IndexEntry(typ, const XMLCh*, const XMLCh*,const XMLCh*,const XMLCh*,const XMLCh*,const XMLCh*,int);
	int IndexEntry(typ, UnicodeString, UnicodeString,UnicodeString,UnicodeString,UnicodeString,UnicodeString,int,UnicodeString*,int,int,int,int,int);
	int IndexEntryW(typ, UnicodeString, UnicodeString,UnicodeString,UnicodeString,UnicodeString,UnicodeString,int,UnicodeString*,int,int,int,int,UnicodeString,int);
	int IndexEntryW(typ, UnicodeString, UnicodeString,UnicodeString,UnicodeString,UnicodeString,UnicodeString,int,UnicodeString*,int,int,int,int,int,int);
	void InitTax();
	void UndefTax(int,int,int);
	CTokeniser* m_tok;
	int m_l0;
	bool m_bLine;
	int m_lab;
	void TryLineLabel(int,int);
	int m_lll;
	int *m_w1, *m_w2;
	void SCWrite(struct _taxonomy*,UnicodeString);
	_ctxt* m_stack;
	int m_errs;
	int m_brkfi, m_brkloc;
	bool m_bBrk;
	int m_band;
	UnicodeString* m_cont;
	CDeclarationManager m_dm;
	int m_kstr;
};

UnicodeString IntToWide(int);
UnicodeString FloatToWide(double);

#endif // !defined(AFX_INDEXER_H__8E895388_52B8_4170_A015_17BC0E78847D__INCLUDED_)
