// SaraHeaderReader.cpp: implementation of the SaraHeaderReader class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_TOOLS) && defined(_WIN32)
#include "..\indexer\StdAfx.h"
#else
#include "StdAfx.h"
#endif

#ifdef _WIN32
typedef wchar_t xairo_uchar;

#define L(x) L ## x
#endif


//#include "indextools.h"
#ifdef _WIN32
#include "../indexer/xarastruc.h"
#else
#include "xarastruc.h"
#endif
#include "SaraHeaderReader.h"

#if defined(_DEBUG) && !defined(_TOOLS)
#define _CRTDBG_MAP_ALLOC
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif

UnicodeString StringsToString(int,UnicodeString*);
void StringToStrings(UnicodeString,int&,UnicodeString*&);


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

FILE* iopen(UnicodeString,const UChar*);

int swcmp(const char* c,const xairo_uchar* w)
{
	while (*c || *w) {
		if (*c != *w) return *c-*w;
		c++;
		w++;
	}
	return 0;
}

CSaraHeaderReader::CSaraHeaderReader()
{
	m_nlabels=0;
	m_bib=NULL;
	m_nbib=0;
	m_deflang=NULL;
	m_ntags=0;
	m_nglobatts=0;
	m_tagusage=NULL;
	m_ntax=0;
	m_uranges=NULL;
	m_bConverting=false;
	m_error_handler=NULL;
	m_nreg=0;
}

CSaraHeaderReader::~CSaraHeaderReader()
{
	DelTagUsage();
	DelTaxonomies();
}

static UChar s_ident[]={'i','d','e','n','t',0};
static UChar s_codebooks[]={'c','o','d','e','b','o','o','k','s',0};
static UChar s_rb[]={'r','b',0};
static UChar s_teiHeader[]={'t','e','i','H','e','a','d','e','r',0};
static UChar s_xara_indexspec[]={'x','a','r','a',':','i','n','d','e','x','s','p','e','c',0};
static UChar s_indexSpec[]={'i','n','d','e','x','S','p','e','c',0};
static UChar s_target[]={'t','a','r','g','e','t',0};
static UChar s_iip[]={'I','n','d','e','x','t','o','o','l','s',' ','m','a','i','n','t','a','i','n','e','d',' ',
'l','i','s','t',' ','o','f',' ','i','n','d','e','x',' ','p','o','l','i','c','i','e','s',0};
//static UChar s_xaira_url[]={'h','t','t','p',':','/','/','w','w','w','.','x','a','i','r','a','.','o','r','g','/','n','s','/','1','.','0','/','s','p','e','c',0};
static UChar s_lemmaScheme[]={'l','e','m','m','a','S','c','h','e','m','e',0};
static UChar s_region[]={'r','e','g','i','o','n',0};
static UChar s_name[]={'n','a','m','e',0};
static UChar s_nameList[]={'n','a','m','e','L','i','s','t',0};
static UChar s_default[]={'d','e','f','a','u','l','t',0};
static UChar s_valList[]={'v','a','l','L','i','s','t',0};
static UChar s_valItem[]={'v','a','l','I','t','e','m',0};
static UChar s_xmlns[]={'h','t','t','p',':','/','/','w','w','w','.','w','3','.','o','r','g','/','X','M','L','/','1','9','9','8','/','n','a','m','e','s','p','a','c','e',0};
static UChar s_ns[]={'n','s',0};
static UChar s_xara_codebook[]={'x','a','r','a',':','c','o','d','e','b','o','o','k',0};
static UChar s_btagsDecl[]={'b','t','a','g','s','D','e','c','l',0};
static UChar s_[]={0};
static UChar s_pseudo[]={'p','s','e','u','d','o',0};
//static UChar s_element[]={'e','l','e','m','e','n','t',0};
//static UChar s_attribute[]={'a','t','t','r','i','b','u','t','e',0};
static UChar s_element[]={'e','l','e','m','e','n','t',0};
static UChar s_attribute[]={'a','t','t','r','i','b','u','t','e',0};
static UChar s_asterisk[]={'*',0};
static UChar s_attList[]={'a','t','t','L','i','s','t',0};
static UChar s_attDef[]={'a','t','t','D','e','f',0};
static UChar s_editorialDecl[]={'e','d','i','t','o','r','i','a','l','D','e','c','l',0};
static UChar s_segmentation[]={'s','e','g','m','e','n','t','a','t','i','o','n',0};
static UChar s_gilist[]={'g','i','l','i','s','t',0};
static UChar s_aklist[]={'a','k','l','i','s','t',0};
static UChar s_valSource[]={'v','a','l','S','o','u','r','c','e',0};
static UChar s_defaultVal[]={'d','e','f','a','u','l','t','V','a','l',0};
static UChar s_source[]={'s','o','u','r','c','e',0}; 
static UChar s_caseFold[]={'c','a','s','e','F','o','l','d',0};
static UChar s_true[]={'t','r','u','e',0};
static UChar s_name_par[]={'n','a','m','e','(',')',0};
static UChar s_joinTo[]={'j','o','i','n','T','o',0};
static UChar s_version[]={'v','e','r','s','i','o','n',0};
static UChar s_xara_bibdesc[]={'x','a','r','a',':','b','i','b','d','e','s','c',0};


int CSaraHeaderReader::Read(UnicodeString path)
{
	ZeroErrCount();
	m_bExists=false;
	FILE* f=iopen(path,s_rb);
	if (f==NULL) return 1;
	fclose(f);
	static XercesDOMParser d;
	d.setDoNamespaces(true);
//	SHandle e;
//	d.setErrorHandler(&e);
	d.parse((const xairo_uchar*) path.getTerminatedBuffer());
	if (d.getErrorCount()>0) return 2;
	m_doc=d.getDocument();
	DOMElement* el=m_doc->getDocumentElement();
	if (wcscmp(el->getTagName(),s_teiHeader) != 0) 
		HeaderError(HE_ROOT,false,L("Root of header is not teiHeader"));
	if (ErrCount() > 0) return 3;
	m_bExists=true;
	m_bUpdate=false;
	return 0;
}

UnicodeString CSaraHeaderReader::ipname[IP_MAX]={"CDATA","NONE","TAX","JOIN-FROM","JOIN-TO","CONTENT","LEMMA","CONTENTA"};
UnicodeString CSaraHeaderReader::ipnname[IP_MAX]={"content","none","taxonomy","joinfrom","jointo","CONTENT","LEMMA","CONTENTA"};

indexpol CSaraHeaderReader::LookupIP(const UChar* s)
{
	for (int i=0;i<IP_MAX;i++) {
		if (ipname[i]==s) return (indexpol) i;
	}
	return IP_MAX;
}

static UChar s_list[]={'l','i','s','t',0};
static UChar s_xaira_list[]={'x','a','i','r','a','L','i','s','t',0};
static UChar s_attlist[]={'a','t','t','l','i','s','t',0};
static UChar s_type[]={'t','y','p','e',0};
static UChar s_item[]={'i','t','e','m',0};
static UChar s_xaira_item[]={'x','a','i','r','a','I','t','e','m',0};
static UChar s_n[]={'n',0};
static UChar s_uri[]={'u','r','i',0};
static UChar s_desc[]={'d','e','s','c',0};
static UChar s_note[]={'n','o','t','e',0};
static UChar s_indexPol[]={'i','n','d','e','x','P','o','l',0};
static UChar s_attributePolicy[]={'a','t','t','r','i','b','u','t','e','P','o','l','i','c','y',0};
static UChar s_elementPolicy[]={'e','l','e','m','e','n','t','P','o','l','i','c','y',0};
static UChar s_elementSpec[]={'e','l','e','m','e','n','t','S','p','e','c',0};

void CSaraHeaderReader::GetAttUsage(DOMElement* e,int& n,_attusage* & a)
{
	n=0;
	if (e==NULL) return;
	if (e->getNodeType() != DOMNode::ELEMENT_NODE) return;
	if (wcscmp(e->getTagName(),s_list) != 0) return;
	if (wcscmp(e->getAttribute(s_type),s_attlist) != 0) return;
	DOMNodeList* l=e->getChildNodes();
	n=l->getLength(); _attusage* a0=mynew _attusage[n];
	int k=0; int i;
	for (i=0;i<n;i++) {
		DOMNode* m=l->item(i);
		if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* me=(DOMElement*) m;
		if (wcscmp(me->getTagName(),s_item)!=0) continue;
		a0[k].name=(const UChar*) me->getAttribute(s_n);
		a0[k].uri=(const UChar*) me->getAttribute(s_ns);
		DOMElement* gl=GetChildByTag(me,s_desc);
		a0[k].gloss="";
		if (gl != NULL) {
			a0[k].gloss=GetContentNS(gl);
		}
		k++;
		
	}
	a=mynew _attusage[k];
	n=k;
	for (i=0;i<k;i++) {
		a[i]=a0[i];
	}
	mydelete[] a0;

}



void CSaraHeaderReader::GetAttUsage2(DOMElement* e,int& n,_attusage* & a)
{
	DOMNodeList* l0=e->getElementsByTagName(s_attList);
	if (l0==NULL || l0->getLength()==0) return;
	DOMElement* al=(DOMElement*) l0->item(0);
	DOMNodeList* l=e->getElementsByTagName(s_attDef);
	n=l->getLength();
	_attusage* a0=mynew _attusage[n];
	int i;
	for (i=0;i<n;i++) {
		DOMElement* m=(DOMElement*) l->item(i);
		a0[i].name=(const UChar*) m->getAttribute(s_ident);
		a0[i].uri=(const UChar*) m->getAttribute(s_ns);
		DOMElement* gl=GetChildByTag(m,s_desc);
		a0[i].gloss="";
		if (gl != NULL) {
			a0[i].gloss=GetContentNS(gl);
		}
		GetCodeBook(&a0[i],m);
	}
	a=a0;

}


static UChar s_tagsDecl[]={'t','a','g','s','D','e','c','l',0};
static UChar s_tagUsage[]={'t','a','g','U','s','a','g','e',0};
static UChar s_gi[]={'g','i',0};
static UChar s_hash_global[]={'#','g','l','o','b','a','l',0};
static UChar s_occurs[]={'o','c','c','u','r','s',0};

void CSaraHeaderReader::GetTagUsageOld()
{
	m_ntags=0; m_nglobatts=0;
	if (!m_bExists) return;
	DOMNodeList* l=m_doc->getElementsByTagName(s_tagsDecl);
	if (l->getLength()==0) return;
	DOMNode* root=l->item(0); // should only be one but what the hell
	DOMNodeList* ll=root->getChildNodes();
	int k=ll->getLength();
	_tagusage* t0=mynew _tagusage [k];
//	m_tagusage=mynew _tagusage [k];
	int j=0; int i;
	for (i=0;i<k;i++) {
		DOMNode* n0=ll->item(i);
		if (n0->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* tag=(DOMElement*) n0;
		if (wcscmp(tag->getTagName(),s_tagUsage) != 0) continue;
		if (wcscmp(tag->getAttribute(s_gi),s_hash_global)==0)
		{
			DOMElement* sub=GetChildByTag(tag,s_list);
			GetAttUsage(sub,m_nglobatts,m_globals);
			continue;
		}
		t0[j].name=(const UChar*) tag->getAttribute(s_gi);
		t0[j].uri=(const UChar*) tag->getAttribute(s_ns);
		const UChar* occurs=(const UChar*) tag->getAttribute(s_occurs);
		t0[j].occurs=_wtoi((const xairo_uchar*) occurs);
		DOMElement* sub=GetChildByTag(tag,s_desc);
		t0[j].gloss="";
		if (sub != NULL) {
			t0[j].gloss=GetContentNS(sub);
		}
		sub=GetNamedList(tag,s_attlist); // should check type too
		GetAttUsage(sub, t0[j].natts,t0[j].atts);
		j++;
	}
	m_tagusage=mynew _tagusage[j];
	m_ntags=j;
	for (i=0;i<j;i++) {
		m_tagusage[i]=t0[i];
	}
//	m_ncont=0;
	m_lemma=0;
//	for (i=0;i<m_nglobatts;i++) {
//		if (m_globals[i].ip==IP_CONTENT||m_globals[i].ip==IP_LEMMA
//			||m_globals[i].ip==IP_CONTENTA) m_ncont++;
//	}
//	m_cont = mynew xairo_uchar* [m_ncont];
//	m_acont = mynew const char* [m_ncont];
//	for (i=0;i<m_nglobatts;i++) {
//		if (m_globals[i].ip==IP_CONTENT||m_globals[i].ip==IP_LEMMA||m_globals[i].ip==IP_CONTENTA) 
//			m_acont[i]=m_globals[i].name;
//		if (m_globals[i].ip==IP_LEMMA) 
//			m_lemma=i;
//	}
}

#define NBI 0

void CSaraHeaderReader::GetTagUsage()
{
	if (m_bConverting) {
		GetTagUsageOld();
		return;
	}
	m_ntags=0; m_nglobatts=0;
	if (!m_bExists) return;
	DOMElement* root=GetHeaderSection(HD_TAGUSAGE);
	if (root==NULL) {
		HeaderError(HE_MISSING_SECTION,false,L("No elementSpec"));
		return;
	}
	DOMNodeList* ll=root->getElementsByTagName(s_xaira_item);
	int k=ll->getLength();
	int t=NBI;
	for (int i=0;i<k;i++) {
		DOMElement* m=(DOMElement*) ll->item(i);
		const xairo_uchar* ident=m->getAttribute(s_ident);
		const xairo_uchar* type=m->getAttribute(s_type);
		if (wcscmp(type,s_element)==0) {
			if (wcscmp(ident,s_asterisk) != 0) t++;
		}
		else HeaderError(HE_BAD_ELT,true,L("Bad item type (%s) in element spec"),type);
	}
	_tagusage* t0=mynew _tagusage [t];
	m_ntags=t; t=0;
//	BuiltIns(t0);
	t=NBI;
	for (int i=0;i<k;i++) {
		DOMElement* m=(DOMElement*) ll->item(i);
		const xairo_uchar* ident=m->getAttribute(s_ident);
		const xairo_uchar* type=m->getAttribute(s_type);
		const xairo_uchar* ns=m->getAttribute(s_ns);
		if (ns==NULL) ns=s_;
		if (wcscmp(type,s_element)!=0) continue;
		if (wcscmp(ident,s_asterisk) == 0) {
			GetAttUsage2(m,m_nglobatts,m_globals);
			continue;
		}
		t0[t].name=ident;
		t0[t].uri=ns;
		t0[t].occurs=0;
		DOMElement* sub=GetChildByTag(m,s_desc);
		t0[t].gloss="";
		if (sub != NULL) {
			t0[t].gloss=GetContentNS(sub);
		}
		GetAttUsage2(m, t0[t].natts,t0[t].atts);
		t++;
	}
	m_tagusage=t0;

	m_lemma=0;
}

static UChar s_encodingDesc[]={'e','n','c','o','d','i','n','g','D','e','s','c',0};
static UChar s_fileDesc[]={'f','i','l','e','D','e','s','c',0};
static UChar s_sysid_par[]={'s','y','s','i','d','(',')',0};
static UChar s_lineno_par[]={'l','i','n','e','n','o','(',')',0};
static UChar s_count_par[]={'c','o','u','n','t','(',')',0};

DOMElement* CSaraHeaderReader::GetEncodingDesc()
{
	DOMElement* n=GetUniqueNode(s_teiHeader);
	if (n==NULL) return NULL;
	DOMNodeList* l=n->getChildNodes();
	int k=l->getLength();
	DOMNode* mf=NULL;
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		if (wcscmp(m->getNodeName(),s_encodingDesc)==0) 
			return (DOMElement*) m;
		if (wcscmp(m->getNodeName(),s_fileDesc)==0) 
			mf=m;
	}
	if (mf) mf=mf->getNextSibling();
	DOMElement* m0=m_doc->createElement(s_encodingDesc);
	n->insertBefore(m0,mf);
	return m0;
}




static UChar s_refsDecl[]={'r','e','f','s','D','e','c','l',0};
static UChar s_xara_refspec[]={'x','a','r','a',':','r','e','f','s','p','e','c',0};
static UChar s_refSpec[]={'r','e','f','S','p','e','c',0};
static UChar s_p[]={'p',0};
static UChar s_text[]={'t','e','x','t',0};
static UChar s_unit[]={'u','n','i','t',0};
static UChar s_ans[]={'a','n','s',0};
static UChar s_format[]={'f','o','r','m','a','t',0};
static UChar s_xml_id[]={'x','m','l',':','i','d',0};
static UChar s_irefs[]={'I','n','d','e','x','t','o','o','l','s',' ','m','a','i','n','t','a','i','n','e','d',' ',
'l','i','s','t',' ','o','f',' ','r','e','f','e','r','e','n','c','e','s',0};
static UChar s_textRef[]={'t','e','x','t','R','e','f',0};
static UChar s_unitRef[]={'u','n','i','t','R','e','f',0};
static UChar s_scopeRef[]={'s','c','o','p','e','R','e','f',0};
static UChar s_labelGen[]={'l','a','b','e','l','G','e','n',0};
static UChar s_change[]={'c','h','a','n','g','e',0};
static UChar s_onStart[]={'o','n','S','t','a','r','t',0};

void CSaraHeaderReader::SetLabelsOld()
{
	DOMElement* root=SetHeaderSection(HD_REFS);
	for (int i=0;i<m_nlabels;i++) {
		_lab* l=&m_labels[i];
		DOMElement* m=m_doc->createElement(s_xaira_item);
		m->setAttribute(s_n,i==0?s_text:s_unit);
		DOMElement* mm=m_doc->createElement(s_ident);
		UnicodeString c, t;
		EncodeLabel(l,t,c);
		mm->setAttribute(s_type,(const xairo_uchar*) t.getTerminatedBuffer());
		if (!l->ns.isEmpty()) {
			mm->setAttribute(s_ns,(const xairo_uchar*) l->ns.getTerminatedBuffer());
		}
		if (!l->ans.isEmpty()) {
			mm->setAttribute(s_ans,(const xairo_uchar*) l->ans.getTerminatedBuffer());
		}
		DOMText* tx=m_doc->createTextNode((const xairo_uchar*) c.getTerminatedBuffer());
		m->appendChild(mm);
		mm->appendChild(tx);
		root->appendChild(m);
		if (!l->fmt.isEmpty()) {
			DOMElement* nn=m_doc->createElement(s_note);
			nn->setAttribute(s_type,s_format);
			DOMText* nx=m_doc->createTextNode((const xairo_uchar*) l->fmt.getTerminatedBuffer());
			nn->appendChild(nx);
			m->appendChild(nn);
		}
	}
/*	DOMElement* m=m_doc->createElement(L"item");
	m->setAttribute(L"n",L"ulfs");
	DOMText* tx=m_doc->createTextNode(m_ulfs.getTerminatedBuffer());
	m->appendChild(tx);
	root->appendChild(m);
*/}

void CSaraHeaderReader::SetLabels()
{
	DOMElement* root=SetHeaderSection(HD_REFS);
	for (int i=0;i<m_nlabels;i++) {
		_lab* l=&m_labels[i];
		DOMElement* m=m_doc->createElement(s_xaira_item);
		const xairo_uchar* ty=s_;
		if (i==0) ty=s_textRef; else ty=s_unitRef;
		if (i > 0 && l->isScope) ty=s_scopeRef;
		m->setAttribute(s_type,ty);
		_valsource vs;
		if (l->isSpec) {
			vs.type=s_pseudo;
			if (l->e==SPE_FILE)
				vs.ident=s_sysid_par;
			else 
				vs.ident=s_lineno_par;
		} else if (l->isAuto) {
			vs.type=s_pseudo;
			vs.ident=s_count_par;
		} else {
			vs.type=s_attribute;
			vs.ns=l->ans.getTerminatedBuffer();
			vs.ident=l->att.getTerminatedBuffer();
		}
		vs.ngi=l->ne;
		vs.gi=new const xairo_uchar* [l->ne];
		for (int jj=0;jj<l->ne;jj++) {
			vs.gi[jj]=l->elt[jj].getTerminatedBuffer();
		}
		DOMElement* mm=MakeValSource(&vs);
		m->appendChild(mm);
		if (!l->fmt.isEmpty() || l->isMilestone) {
			DOMElement* nn=m_doc->createElement(s_labelGen);
			if (l->isMilestone) nn->setAttribute(s_change,s_onStart);
			DOMText* nx=m_doc->createTextNode((const xairo_uchar*) l->fmt.getTerminatedBuffer());
			nn->appendChild(nx);
			m->appendChild(nn);
		}
		root->appendChild(m);
	}
/*	DOMElement* m=m_doc->createElement(L"item");
	m->setAttribute(L"n",L"ulfs");
	DOMText* tx=m_doc->createTextNode(m_ulfs.getTerminatedBuffer());
	m->appendChild(tx);
	root->appendChild(m);
*/}

static UChar s_hash_sysid[]={'#','s','y','s','i','d',0};
static UChar s_hash_line[]={'#','l','i','n','e',0};
static UChar s_pc_at_pc[]={'%','s','/','@','%','s',0};
//static UChar s_hash_global[]={'#','g','l','o','b','a','l',0};

void CSaraHeaderReader::EncodeLabel(_lab* l,UnicodeString& t,UnicodeString& c) 
{
		if (l->isScope) c="*";
		if (l->isSpec) {
			t="kw";
			c+=(const UChar*) ((l->e==SPE_FILE)?s_hash_sysid:s_hash_line);
		} else if (l->isAuto) {
			t="gi-seq";
			if (l->isMilestone) c+="+";
			c+=StringsToString(l->ne,l->elt);
		} else {
			xairo_uchar x[100];
			t="att";
			if (l->bGlob)
				swprintf(x,s_pc_at_pc,s_hash_global,l->att.getTerminatedBuffer());
			else
				swprintf(x,s_pc_at_pc,StringsToString(l->ne,l->elt).getTerminatedBuffer(),l->att.getTerminatedBuffer());
			if (l->isMilestone) c+="+";
			c+=(const UChar*) x;
		}

}

static UChar s_kw[]={'k','w',0};

void CSaraHeaderReader::GetLabelsOld()
{
	if (!m_bExists) return;
	DOMElement* root=GetHeaderSection(HD_REFS);
	int k=0;
	DOMNodeList* ll=NULL;
	if (root) {
		ll=root->getChildNodes();
		k=ll->getLength();
	}
	int tt=0; int tu=0;
	m_nulllab="none";
	for (int i=0;i<k;i++) {
		DOMNode* nn=ll->item(i);
		if (nn->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* nne=(DOMElement*) nn;
		if (wcscmp(nne->getTagName(),s_item)!=0 && wcscmp(nne->getTagName(),s_xaira_item)!=0) continue;
		if (wcscmp(nne->getAttribute(s_n),s_text)==0)
		{
			tt++;
		}
		else if (wcscmp(nne->getAttribute(s_n),s_unit)==0) 
		{
			tu++;
		}
	}
	if (tu==0) tu++;
	m_nlabels=tu+1;
	m_labels=mynew _lab [tu+1];
	tt=0; tu=0; bool bSc=false;
	for (int i=0;i<k;i++) {
		DOMNode* nn=ll->item(i);
		if (nn->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* nne=(DOMElement*) nn;
		if (wcscmp(nne->getTagName(),s_item)!=0 && wcscmp(nne->getTagName(),s_xaira_item)!=0) continue;
		const xairo_uchar* ty=nne->getAttribute(s_n);
		_lab* l;
		if (wcscmp(ty,s_text)==0) {
			if (tt>0) continue;
			l=&m_labels[0];
			tt++;
		}
		else if (wcscmp(ty,s_unit)==0) {
			l=&m_labels[tu+1];
			tu++;
		}
/*		else if (wcscmp(ty,L"ulfs")==0) {
			m_ulfs=GetContent(nne);
			continue;
		}
*/		else continue;
		DOMElement* me=GetChildByTag(nn,s_ident);
		ty=s_kw;
		if (me) ty=me->getAttribute(s_type);
		if (me) l->ns=(const UChar*) me->getAttribute(s_ns);
		if (me) l->ans=(const UChar*) me->getAttribute(s_ans);
		const xairo_uchar* c=s_;
		if (me) c=GetContent(me);
		DecodeLabel(ty,c,l,bSc);
		DOMElement* mn=GetChildByTag(nn,s_note);
		if (mn) l->fmt=(const UChar*) GetContent(mn);
	}
	if (tt==0) {
		_lab* l=&m_labels[0];
		l->isSpec=true;
		l->e=SPE_FILE;
	}
	if (tu==0) {
		_lab* l=&m_labels[1];
		l->isSpec=true;
		l->e=SPE_LINE;
	}
	if (!bSc) {
		_lab* l=&m_labels[m_nlabels-1];
		l->isScope=true;
	}
}

void CSaraHeaderReader::GetLabels()
{
	if (m_bConverting) {
		GetLabelsOld();
		return;
	}
	if (!m_bExists) return;
	DOMElement* root=GetHeaderSection(HD_REFS);
	if (root==NULL) return;
	DOMNodeList* ll=root->getElementsByTagName(s_xaira_item);
	int k=ll->getLength();
	int tt=0; int tu=0;
	m_nulllab="none";
	for (int i=0;i<k;i++) {
		DOMElement* nn=(DOMElement*) ll->item(i);
		const xairo_uchar* ref=nn->getAttribute(s_type);
		if (wcscmp(ref,s_textRef)==0)
		{
			tt++;
		}
		else if (wcscmp(ref,s_unitRef)==0||wcscmp(ref,s_scopeRef)==0) 
		{
			tu++;
		}
	}
	if (tu==0) tu++;
	m_nlabels=tu+1;
	m_labels=mynew _lab [tu+1];
	tt=0; tu=0; bool bSc=false;
	for (int i=0;i<k;i++) {
		DOMElement* nn=(DOMElement*) ll->item(i);
		const xairo_uchar* ty=nn->getAttribute(s_type);
		_lab* l;
		if (wcscmp(ty,s_textRef)==0) {
			if (tt>0) continue;
			l=&m_labels[0];
			tt++;
		}
		else if (wcscmp(ty,s_unitRef)==0) {
			l=&m_labels[tu+1];
			tu++;
		}
		else if (wcscmp(ty,s_scopeRef)==0) {
			l=&m_labels[tu+1];
			tu++;
			bSc=true;
			l->isScope=true;
		}
		else continue;
		DOMElement* vs=GetChildByTag(nn,s_valSource);
		_valsource v;
		GetValSource(vs,&v);
		DecodeLabel2(&v,l);
		DOMElement* mn=GetChildByTag(nn,s_labelGen);
		if (mn) {
			l->fmt=(const UChar*) GetContent(mn);
			const xairo_uchar* os=mn->getAttribute(s_change);
			if (os && wcscmp(os,s_onStart)==0) l->isMilestone=true;
		}
	}
	if (tt==0) {
		_lab* l=&m_labels[0];
		l->isSpec=true;
		l->e=SPE_FILE;
	}
	if (tu==0) {
		_lab* l=&m_labels[1];
		l->isSpec=true;
		l->e=SPE_LINE;
	}
	if (!bSc) {
		_lab* l=&m_labels[m_nlabels-1];
		l->isScope=true;
	}
}

static UChar s_gi_seq[]={'g','i','-','s','e','q',0};
static UChar s_att[]={'a','t','t',0};
static UChar s_attc[]={'a','t','t','c',0};

void CSaraHeaderReader::DecodeLabel2(_valsource* v,_lab* l)
{
	if (wcscmp(v->type,s_pseudo)==0) {
		if (wcscmp(v->ident,s_sysid_par)==0) {
			l->isSpec=true;
			l->e=SPE_FILE;
		}
		else if (wcscmp(v->ident,s_lineno_par)==0) {
			l->isSpec=true;
			l->e=SPE_LINE;
		}
		else if (wcscmp(v->ident,s_count_par)==0) {
			l->isAuto=true;
		}
	} else {
			l->att=v->ident;
			l->ans=v->ns;
	}
	l->ne=v->ngi;
	l->elt=new UnicodeString [l->ne];
	for (int i=0;i<l->ne;i++) {
		l->elt[i]=v->gi[i];
	}
	l->bGlob=(l->ne==0);
}

void CSaraHeaderReader::DecodeLabel(const xairo_uchar* ty,const xairo_uchar* c,_lab* l,bool& bSc)
{
		xairo_uchar a[100]={0};
		xairo_uchar e[100]={0};
		bool bM=false;
		bool bTxt=(wcscmp(ty,s_text)==0);
		if (c[0]=='*') {
			l->isScope=true;
			bSc=true;
			c++;
		}
		if (c[0]=='+') {
			bM=true;
			c++;
		}
		const xairo_uchar* kx=wcschr(c,'/');
		if (kx) {
			wcsncpy(e,c,kx-c);
			e[kx-c]=0;
			if (wcslen(kx)<2)
				kx=NULL;
			else
				wcscpy(a,kx+2);
		}
		if (wcscmp(ty,s_gi_seq)==0) {
			l->isAuto=true;
			l->ne=0;
			if (wcscmp(c,s_hash_global)==0) 
				l->bGlob=true;
			else
				StringToStrings((const UChar*) c,l->ne,l->elt);
			l->isMilestone=bM;
		} else if (wcscmp(ty,s_att)==0 && kx != NULL) {
			l->ne=0;
			if (wcscmp(e,s_hash_global)==0) 
				l->bGlob=true;		
			StringToStrings((const UChar*) e,l->ne,l->elt);
			l->att=_wcsdup(a);
			l->isMilestone=bM;
		} else // if (wcscmp(ty,L"kw")==0) 
		{
			l->isSpec=true;
			if (wcscmp(c,s_hash_sysid)==0)
				l->e=SPE_FILE;
			else
				l->e=SPE_LINE;
		}
}


DOMElement* GetChildByTag(DOMNode* e,const XMLCh* name)
{
	DOMNodeList* l=e->getChildNodes();
	int n=l->getLength();
	for (int i=0;i<n;i++) {
		DOMNode* m=l->item(i);
		if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* me=(DOMElement*) m;
		if (wcscmp(me->getTagName(),name)==0) return me;
	}
	return NULL;
}

const XMLCh* GetContent(DOMNode* e)
{
	DOMNodeList* l=e->getChildNodes();
	int n=l->getLength();
	for (int i=0;i<n;i++) {
		DOMNode* m=l->item(i);
		if (m->getNodeType() != DOMNode::TEXT_NODE) continue;
		DOMText* me=(DOMText*) m;
		return me->getData();
	}
	return s_;
}

unsigned CSaraHeaderReader::GetIP0(UnicodeString e,UnicodeString url)
{
	int i;
	for (i=0;i<m_nwtags;i++) {
		if (e==m_wtags[i].s) return EIP_MAKE(EIP_NO,EIP_INHERIT,EIP_INHERIT);
	}
	for (i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		if (e==t->name && url==t->uri) return t->ip;
	}
	return EIP_DEFAULT;
}

indexpol CSaraHeaderReader::GetIP(UnicodeString e,UnicodeString a) 
{
	_tagusage* t;
	int i;
	for (i=0;i<m_nglobatts;i++) {
		_attusage* au=&m_globals[i];
		if (a==au->name) return au->ip;
	}
	for (i=0;i<m_ntags;i++) {
		t=&m_tagusage[i];
		if (e==t->name) break;
	}
	if (i==m_ntags) return IP_MAX;
	for (i=0;i<t->natts;i++) {
		_attusage* au=&t->atts[i];
		if (a==au->name) return au->ip;
	}
	return IP_MAX;
}
/*
indexpol0 CSaraHeaderReader::GetIP0(UnicodeString e) 
{
	_tagusage* t;
	for (int i=0;i<m_ntags;i++) {
		t=&m_tagusage[i];
		if (e==t->name) return t->ip;
	}
	return IP0_CONTENT;
}
*/
int CSaraHeaderReader::GetTaxClasses(UnicodeString s,UnicodeString*& vals)
{
	int n=0;
	UnicodeString t; UnicodeString ss=s;
	while (t=GetToken(ss),!t.isEmpty()) {
		n++;
	}
	vals=mynew UnicodeString [n];
	int i=0;
	while (t=GetToken(s),!t.isEmpty()) {
		vals[i++]=t;
	}
	return n;
}

bool CSaraHeaderReader::GetTaxClass(_taxonomy* t,int n,int &nj,UnicodeString* vals,bool* bUsed,UnicodeString& x)
{
	int dmax; int d; nj=0; bool b=false;
	for (int i=0;i<n;i++) {
		dmax=-1;
		for (int j=0;j<t->n;j++) {
			if (FindClass(vals[i],&t->l[j],0,d) && d>dmax) {
				x=vals[i];
				dmax=d;
				nj=j;
				b=true;
				bUsed[i]=true;
			}
		}
	}
	if (x.isEmpty()) {
		x=t->unclass;
		nj=t->n;
	}
	return b;
}



bool CSaraHeaderReader::FindClass(UnicodeString s,_class* c,int d,int& dd)
{
	dd=d;
//	if (c->n==0) return s==c->id?true:false; // crappy ICU types
	if (s==c->id) return true;
	if (c->n==0) return false;
	for (int i=0;i<c->n;i++) {
		if (FindClass(s,&c->l[i],d+1,dd)) return true;
	}
	return false;
}


int CSaraHeaderReader::CountTax()
{
	int k=0;
	for (int i=0;i<m_ntax;i++) {
		_taxonomy* t=&m_taxonomies[i];
		k+=CountTax(t->n,t->l);
	}
	return k;
}

int CSaraHeaderReader::CountTax(int n,_class* cl)
{
	int k=1;
	for (int i=0;i<n;i++) {
		_class* c0=&cl[i];
		k+=CountTax(c0->n,c0->l);
	}
	return k;
}


/*
int CSaraHeaderReader::GetTaxClass(int k,UnicodeString* c)
{
	int k=0;
	for (int i=0;i<m_ntax;i++) {
		_taxonomy* t=&m_taxonomies[i];
		k+=GetTaxClass(c,t->n,t->l,t->id,cl);
	}
	return k;	
}

int CSaraHeaderReader::GetTaxClass(UnicodeString c,int n,_class* cl, UnicodeString id0)
{
	for (int i=0;i<n;i++) {
		_class* c0=&cl[i];
		if (c==c0->id) {
			id=id0;
			return i;
		}
		int k=GetTaxClass(c,c0->n,c0->l,c0->id,id);
		if (k >= 0) return k;
	}
	return -1;
}
*/

static UChar s_category[]={'c','a','t','e','g','o','r','y',0};
static UChar s_catDesc[]={'c','a','t','D','e','s','c',0};
static UChar s_id[]={'i','d',0};

int CSaraHeaderReader::GetClassList(DOMElement* e,_class*& cl)
{
	DOMNodeList* l=e->getChildNodes();
	int k=l->getLength();
	int n=0; int i;
	for (i=0;i<k;i++) {
		DOMNode* nd=l->item(i);
		if (nd->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ee=(DOMElement*) nd;
		if (wcscmp(ee->getTagName(),s_category) != 0) continue;
		n++;
	}
	if (n==0) return NULL;
	cl =mynew _class [n];
	n=0;
	for (i=0;i<k;i++) {
		DOMNode* nd=l->item(i);
		if (nd->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ee=(DOMElement*) nd;
		if (wcscmp(ee->getTagName(),s_category) != 0) continue;
		_class* c=&cl[n];
		c->id=(const UChar*) ee->getAttributeNS(s_xmlns,s_id);
		DOMNodeList* l1=ee->getElementsByTagName(s_catDesc);
		if (l1->getLength() > 0) {
			DOMNode* n=l1->item(0);
			c->desc=GetContentNS(n);
		} else c->desc=c->id;
		c->n=GetClassList(ee,c->l);
		n++;
	}
	return n;
}

static UChar s_classDecl[]={'c','l','a','s','s','D','e','c','l',0};
static UChar s_taxonomy[]={'t','a','x','o','n','o','m','y',0};

void CSaraHeaderReader::GetTaxonomies()
{
	if (!m_bExists) return;
//	m_taxonomies.RemoveAll();
	m_ntax=0;
	DOMNodeList* l=m_doc->getElementsByTagName(s_classDecl);
	if (l->getLength()==0) return;
	DOMNode* root=l->item(0); // should only be one but what the hell
	DOMNodeList* ll=root->getChildNodes();
	int k=ll->getLength();
	int kk=0;
	for (int i=0;i<k;i++) {
		DOMNode* n=ll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) n;
		if (wcscmp(e->getTagName(),s_taxonomy) != 0) continue;
		const XMLCh* id=e->getAttributeNS(s_xmlns,s_id);
		if (id[0]==0) continue;
		if (GetChildByTag(e,s_category)==NULL) continue;
		kk++;
	}
	m_taxonomies=mynew _taxonomy[kk];
	m_ntax=kk;
	kk=0;
	for (int i=0;i<k;i++) {
		DOMNode* n=ll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) n;
		if (wcscmp(e->getTagName(),s_taxonomy) != 0) continue;
		if (GetChildByTag(e,s_category)==NULL) continue;
		const XMLCh* id=e->getAttributeNS(s_xmlns,s_id);
		if (id[0]==0) continue;
		const XMLCh* desc=NULL;
		DOMNodeList* l1=e->getElementsByTagName(s_desc);
		if (l1->getLength() > 0) {
			DOMNode* n=l1->item(0);
			desc=_wcsdup(GetContentNS(n).getTerminatedBuffer());
		} else desc=id;
		_taxonomy* t =&m_taxonomies[kk];
		t->id=(const UChar*) id;
		t->desc=(const UChar*) desc;
		t->unclass="#unc";
		int nk=GetClassList(e,t->l);
//		if (nk==0) continue;
		t->n=nk;
		kk++;
	}
}


bool CSaraHeaderReader::IsGlobal(const xairo_uchar* e,const xairo_uchar* u,const xairo_uchar* s,const xairo_uchar* ua)
{
	for (int i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		if (t->name != e || t->uri != u) continue;
		int j; 
		for (j=0; j<t->natts;j++) {
			_attusage* at0=&t->atts[j];
			if (at0->name==s) return false;
		}
	}
	for (int i=0;i<m_nglobatts;i++) {
		_attusage* at0=&m_globals[i];
		if (at0->name==s && at0->uri==ua) {
			return true;
		}
	}
	return false;

}

int CSaraHeaderReader::TestWTag(const xairo_uchar* s)
{
	if (m_nwtags==0) return 0;
	for (int i=0;i<m_nwtags;i++) {
		if (m_wtags[i].s==(const UChar*) s) return 1;
	}
	return 2;
}

bool CSaraHeaderReader::TestWB(const xairo_uchar* s)
{
	for (int i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		if (t->name==(const UChar*) s) return t->wb;
	}
	return false;
}

// Hey! This was completely unrecursive.
int CSaraHeaderReader::GetSCPart(_taxonomy* tax,UnicodeString val)
{
	int n=tax->n;
	_class* l=tax->l;
	int k=0; int kk;
	for (int i=0;i<n;i++) {
		if (ClassToInt(&l[i],val,k,kk)) return kk+1;
	}
	return k+1;
}

bool CSaraHeaderReader::ClassToInt(_class* l,UnicodeString val,int& k0,int& k)
{
	if (val==l->id) {
		k=k0;
		return true;
	}
	k0++;
	for (int i=0;i<l->n;i++) {
		if (ClassToInt(&l->l[i],val,k0,k)) return true;
	}
	return false;
}

bool CSaraHeaderReader::IsTax(UnicodeString tax)
{
	for (int i=0;i<m_ntax;i++)
		if (tax==m_taxonomies[i].id) return true;
	return false;
}

//#define MAXSTRING 100000

const char* Ascii(UnicodeString w)
{
	char* c=mynew char[w.length()+1];
	w.extract(0,w.length(),c);
	return c;
}

const char* Ascii(const XMLCh* w)
{
//	static char a[MAXSTRING];
	int k=wcslen(w);
	char* a=mynew char [k+1];
//	assert (k<MAXSTRING);
	wcstombs(a,w,k+1);
	return a;
}

UnicodeString Wide(const char* s)
{
	return UnicodeString(s);
}

// the mynew structure for addkeys
//
// char* name
// char* desc
// enum akproc proc
// char* gi
// char* att
// bool isLemma;
// enum defproc dproc
// xairo_uchar* defval
// xairo_uchar* cache
//


void CSaraHeaderReader::DefAddKey(_addkey* k)
{
	k->name="form";
	k->desc="default";
	k->proc=_addkey::AKP_FORM;
	k->isLemma=true;
	k->bCaseSensitive=true;
	k->isPseudo=false;
}


static UChar SSV[]={'1','.','0','0',0};
static UChar s_specification[]={'s','p','e','c','i','f','i','c','a','t','i','o','n',0};
static UChar s_xaira_specification[]={'x','a','i','r','a','S','p','e','c','i','f','i','c','a','t','i','o','n',0};
static UChar s_xara_keyspec[]={'x','a','r','a',':','k','e','y','s','p','e','c',0};
static UChar s_keySpec[]={'k','e','y','S','p','e','c',0};
static UChar s_addkey[]={'a','d','d','k','e','y',0};
static UChar s_addKey[]={'a','d','d','K','e','y',0};
static UChar s_lemma[]={'l','e','m','m','a',0};
static UChar s_ls[]={'l','s',0};
static UChar s_lemma0[]={'l','e','m','m','a','0',0};
static UChar s_lemmaSpec[]={'l','e','m','m','a','S','p','e','c',0};
static UChar s_regionSpec[]={'r','e','g','i','o','n','S','p','e','c',0};
static UChar s_form[]={'f','o','r','m',0};
static UChar s_label[]={'l','a','b','e','l',0};
static UChar s_ifk[]={'I','n','d','e','x','t','o','o','l','s',' ','m','a','i','n','t','a','i','n','e','d',' ',
'l','i','s','t',' ','o','f',' ','f','o','r','m','s',' ','a','n','d',' ','k','e','y','s',0};
static UChar s_ils[]={'I','n','d','e','x','t','o','o','l','s',' ','m','a','i','n','t','a','i','n','e','d',' ',
'l','i','s','t',' ','o','f',' ','l','e','m','m','a',' ','s','c','h','e','m','e','s',0};
static UChar s_irs[]={'I','n','d','e','x','t','o','o','l','s',' ','m','a','i','n','t','a','i','n','e','d',' ',
'l','i','s','t',' ','o','f',' ','r','e','g','i','o','n','s',0};

bool CSaraHeaderReader::OldStyle()
{
	DOMElement* n=GetUniqueNode(s_encodingDesc);
	DOMNodeList* l=m_doc->getElementsByTagName(s_xaira_specification);
	return l->getLength()==0;

}

DOMElement* CSaraHeaderReader::GetXairaRoot()
{
	DOMElement* n=GetUniqueNode(s_encodingDesc);
	DOMNodeList* l=m_doc->getElementsByTagName(s_xaira_specification);
	int k=l->getLength();
	if (k) return (DOMElement*) l->item(0);
	DOMElement* n1=m_doc->createElement(s_xaira_specification);
	n1->setAttribute(s_version,SSV);
	n->appendChild(n1);
	return n1;
}



DOMElement* CSaraHeaderReader::GetHeaderSection(hd_section h)
{
	switch (h) {
		case HD_TAGUSAGE:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
//			n1=GetNamedList(n,s_xara_elementSpec);
//			if (n1) return n1;
			n1=GetNamedList2(n,s_elementSpec);
			if (n1) return n1;
//			HeaderError(HE_MISSING_SECTION,false,L("No elementSpec"));
			return NULL;
			}
		case HD_ADDKEYS:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
			n1=GetNamedList(n,s_xara_keyspec);
			if (n1) return n1;
			n1=GetNamedList2(n,s_keySpec);
			if (n1) return n1;
//			HeaderError(HE_MISSING_SECTION,true,L("No keySpec")); not an error
			return NULL;
			}
		case HD_LEMMASCHEMES:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
//			n1=GetNamedList(n,s_xara_keyspec);
//			if (n1) return n1;
			n1=GetNamedList2(n,s_lemmaSpec);
			if (n1) return n1;
//			HeaderError(HE_MISSING_SECTION,"No lemmaSpec"); not an error
			return NULL;
			}
		case HD_REGIONS:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
//			n1=GetNamedList(n,s_xara_keyspec);
//			if (n1) return n1;
			n1=GetNamedList2(n,s_regionSpec);
			if (n1) return n1;
//			HeaderError(HE_MISSING_SECTION,"No lemmaSpec"); not an error
			return NULL;
			}
		case HD_REFS:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
			n1=GetNamedList(n,s_xara_refspec);
			if (n1) return n1;
			n1=GetNamedList2(n,s_refSpec);
			if (n1) return n1;
//			HeaderError(HE_MISSING_SECTION,false,L("No refSpec"));
			return NULL;
			}
		case HD_IPS:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=NULL;
			n1=GetNamedList(n,s_xara_indexspec);
			if (n1) return n1;
			n1=GetNamedList2(n,s_indexSpec);
			if (n1) return n1;
			return NULL;
			}
		case HD_CODEBOOKS:
			{
			DOMElement* n=GetUniqueNode(s_encodingDesc);
			if (n==NULL) return NULL;
			DOMElement* n1=GetNamedList2(n,s_codebooks);
			if (n1) return n1;
			return n;
			}
		case HD_BIBTAGS:
			{
				DOMNodeList* nl=m_doc->getElementsByTagName(s_btagsDecl);
				int k=nl->getLength();
				if (k==NULL) return 0;
				return (DOMElement*) nl->item(0);
			}
		case HD_BIBRULES:
			{
				DOMElement* n=GetUniqueNode(s_fileDesc);
				if (n==NULL) return NULL;
				DOMElement* nn=GetNamedNote(n,s_xara_bibdesc);
				return nn;
			}

	}
	return NULL;
}

DOMElement* CSaraHeaderReader::SetHeaderSection(hd_section h)
{
	DOMElement* old=GetHeaderSection(h);
	if (old && wcscmp(old->getNodeName(),s_encodingDesc) != 0) {
		DOMNode* op=old->getParentNode();
		op->removeChild(old);
	} else if (old && h==HD_CODEBOOKS)
	{
		DOMNodeList* nl=old->getElementsByTagName(s_list);
		int k=nl->getLength();
		for (int i=k-1;i>=0;i--) {
			DOMElement* nn=(DOMElement*) nl->item(i);
			if (wcscmp(nn->getAttribute(s_type),s_xara_codebook) != 0) continue;
			DOMNode* parent=nn->getParentNode();
			parent->removeChild(nn);
		}
	}
	switch (h) {
		case HD_TAGUSAGE:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_elementSpec);
			n->appendChild(n1);
//			DOMComment* c=m_doc->createComment(s_ifk);
//			n1->appendChild(c);
			return n1;
			}
		case HD_ADDKEYS:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_keySpec);
			n->appendChild(n1);
			DOMComment* c=m_doc->createComment(s_ifk);
			n1->appendChild(c);
			return n1;
			}
		case HD_LEMMASCHEMES:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_lemmaSpec);
			n->appendChild(n1);
			DOMComment* c=m_doc->createComment(s_ils);
			n1->appendChild(c);
			return n1;
			}
		case HD_REGIONS:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_regionSpec);
			n->appendChild(n1);
			DOMComment* c=m_doc->createComment(s_irs);
			n1->appendChild(c);
			return n1;
			}
		case HD_REFS:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_refSpec);
			n->appendChild(n1);
			DOMComment* c=m_doc->createComment(s_irefs);
			n1->appendChild(c);
			return n1;
			}
		case HD_IPS:
			{
			DOMElement* n=GetXairaRoot();
			DOMElement* n1=m_doc->createElement(s_xaira_list);
			n1->setAttribute(s_type,s_indexSpec);
			n->appendChild(n1);
			DOMComment* c=m_doc->createComment(s_iip);
			n1->appendChild(c);
			return n1;
			}
		case HD_CODEBOOKS:
			return NULL;
		case HD_BIBRULES:
			return NULL;
	}
	return NULL;
}




void CSaraHeaderReader::GetAddKeys()
{
	if (m_bConverting) {
		GetAddKeysOld();
		return;
	}
	DOMElement* root=GetHeaderSection(HD_ADDKEYS);
	int n=0;
	DOMNodeList* l=NULL;
	if (root) {
		l=root->getElementsByTagName(s_xaira_item);
		n=l->getLength();
	}
	int tt=0; int nk0=-1;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_form)!=0) continue;
		DOMElement *vs=GetChildByTag(e,s_valSource);
		_valsource v;
		if (vs) GetValSource(vs,&v);
		if (wcscmp(v.type,s_element)==0) tt++;
		else nk0=i;
	}
	m_nwtags=tt;
	m_wtags=mynew _wtag [tt];
	tt=0;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_form)!=0) continue;
		DOMElement *vs=GetChildByTag(e,s_valSource);
		_valsource v;
		if (vs) GetValSource(vs,&v);
		if (wcscmp(v.type,s_element)==0) 
			m_wtags[tt].s=v.ident;
		tt++;
	}
	int t=1;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_addKey)==0 && i != nk0) t++;
	}
	_addkey* a=m_addkeys=mynew _addkey [t];
	m_naddkeys=t; t=(nk0<0)?1:0;
	if (nk0<0) DefAddKey(&a[0]);
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_addKey)!=0 && i != nk0) continue;
		const xairo_uchar* ident=e->getAttribute(s_ident);
		if (ident[0]==0) {
			HeaderError(HE_NOIDENT,true,L("Addkey %d has no ident"),t+1);
			xairo_uchar x[10];
			swprintf(x,L("k%d"),t);
			ident=_wcsdup(x);
		}
		_addkey* ak=&a[t];
		ak->name=ident;
		DOMElement *g=GetChildByTag(e,s_desc);
		if (g) ak->desc=GetContent(g);
		DOMElement *vs=GetChildByTag(e,s_valSource);
		_valsource v;
		if (vs) GetValSource(vs,&v);
		if (wcscmp(v.type,s_pseudo)==0) {
			ak->isPseudo=true;
		} else if (wcscmp(v.type,s_attribute)==0) {
			ak->isPseudo=false;
			ak->att=v.ident;
			ak->ans=v.ns;
		}
		ak->bCaseSensitive=!v.bcasefold;
		ak->ngi=v.ngi;
		ak->gi=new UnicodeString [v.ngi];
		for (int i=0;i<v.ngi;i++) {
			ak->gi[i]=v.gi[i];
		}
		if (v.bdefform) {
			ak->dproc=_addkey::DP_USEFORM;
		}
		else {
			ak->dproc=_addkey::DP_USEVAL;
			ak->defval=v.def;
		}
		t++;
	}
	DOMElement* ls0=GetHeaderSection(HD_LEMMASCHEMES);
	if (ls0!=NULL) {
	l=ls0->getElementsByTagName(s_xaira_item);
	n=l->getLength();
	t=0;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_lemmaScheme)==0) t++;
	}
	_ls* ls = m_ls=mynew _ls [t];
	m_nls=t; t=0;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_lemmaScheme)!=0) continue;
		_ls* ls=&m_ls[t];
		ls->name=e->getAttribute(s_ident);
		DOMElement *g=GetChildByTag(e,s_desc);
		if (g) ls->desc=GetContent(g);
		DOMElement* nl=GetChildByTag(e,s_nameList);
		DOMNodeList* ll=nl->getElementsByTagName(s_ident);
		ls->n=ll->getLength();
		ls->a=new UnicodeString [ls->n];
		for (int i=0;i<ls->n;i++) {
			DOMElement* k=(DOMElement*) ll->item(i);
			ls->a[i]=GetContent(k); int j; _addkey* ak=NULL;
			for (j=0;j<m_naddkeys;j++) {
				ak=&m_addkeys[j];
				if (ls->a[i]==ak->name) break;
			}
			if (j >= m_naddkeys) {
				HeaderError(HE_UNDEF_ADDKEY,false,L("Undefined addkey %S in lemma scheme %d"),ls->a[i].getTerminatedBuffer(),i+1);
			}
			if (i==0) ak->isLemma=true;
		
		}
		t++;
	}
	}

	ls0=GetHeaderSection(HD_REGIONS);
	if (ls0==NULL) return;
	l=ls0->getElementsByTagName(s_xaira_item);
	n=l->getLength();
	t=0;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_region)==0) t++;
	}
	_region* rs = m_reg=mynew _region [t];
	m_nreg=t; t=0;
	for (int i=0;i<n;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		const xairo_uchar* ty=e->getAttribute(s_type);
		if (wcscmp(ty,s_region)!=0) continue;
		_region* rs=&m_reg[t];
		rs->name=e->getAttribute(s_ident);
		DOMElement *g=GetChildByTag(e,s_desc);
		if (g) rs->desc=GetContent(g);
		DOMElement* nl=GetChildByTag(e,s_nameList);
		DOMNodeList* ll=nl->getElementsByTagName(s_ident);
		if (ll->getLength() != 1) {
			HeaderError(HE_REGION_ADDKEY,true,L("Missing or duplicate addkey for region %s"),rs->name.getTerminatedBuffer());
			if (ll->getLength()==0) {
				t++;
				continue;
			}
		}
		DOMElement* x=(DOMElement*) ll->item(0);
		UnicodeString ak(GetContent(x));
		int nak=-1;
		for (int i=0;i<m_naddkeys;i++) {
			if (m_addkeys[i].name==ak) nak=i;
		}
		if (nak <0) {
			HeaderError(HE_REGION_ADDKEY,false,L("Bad addkey (%s) for region %s"),ak.getTerminatedBuffer(),rs->name.getTerminatedBuffer());
			t++; 
			continue;
		}
		m_addkeys[nak].isLemma=true;
		rs->nak=nak;
		ll=nl->getElementsByTagName(s_gi);
		rs->n=ll->getLength();
		rs->a=new UnicodeString [rs->n];
		for (int i=0;i<rs->n;i++) {
			DOMElement* k=(DOMElement*) ll->item(i);
			rs->a[i]=GetContent(k);  
		}
		t++;
	}



/*
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		if (TestItem(m,s_ls)||TestItem(m,s_lemmaScheme)) {
			DOMNodeList* lit=m->getChildNodes();
			int kk=lit->getLength();
			for (int j=0;j<kk;j++) {
				DOMNode* mm=lit->item(j);
				if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
				DOMElement* e=(DOMElement*) mm;
				const xairo_uchar* ty=e->getTagName();
				if (wcscmp(ty,s_gloss)==0) {
					ls[ils].desc=(const UChar*) GetContent(e);
				}
				else if (wcscmp(ty,s_label)==0||wcscmp(ty,s_name)==0) {
					ls[ils].name=(const UChar*) GetContent(e);
				}
				else if (wcscmp(ty,s_list)==0||wcscmp(ty,s_nameList)==0) {
					DOMNodeList* gl=e->getChildNodes();
					int kg=gl->getLength();
					int kkg=0;
					for (int jj=0;jj<kg;jj++) {
						DOMNode* eg=gl->item(jj);
						if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
						DOMElement* fg = (DOMElement*) eg;
						bool b1=wcscmp(fg->getTagName(),s_addkey)==0;
						bool b2=wcscmp(fg->getTagName(),s_ident)==0;
						if (b2) b2 = (fg->getAttribute(s_type) != NULL);
						if (b2) b2 = wcscmp(fg->getAttribute(s_type),s_addkey)==0;
						if (!b1 && !b2) continue;
						kkg++;
					}
					ls[ils].n=kkg;
					ls[ils].a=mynew UnicodeString [kkg];
					kkg=0;
					for (int jj=0;jj<kg;jj++) {
						DOMNode* eg=gl->item(jj);
						if (eg->getNodeType() != DOMNode::ELEMENT_NODE) continue;
						DOMElement* fg = (DOMElement*) eg;
						bool b1=wcscmp(fg->getTagName(),s_addkey)==0;
						bool b2=wcscmp(fg->getTagName(),s_ident)==0;
						if (b2) b2 = (fg->getAttribute(s_type) != NULL);
						if (b2) b2 = wcscmp(fg->getAttribute(s_type),s_addkey)==0;
						if (!b1 && !b2) continue;
						ls[ils].a[kkg++]=(const UChar*) GetContent(eg);
					}
				}
			}
			ils++;
			continue;
		}
*/
}

void CSaraHeaderReader::GetAddKeysOld()
{
	DOMNode* n1=GetHeaderSection(HD_ADDKEYS);
	int k=0; 
	DOMNodeList* l=NULL;
	if (n1) {
		l=n1->getChildNodes();
		k=l->getLength();
	}
	int t=0; int tt=0; int tl=0;
	bool bAK0=false;
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		if (TestItem(m,s_addkey)||TestItem(m,s_lemma)) t++;
		if (TestItem(m,s_lemma0)) {t++; bAK0=true;}
		if (TestItem(m,s_form)) tt++;
		if (TestItem(m,s_ls)||TestItem(m,s_lemmaScheme)) tl++;
	}
	if (!bAK0) t++;
	_addkey* a=m_addkeys=mynew _addkey [t];
	m_naddkeys=t;
	_ls* ls = m_ls=mynew _ls [tl];
	m_nls=tl;
	int ii=0; int ils=0;
	if (!bAK0) {
		DefAddKey(&a[0]);
		ii=1;
	}
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		if (TestItem(m,s_ls)||TestItem(m,s_lemmaScheme)) {
			DOMNodeList* lit=m->getChildNodes();
			int kk=lit->getLength();
			for (int j=0;j<kk;j++) {
				DOMNode* mm=lit->item(j);
				if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
				DOMElement* e=(DOMElement*) mm;
				const xairo_uchar* ty=e->getTagName();
				if (wcscmp(ty,s_desc)==0) {
					ls[ils].desc=(const UChar*) GetContent(e);
				}
				else if (wcscmp(ty,s_label)==0||wcscmp(ty,s_name)==0) {
					ls[ils].name=(const UChar*) GetContent(e);
				}
				else if (wcscmp(ty,s_list)==0||wcscmp(ty,s_nameList)==0) {
					DOMNodeList* gl=e->getChildNodes();
					int kg=gl->getLength();
					int kkg=0;
					for (int jj=0;jj<kg;jj++) {
						DOMNode* eg=gl->item(jj);
						if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
						DOMElement* fg = (DOMElement*) eg;
						bool b1=wcscmp(fg->getTagName(),s_addkey)==0;
						bool b2=wcscmp(fg->getTagName(),s_ident)==0;
						if (b2) b2 = (fg->getAttribute(s_type) != NULL);
						if (b2) b2 = wcscmp(fg->getAttribute(s_type),s_addkey)==0;
						if (!b1 && !b2) continue;
						kkg++;
					}
					ls[ils].n=kkg;
					ls[ils].a=mynew UnicodeString [kkg];
					kkg=0;
					for (int jj=0;jj<kg;jj++) {
						DOMNode* eg=gl->item(jj);
						if (eg->getNodeType() != DOMNode::ELEMENT_NODE) continue;
						DOMElement* fg = (DOMElement*) eg;
						bool b1=wcscmp(fg->getTagName(),s_addkey)==0;
						bool b2=wcscmp(fg->getTagName(),s_ident)==0;
						if (b2) b2 = (fg->getAttribute(s_type) != NULL);
						if (b2) b2 = wcscmp(fg->getAttribute(s_type),s_addkey)==0;
						if (!b1 && !b2) continue;
						ls[ils].a[kkg++]=(const UChar*) GetContent(eg);
					}
				}
			}
			ils++;
			continue;
		}
		if (!TestItem(m,s_addkey)&&!TestItem(m,s_lemma0)&&!TestItem(m,s_lemma)) continue;
		a[ii].isLemma=!TestItem(m,s_addkey);
		a[ii].bCaseSensitive=!a[ii].isLemma;
		a[ii].dproc=_addkey::DP_USEFORM;
		DOMNodeList* lit=m->getChildNodes();
		int kk=lit->getLength();
		for (int j=0;j<kk;j++) {
			DOMNode* mm=lit->item(j);
			if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			DOMElement* e=(DOMElement*) mm;
			const xairo_uchar* ty=e->getTagName();
			if (wcscmp(ty,s_ident)==0) {
				if (wcscmp(e->getAttribute(s_type),s_attc)==0) {
					a[ii].att=(const UChar*) GetContent(e);
					a[ii].proc=_addkey::AKP_TAG;
					a[ii].isPseudo=false;
					a[ii].bCaseSensitive=false;
				}
				else if (wcscmp(e->getAttribute(s_type),s_att)==0) {
					a[ii].att=(const UChar*) GetContent(e);
					a[ii].proc=_addkey::AKP_TAG;
					a[ii].isPseudo=false;
					a[ii].bCaseSensitive=true;
				} else if (wcscmp(e->getAttribute(s_type),s_pseudo)==0) {
					a[ii].isPseudo=true;
				}
			}
			else if (wcscmp(ty,s_desc)==0) {
				a[ii].desc=(const UChar*) GetContent(e);
			}
			else if (wcscmp(ty,s_label)==0||wcscmp(ty,s_name)==0) {
				a[ii].name=(const UChar*) GetContent(e);
			}
			else if (wcscmp(ty,s_note)==0||wcscmp(ty,s_default)==0) {
				a[ii].dproc=_addkey::DP_USEVAL;
				a[ii].defval=(const UChar*) _wcsdup(GetContent(e));
			}
			else if (wcscmp(ty,s_list)==0||wcscmp(ty,s_nameList)==0) {
				DOMNodeList* gl=e->getChildNodes();
				int kg=gl->getLength();
				int kkg=0;
				for (int jj=0;jj<kg;jj++) {
					DOMNode* eg=gl->item(jj);
					if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
					DOMElement* fg = (DOMElement*) eg;
					if (wcscmp(fg->getTagName(),s_gi) != 0) continue;
					kkg++;
				}
				a[ii].ngi=kkg;
				a[ii].gi=mynew UnicodeString [kkg];
				kkg=0;
				for (int jj=0;jj<kg;jj++) {
					DOMNode* eg=gl->item(jj);
					if (eg->getNodeType() != DOMNode::ELEMENT_NODE) continue;
					DOMElement* fg = (DOMElement*) eg;
					if (wcscmp(fg->getTagName(),s_gi) != 0) continue;
					a[ii].gi[kkg++]=(const UChar*) GetContent(eg);
				
				}
			}
		}
/*		if (a[ii].isLemma) {
			ls[ils].name=a[ii].name;
			ls[ils].desc=a[ii].desc;
			ls[ils].n=1;
			ls[ils].a=new UnicodeString [1];
			ls[ils].a[0]=a[ii].name;
			ils++;
		}
*/		ii++;
	}
	m_nwtags=tt;
	m_wtags=mynew  _wtag [tt];
	ii=0;
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		if (!TestItem(m,s_form)) continue;
		DOMNodeList* lit=m->getChildNodes();
		int kk=lit->getLength();
		for (int j=0;j<kk;j++) {
			DOMNode* mm=lit->item(j);
			if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			DOMElement* e=(DOMElement*) mm;
			const xairo_uchar* ty=e->getTagName();
			if (wcscmp(ty,s_ident)==0) {
				if (wcscmp(e->getAttribute(s_type),s_gi)==0) {
					m_wtags[ii].s=(const UChar*) _wcsdup(GetContent(e));
				}
			}
		}
		ii++;
	}
}

DOMElement* CSaraHeaderReader::GetUniqueNode(const xairo_uchar* w)
{
	DOMNodeList* l=m_doc->getElementsByTagName(w);
	int k=l->getLength();
	if (k != 1) return NULL;
	return (DOMElement*) l->item(0);
}

DOMElement* CSaraHeaderReader::GetNamedList(DOMElement* n ,const xairo_uchar* w)
{
	DOMNodeList* l=n->getElementsByTagName(s_list);
	int k=l->getLength();
	for (int i=0;i<k;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		if (wcscmp(e->getAttribute(s_type),w)==0) return e;
	}
	return NULL;
}

DOMElement* CSaraHeaderReader::GetNamedList2(DOMElement* n ,const xairo_uchar* w)
{
	DOMNodeList* l=n->getElementsByTagName(s_xaira_list);
	int k=l->getLength();
	for (int i=0;i<k;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		if (wcscmp(e->getAttribute(s_type),w)==0) return e;
	}
	return NULL;
}

DOMElement* CSaraHeaderReader::GetNamedNote(DOMElement* n ,const xairo_uchar* w)
{
	DOMNodeList* l=n->getElementsByTagName(s_note);
	int k=l->getLength();
	for (int i=0;i<k;i++) {
		DOMElement* e=(DOMElement*) l->item(i);
		if (wcscmp(e->getAttribute(s_type),w)==0) return e;
	}
	return NULL;
}

void CSaraHeaderReader::CopyOldAKStruc()
{
}

bool CSaraHeaderReader::TestItem(DOMNode* n,const xairo_uchar* w)
{
	if (n->getNodeType() != DOMNode::ELEMENT_NODE) return false;
	if (wcscmp(n->getNodeName(),s_item)!=0 && wcscmp(n->getNodeName(),s_xaira_item)!=0) return false;
	DOMElement* e=(DOMElement*) n;
	return wcscmp(e->getAttribute(s_n),w)==0;
}

const xairo_uchar* CSaraHeaderReader::GetContent(DOMNode* n)
{
	DOMNodeList* l=n->getChildNodes();
	int k=l->getLength();
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		int t=m->getNodeType();
		if (t!=DOMNode::TEXT_NODE && t!=DOMNode::CDATA_SECTION_NODE) continue;
		return m->getNodeValue();
	}
	return s_;
}

UnicodeString NS(UnicodeString s)
{
	int k=s.length();
	UnicodeString t; bool bSp=true; int j=0;
	for (int i=0;i<k;i++) {
		if (s.charAt(i) <= 32) {
			if (!bSp) {
				t+=' ';
				bSp=true;
			} else ;
		} else {
			t+=s.charAt(i);
			bSp=false;
		}
	}
	if (bSp && t.length()) t.truncate(t.length()-1);
	return t;
}

UnicodeString CSaraHeaderReader::GetContentNS(DOMNode* n)
{
	DOMNodeList* l=n->getChildNodes();
	int k=l->getLength();
	for (int i=0;i<k;i++) {
		DOMNode* m=l->item(i);
		int t=m->getNodeType();
		if (t!=DOMNode::TEXT_NODE && t!=DOMNode::CDATA_SECTION_NODE) continue;
		return NS(m->getNodeValue());
	}
	return s_;
}

void CSaraHeaderReader::GetValSource(DOMElement* e,_valsource* v)
{
	v->ident=e->getAttribute(s_ident);
	v->ns=e->getAttribute(s_ns);
	v->type=e->getAttribute(s_type);
	const xairo_uchar* cf=e->getAttribute(s_caseFold);
	if (cf && wcscmp(cf,s_true)==0) v->bcasefold=true;
	DOMElement* n=GetChildByTag(e,s_nameList);
	if (n) {
		DOMNodeList* nl=n->getElementsByTagName(s_gi);
		v->ngi=nl->getLength();
		v->gi=new const xairo_uchar* [v->ngi];
		for (int i=0;i<v->ngi;i++) {
			DOMElement* gi=(DOMElement*) nl->item(i);
			v->gi[i]=GetContent(gi);
		}
	}
	if (wcscmp(v->type,s_attribute)==0) {
		_attusage* a=NULL;
		a=GetAttUsage(s_,v->ident);
		if (a==NULL) {
		for (int i=0;i<v->ngi;i++) {
			_attusage* a=GetAttUsage(v->gi[i],v->ident);
			if (a==NULL)
				HeaderError(HE_UNDEF_ATTRIBUTE,false,L("Value source uses undeclared attribute (%s/@%s)"),v->gi[i],v->ident);
		}
		}
	}
	DOMElement* d=GetChildByTag(e,s_defaultVal);
	if (d) {
		v->def=GetContent(d);
		const xairo_uchar* uf=d->getAttribute(s_source);
		if (uf && wcscmp(uf,s_form)==0) v->bdefform=true;
	}
}

DOMElement* CSaraHeaderReader::MakeValSource(_valsource* vs)
{
	DOMElement* e=m_doc->createElement(s_valSource);
	e->setAttribute(s_type,vs->type);
	e->setAttribute(s_ident,vs->ident);
	if (vs->ns[0]) e->setAttribute(s_ns,vs->ns);
	if (vs->bcasefold) e->setAttribute(s_caseFold,s_true);
	if (vs->ngi) {
	DOMElement* it9=m_doc->createElement(s_nameList); // was list type=gilist
		for (int jj=0;jj<vs->ngi;jj++) {
			DOMElement* it10=m_doc->createElement(s_gi);
			DOMText* it11=m_doc->createTextNode(vs->gi[jj]);
			it10->appendChild(it11);
			it9->appendChild(it10);
		}
		e->appendChild(it9);
	}
	if (vs->bdefform || vs->def[0]) {
		DOMElement* it7=m_doc->createElement(s_defaultVal); // was note
		if (vs->bdefform) it7->setAttribute(s_source,s_form);
		e->appendChild(it7);
		DOMText* it8=m_doc->createTextNode(vs->def);
		it7->appendChild(it8);
	}
	return e;
}

void CSaraHeaderReader::SetAddKeys()
{
	DOMElement* e=SetHeaderSection(HD_ADDKEYS);
	for (int i=0;i<m_nwtags;i++) {
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_type,s_form);
		_valsource vs;
		vs.type=s_element;
		vs.ident=m_wtags[i].s.getTerminatedBuffer();
		vs.ns=s_;
		DOMElement* it1=MakeValSource(&vs); 
		it->appendChild(it1);
		e->appendChild(it);
	}
	for (int i=0;i<m_naddkeys;i++) {
		_addkey* a=&m_addkeys[i];
		if (i==0 && a->proc==_addkey::AKP_FORM) continue;
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_type,i==0?s_form:s_addKey);
		it->setAttribute(s_ident,(const xairo_uchar*) a->name.getTerminatedBuffer());
		if (a->desc[0]) {
			DOMElement* it3=m_doc->createElement(s_desc);
			it->appendChild(it3);
			DOMText* it4=m_doc->createTextNode((const xairo_uchar*) a->desc.getTerminatedBuffer());
			it3->appendChild(it4);
		}
		_valsource vs;
		if (a->isPseudo) 
		{
			vs.type=s_pseudo;
			vs.ident=s_name_par;
			vs.ns=s_;
		} else {
			vs.type=s_attribute;
			vs.ident=(const xairo_uchar*) a->att.getTerminatedBuffer();
			vs.ns=(const xairo_uchar*) a->ans.getTerminatedBuffer();
		}
		vs.bcasefold=!a->bCaseSensitive;
		vs.ngi=a->ngi;
		vs.gi=new const xairo_uchar* [a->ngi];
		for (int jj=0;jj<a->ngi;jj++) {
			vs.gi[jj]=(const xairo_uchar*) a->gi[jj].getTerminatedBuffer();
		}
		vs.def=(const xairo_uchar*) a->defval.getTerminatedBuffer();
		vs.bdefform=(a->dproc==_addkey::DP_USEFORM);
		DOMElement* it1=MakeValSource(&vs);
		e->appendChild(it);
		it->appendChild(it1);
	}
	DOMElement* e1=SetHeaderSection(HD_LEMMASCHEMES);
	for (int i=0;i<m_nls;i++) {
		_ls* ls=&m_ls[i];
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_type,s_lemmaScheme);
		it->setAttribute(s_ident,(const xairo_uchar*) ls->name.getTerminatedBuffer());
		if (!ls->desc.isEmpty()) {
			DOMElement* it3=m_doc->createElement(s_desc);
			it->appendChild(it3);
			DOMText* it4=m_doc->createTextNode((const xairo_uchar*) ls->desc.getTerminatedBuffer());
			it3->appendChild(it4);
		}
		DOMElement* it9=m_doc->createElement(s_nameList); // was nameList
		for (int jj=0;jj<ls->n;jj++) {
			DOMElement* it10=m_doc->createElement(s_ident);  // was addkey !
			DOMText* it11=m_doc->createTextNode((const xairo_uchar*) ls->a[jj].getTerminatedBuffer());
			it10->appendChild(it11);
			it9->appendChild(it10);
		}
		it->appendChild(it9);
		e1->appendChild(it);
	}

	e1=SetHeaderSection(HD_REGIONS);
	for (int i=0;i<m_nreg;i++) {
		_region* rs=&m_reg[i];
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_type,s_region);
		it->setAttribute(s_ident,(const xairo_uchar*) rs->name.getTerminatedBuffer());
		if (!rs->desc.isEmpty()) {
			DOMElement* it3=m_doc->createElement(s_desc);
			it->appendChild(it3);
			DOMText* it4=m_doc->createTextNode((const xairo_uchar*) rs->desc.getTerminatedBuffer());
			it3->appendChild(it4);
		}
		DOMElement* it9=m_doc->createElement(s_nameList); // was nameList
		DOMElement* it10=m_doc->createElement(s_ident);  // was addkey !
		UnicodeString ak(m_addkeys[rs->nak].name);
		DOMText* it11=m_doc->createTextNode((const xairo_uchar*) ak.getTerminatedBuffer());
		it10->appendChild(it11);
		it9->appendChild(it10);
		for (int jj=0;jj<rs->n;jj++) {
			DOMElement* it10=m_doc->createElement(s_gi);  // was addkey !
			DOMText* it11=m_doc->createTextNode((const xairo_uchar*) rs->a[jj].getTerminatedBuffer());
			it10->appendChild(it11);
			it9->appendChild(it10);
		}
		it->appendChild(it9);
		e1->appendChild(it);
	}
	return;
}

void CSaraHeaderReader::SetAddKeysOld()
{
	DOMElement* e=SetHeaderSection(HD_ADDKEYS);
	for (int i=0;i<m_nwtags;i++) {
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_n,s_form);
		DOMElement* it1=m_doc->createElement(s_ident);
		it1->setAttribute(s_type,s_gi);
		it->appendChild(it1);
		DOMText* it2=m_doc->createTextNode((const xairo_uchar*) m_wtags[i].s.getTerminatedBuffer());
		it1->appendChild(it2);
		e->appendChild(it);
	}
	for (int i=0;i<m_naddkeys;i++) {
		_addkey* a=&m_addkeys[i];
		if (i==0 && a->proc==_addkey::AKP_FORM) continue;
		DOMElement* it=m_doc->createElement(s_xaira_item);
		if (a->isLemma) 
			it->setAttribute(s_n,i==0?s_lemma0:s_lemma);
		else
			it->setAttribute(s_n,s_addkey);
		DOMElement* it1=m_doc->createElement(s_ident);
		const xairo_uchar* sa=s_att;
		if (a->isPseudo) sa=s_pseudo;
		if (!a->bCaseSensitive) sa=s_attc;
		it1->setAttribute(s_type,sa);
		it->appendChild(it1);
		DOMText* it2=m_doc->createTextNode((const xairo_uchar*) a->att.getTerminatedBuffer());
		it1->appendChild(it2);
		if (a->desc[0]) {
			DOMElement* it3=m_doc->createElement(s_desc);
			it->appendChild(it3);
			DOMText* it4=m_doc->createTextNode((const xairo_uchar*) a->desc.getTerminatedBuffer());
			it3->appendChild(it4);
		}
		DOMElement* it5=m_doc->createElement(s_name); // was label
		it->appendChild(it5);
		DOMText* it6=m_doc->createTextNode((const xairo_uchar*) a->name.getTerminatedBuffer());
		it5->appendChild(it6);
		if (a->dproc==_addkey::DP_USEVAL) {
			DOMElement* it7=m_doc->createElement(s_default); // was note
			it->appendChild(it7);
			DOMText* it8=m_doc->createTextNode((const xairo_uchar*) a->defval.getTerminatedBuffer());
			it7->appendChild(it8);
		}
		if (a->ngi) {
			DOMElement* it9=m_doc->createElement(s_nameList); // was list type=gilist
//			it9->setAttribute(s_type,s_gilist);
			for (int jj=0;jj<a->ngi;jj++) {
				DOMElement* it10=m_doc->createElement(s_gi);
				DOMText* it11=m_doc->createTextNode((const xairo_uchar*) a->gi[jj].getTerminatedBuffer());
				it10->appendChild(it11);
				it9->appendChild(it10);
			}
			it->appendChild(it9);
		}
		e->appendChild(it);
	}
	for (int i=0;i<m_nls;i++) {
		_ls* ls=&m_ls[i];
		DOMElement* it=m_doc->createElement(s_xaira_item);
		it->setAttribute(s_n,s_lemmaScheme);
		if (!ls->desc.isEmpty()) {
			DOMElement* it3=m_doc->createElement(s_desc);
			it->appendChild(it3);
			DOMText* it4=m_doc->createTextNode((const xairo_uchar*) ls->desc.getTerminatedBuffer());
			it3->appendChild(it4);
		}
		DOMElement* it5=m_doc->createElement(s_name); // was label
		it->appendChild(it5);
		DOMText* it6=m_doc->createTextNode((const xairo_uchar*) ls->name.getTerminatedBuffer());
		it5->appendChild(it6);
		DOMElement* it9=m_doc->createElement(s_nameList); // was nameList
//		it9->setAttribute(s_type,s_aklist);
		for (int jj=0;jj<ls->n;jj++) {
			DOMElement* it10=m_doc->createElement(s_ident);  // was addkey !
			it10->setAttribute(s_type,s_addkey);
			DOMText* it11=m_doc->createTextNode((const xairo_uchar*) ls->a[jj].getTerminatedBuffer());
			it10->appendChild(it11);
			it9->appendChild(it10);
		}
		it->appendChild(it9);
		e->appendChild(it);
	}
	return;
}

static UChar s_langUsage[]={'l','a','n','g','U','s','a','g','e',0};
static UChar s_language[]={'l','a','n','g','u','a','g','e',0};
static UChar s_xara_lang[]={'x','a','r','a',':','l','a','n','g',0};
static UChar s_xara_lang_default[]={'x','a','r','a',':','l','a','n','g','-','d','e','f','a','u','l','t',0};
static UChar s_hash_def[]={'#','d','e','f',0};
static UChar s_xara_wordbreak_rules[]={'x','a','r','a',':','w','o','r','d','b','r','e','a','k','-','r','u','l','e','s',0};
static UChar s_code[]={'c','o','d','e',0};

void CSaraHeaderReader::GetLangs()
{
	int k=0;
	DOMElement* n=GetUniqueNode(s_langUsage);
	DOMNodeList* nl=NULL;
	if (n) {
		nl=n->getChildNodes();
		k=nl->getLength();
	}
	int t=0; int i;
	for (i=0;i<k;i++) {
		DOMNode* m=nl->item(i);
		if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) m;
		if (wcscmp(e->getTagName(),s_language) != 0) continue;
		t++;
	}
	m_langs=mynew _lang [t];
	m_nlangs=t;
	t=0;
	for (int i=0;i<k;i++) {
		DOMNode* m=nl->item(i);
		if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) m;
		if (wcscmp(e->getTagName(),s_language) != 0) continue;
		wcscpy((xairo_uchar*) m_langs[t].id,e->getAttribute(s_id));
		m_langs[t].desc=(const UChar*) GetContent(e);
		DOMNodeList* nl=e->getChildNodes();
		int kk=nl->getLength();
		for (int j=0;j<kk;j++) {
			DOMNode* m=nl->item(j);
			if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			DOMElement* em=(DOMElement*) m;
			if (wcscmp(em->getTagName(),s_ident) == 0 &&
				wcscmp(em->getAttribute(s_type),s_xara_lang) == 0) {
					wcstombs(m_langs[t].locale,GetContent(em),MAXLOCALE);
				} else if (wcscmp(em->getTagName(),s_ident) == 0 &&
					wcscmp(em->getAttribute(s_type),s_xara_lang_default) == 0)
				{
					wcstombs(m_langs[t].locale,GetContent(em),MAXLOCALE);
					m_deflang=&m_langs[t];
				}
				else if (wcscmp(em->getTagName(),s_code) == 0 &&
					wcscmp(em->getAttribute(s_type),s_xara_wordbreak_rules) == 0)
				{
					m_langs[t].bUserTokenRules=true;
					m_langs[t].rules=(const UChar*) GetContent(em);
				}
		}
		if (strcmp(m_langs[t].locale,"#def")==0) m_langs[t].locale[0]=0;
		t++;
	}
}

void CSaraHeaderReader::Clean(DOMNode* n)
{
	DOMNodeList* l=n->getChildNodes();
	DOMNode* m=n->getFirstChild();
	DOMNode* mm;
	while (m) {
		mm=m->getNextSibling();
		n->removeChild(m);
		m=mm;
	}
}

static UChar s_ill[]={'I','n','d','e','x','t','o','o','l','s',' ','g','e','n','e','r','a','t','e','d',' ',
'l','a','n','g','u','a','g','e',' ','l','i','s','t',0};

void CSaraHeaderReader::SetLangs()
{
	DOMElement* n=GetUniqueNode(s_langUsage);
// if it doesn't exist create it, otherwise clean it
// all the set routines must do this
	if (n) {
		Clean(n);
	} else {
		DOMElement* n0=GetEncodingDesc();
		n=m_doc->createElement(s_langUsage);
		n0->appendChild(n);
	}
	if (m_nlangs==0) {
		DOMNode* p=n->getParentNode();
		p->removeChild(n);
		return;
	}
	DOMComment* c=m_doc->createComment(s_ill);
	n->appendChild(c);
	for (int i=0;i<m_nlangs;i++) {
		_lang* l=&m_langs[i];
		DOMElement* e=m_doc->createElement(s_language);
		e->setAttribute(s_id,(const xairo_uchar*) l->id);
		DOMText* t=m_doc->createTextNode((const xairo_uchar*) l->desc.getTerminatedBuffer());
		e->appendChild(t);
		n->appendChild(e);
		DOMElement* ee=m_doc->createElement(s_ident);
		ee->setAttribute(s_type,(l==m_deflang)?s_xara_lang_default:s_xara_lang);
		e->appendChild(ee);
		xairo_uchar lc[MAXLOCALE];
		mbstowcs(lc,l->locale,MAXLOCALE);
		if (lc[0]==0) wcscpy(lc,s_hash_def);
		DOMText* et=m_doc->createTextNode(lc);
		ee->appendChild(et);
		if (l->bUserTokenRules) {
			DOMElement *wb=m_doc->createElement(s_code);
			wb->setAttribute(s_type,s_xara_wordbreak_rules);
			DOMCDATASection* nnn=m_doc->createCDATASection((const xairo_uchar*) l->rules.getTerminatedBuffer());
			wb->appendChild(nnn);
			e->appendChild(wb);
		}
	}
}

void CSaraHeaderReader::StructureError(const char*)
{
}


void CSaraHeaderReader::GetCodeBooks()
{
//	Must get AttUsages first!
	DOMElement* n=GetHeaderSection(HD_CODEBOOKS);
	if (n==NULL) return;
	bool bOld=(wcscmp(n->getNodeName(),s_encodingDesc)==0);
	DOMNodeList* nl=NULL;
	if (bOld) nl=n->getElementsByTagName(s_list);
	else nl=n->getElementsByTagName(s_xaira_item);
	int k=nl->getLength();
	for (int i=0;i<k;i++) {
		DOMNode* nn=nl->item(i);
		if (nn->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) nn;
		if (bOld && wcscmp(e->getTagName(),s_list) != 0) continue;
		if (bOld && wcscmp(e->getAttribute(s_type),s_xara_codebook) != 0) continue;
		const UChar* att=s_;
		const UChar* gi=s_;
		if (bOld) {
			att=(const UChar*) e->getAttribute(s_n);
			gi=(const UChar*) e->getAttribute(s_gi);
			if (gi==NULL) gi=(const UChar*) s_;
		} else {
			DOMNodeList* f=e->getElementsByTagName(s_ident);
			int fn=f->getLength();
			if (fn > 0) {
				DOMNode* f1=f->item(0);
				const xairo_uchar* x=GetContent(f1);
				const UChar* y=wcschr(x,'@');
				UChar* g=mynew UChar [y-x+1];
				wcsncpy((UChar*) g,x,y-x);
				g[y-x]=0;
				gi=g;
				att=y+1;
			}
			f=e->getElementsByTagName(s_valList);
			fn=f->getLength();
			if (fn > 0) e=(DOMElement*) f->item(0);
				DOMNode* f1=f->item(0);
		}
		_attusage* at=GetAttUsage(gi,att);
		if (at==NULL) continue;
		at->bCB=true;
		DOMNodeList* nnl=e->getChildNodes();
		int t=0;
		for (int j=0;j<(int)nnl->getLength();j++) {
			DOMNode* m=nnl->item(j);
			if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			if (wcscmp(m->getNodeName(),s_item) != 0 && wcscmp(m->getNodeName(),s_valItem) != 0 ) continue;
			t++;
		}
		at->cb=mynew _code [t];
		at->nc=t;
		t=0;
		for (int j=0;j<(int) nnl->getLength();j++) {
			DOMNode* m=nnl->item(j);
			if (m->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			if (wcscmp(m->getNodeName(),s_item) != 0 && wcscmp(m->getNodeName(),s_valItem) != 0) continue;
			DOMElement* me=(DOMElement*) m;
			if (bOld) 
				at->cb[t].code=(const UChar*) me->getAttribute(s_id);
			else 
				at->cb[t].code=(const UChar*) me->getAttributeNS(s_xmlns,s_id);
			at->cb[t].desc=(const UChar*) GetContent(me);
			t++;
		}
	}
}

static UChar s_icb[]={'I','n','d','e','x','t','o','o','l','s',' ','c','o','d','e','b','o','o','k',0};

void CSaraHeaderReader::GetCodeBook(_attusage* at,DOMNode* e)
{
	DOMNodeList* l=((DOMElement*) e)->getElementsByTagName(s_valList);
	if (l->getLength()==0) return;
	at->bCB=true;
	DOMElement* cb=(DOMElement*) l->item(0);
	DOMNodeList* lv=cb->getElementsByTagName(s_valItem);
	int n=lv->getLength();
	_code* cl=new _code [n];
	for (int i=0;i<n;i++) {
		DOMElement* m=(DOMElement*) lv->item(i);
		cl[i].code=m->getAttribute(s_ident);
		for (int j=0;j<i;j++) {
			if (cl[j].code==cl[i].code)
				HeaderError(HE_DUP_CODE,true,L("Duplicate code %s in code book for %s"),cl[i].code.getTerminatedBuffer(),at->name.getTerminatedBuffer());
		}
		DOMNodeList* lv1=m->getElementsByTagName(s_desc);
		int n1=lv1->getLength();
		if (n1 > 0) {
			DOMElement* m0=(DOMElement*) lv1->item(0);
			cl[i].desc=GetContentNS(m0);
		}
		else 
			cl[i].desc=s_;
	}
	at->cb=cl;
	at->nc=n;
}

void CSaraHeaderReader::SetCodeBook(_attusage* at,DOMNode* e)
{
	DOMComment* c=m_doc->createComment(s_icb); 
	e->appendChild(c);
	for (int j=0;j<at->nc;j++) {
		DOMElement* ee=m_doc->createElement(s_valItem);
		ee->setAttribute(s_ident,(const xairo_uchar*) at->cb[j].code.getTerminatedBuffer());
		DOMElement* ed=m_doc->createElement(s_desc);
		ee->appendChild(ed);
		DOMText* t=m_doc->createTextNode((const xairo_uchar*) at->cb[j].desc.getTerminatedBuffer());
		ed->appendChild(t);
		e->appendChild(ee);
	}
}

void CSaraHeaderReader::SetCodeBooks()
{
//	return;
	DOMElement* parent=SetHeaderSection(HD_CODEBOOKS);
	return;
/*	DOMNodeList* nl=n->getElementsByTagName(s_list);
	int k=nl->getLength();
	DOMNode** del=mynew DOMNode* [k];
	for (int i=0;i<k;i++)
		del[i]=NULL;
	for (int i=0;i<m_nglobatts;i++) {
		_attusage* a=&m_globals[i];
		a->bT=false;
	}
	for (int i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		for (int j=0;j<t->natts;j++) {
			_attusage* a=&t->atts[j];
			a->bT=false;
		}
	}
	DOMNode* last=NULL;
// first update existing codebooks
	for (int i=0;i<k;i++) {
		DOMNode* nn=nl->item(i);
		if (nn->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) nn;
		if (wcscmp(e->getTagName(),s_list) != 0) continue;
		if (wcscmp(e->getAttribute(s_type),s_xara_codebook) != 0) continue;
		const UChar* att=(const UChar*) e->getAttribute(s_n);
		const UChar* gi=(const UChar*) e->getAttribute(s_gi);
		if (gi==NULL) gi=(const UChar*) s_;
		_attusage *at=GetAttUsage(gi,att);
		if (at==NULL||!at->bCB) {
			del[i]=e;
			continue;
		}
		Clean(e);
		SetCodeBook(at,e);
		at->bT=true;
		last=e;
	}
	for (int i=0;i<k;i++) {
		DOMNode* d=del[i];
		if (d==NULL) continue;
		DOMNode* d0=d->getParentNode();
		d0->removeChild(d);
	}
// Now add any new code books. If we found any code books the new ones should be made
// siblings.
// The <code>list</code> element containing this specification may be located 
// within a <gi>p</gi> element directly  inside the <gi>encodingDecl</gi> element,
// or more specifically within the <gi>interpretation</gi> element, within 
// the <gi>editorialDecl</gi> element of the <gi>encodingDesc</gi>
	DOMNode* parent;
	if (last)
		parent=last->getParentNode();
	else {
		parent=m_doc->createElement(s_p);
		n->appendChild(parent);
	}
*/
	for (int ii=0;ii<m_nglobatts;ii++) {
		_attusage* at=&m_globals[ii];
		if (/* at->bT || */ !at->bCB) continue;
		DOMElement* e=m_doc->createElement(s_xaira_item);  // wsa list
//		e->setAttribute(s_type,s_xara_codebook);
//		e->setAttribute(s_n,(const xairo_uchar*) at->name.getTerminatedBuffer());
		UnicodeString x("@");
		x+=at->name;
		DOMElement* id=m_doc->createElement(s_ident);
		DOMText* idt=m_doc->createTextNode(x.getTerminatedBuffer());
		id->appendChild(idt);
		e->appendChild(id);
		DOMElement* ev=m_doc->createElement(s_valList);
		e->appendChild(ev);
		SetCodeBook(at,ev);
		parent->appendChild(e);
	}
	for (int ii=0;ii<m_ntags;ii++) {
		_tagusage* t=&m_tagusage[ii];
		for (int iii=0;iii<t->natts;iii++) {
			_attusage* at=&t->atts[iii];
			if (/* at->bT || */ !at->bCB) continue;
			DOMElement* e=m_doc->createElement(s_xaira_item);
//			e->setAttribute(s_type,s_xara_codebook);
//			e->setAttribute(s_gi,(const xairo_uchar*) t->name.getTerminatedBuffer());
//			e->setAttribute(s_n,(const xairo_uchar*) at->name.getTerminatedBuffer());
			UnicodeString x=t->name;
			x+="@";
			x+=at->name;
			DOMElement* id=m_doc->createElement(s_ident);
			DOMText* idt=m_doc->createTextNode(x.getTerminatedBuffer());
			id->appendChild(idt);
			e->appendChild(id);
			DOMElement* ev=m_doc->createElement(s_valList);
			e->appendChild(ev);
			SetCodeBook(at,ev);
			parent->appendChild(e);
		}
	}
}


void CSaraHeaderReader::GetIPSpecOld()
{
//	Must get AttUsages first!
	DOMElement* m=GetHeaderSection(HD_IPS);
	if (m==NULL) return;
	DOMNodeList* nl=m->getChildNodes();
	int kl=nl->getLength();
	for (int i=0; i<kl; i++) {
		DOMNode* mm=nl->item(i);
		if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) mm;
		if (wcscmp(e->getTagName(),s_item) != 0 && wcscmp(e->getTagName(),s_xaira_item) != 0) continue;
		DOMNodeList* nle=e->getChildNodes();
		const xairo_uchar *id=s_, *kw=s_, *tg=s_, *url=s_;
		for (int j=0;j<(int) nle->getLength();j++) {
			DOMNode* mm=nle->item(j);
			if (mm->getNodeType() != DOMNode::ELEMENT_NODE) continue;
			if (wcscmp(mm->getNodeName(),s_ident)==0)
				id=GetContent(mm);
			else if (wcscmp(mm->getNodeName(),s_ns)==0)
				url=GetContent(mm);
			else if (wcscmp(mm->getNodeName(),s_kw)==0)
				kw=GetContent(mm);
			else if (wcscmp(mm->getNodeName(),s_target)==0||wcscmp(mm->getNodeName(),s_gi)==0)
				tg=GetContent(mm);
		}
		ParseIP((const UChar*) id,(const UChar*) url,(const UChar*) kw,(const UChar*) tg);
	}
}

void CSaraHeaderReader::GetIPSpec()
{
	if (m_bConverting) {
		GetIPSpecOld();
		return;
	}
//	Must get AttUsages first!
	DOMElement* m=GetHeaderSection(HD_IPS);
	if (m==NULL) return;
	DOMNodeList* ll=m->getElementsByTagName(s_xaira_item);
	int k=ll->getLength();
	for (int i=0;i<k;i++) {
		DOMElement* ip=(DOMElement*) ll->item(i);
		const xairo_uchar* ty=ip->getAttribute(s_type);
		if (wcscmp(ty,s_indexPol)!=0) continue;
		DOMElement* pol=GetChildByTag(ip,s_elementPolicy);
		if (pol) {
			const xairo_uchar* ident=pol->getAttribute(s_ident);
			const xairo_uchar* type=pol->getAttribute(s_type);
			const xairo_uchar* ns=pol->getAttribute(s_ns);
			_tagusage* t=GetTagUsage(ident,ns);
			if (t) t->ip=DecodeEIP2(type);
			else 
				HeaderError(HE_UNDEF_ELEMENT,true,L("Index policy for undeclared element (%s)"),ident);

		}
		DOMElement* apol=GetChildByTag(ip,s_attributePolicy);
		if (apol) {
			const xairo_uchar* ident=apol->getAttribute(s_ident);
			const xairo_uchar* type=apol->getAttribute(s_type);
			const xairo_uchar* ns=apol->getAttribute(s_ns);
			indexpol ipx=IP_CDATA; int i;
			for (i=0;i<IP_MAX;i++) {
				if (ipnname[i]==type) break;
			}
			if (i<IP_MAX)ipx=(indexpol) i;
			else 
			{
				HeaderError(HE_UNDEF_POLICY,true,L("Unknown index policy (%s)"),type);
			}
			DOMElement* j=GetChildByTag(apol,s_joinTo);
			DOMElement* jj=NULL;
			if (j) jj=GetChildByTag(j,s_gi);
			const xairo_uchar* target=s_;
			if (jj) {
				target=GetContent(jj);
			}
			DOMElement* nl=GetChildByTag(apol,s_nameList);
			if (ns[0]) {
			_attusage* a=GetAttUsage(s_,ident,ns);
				if (a) {
					a->ip=ipx;
					a->target=target;
				} else {
					HeaderError(HE_UNDEF_ATTRIBUTE,true,L("Index policy for undeclared global attribute (%s:%s)"),ns,ident);
				}
			} else 
			if (nl) {
				DOMNodeList* gl=nl->getElementsByTagName(s_gi);
				for (int i=0;i<(int) gl->getLength();i++) {
					const xairo_uchar* elt=GetContent(gl->item(i));
					_attusage* a=GetAttUsage(elt,ident);
					if (a) {
						a->ip=ipx;
						a->target=target;
					} else {
						HeaderError(HE_UNDEF_ATTRIBUTE,true,L("Index policy for undeclared attribute (%s/@%s)"),elt,ident);
					}
				}
			}

		}
	}
}

void CSaraHeaderReader::ParseIP(const UChar* id,const UChar* url,const UChar* kw,const UChar* tg)
{
	const UChar* x=wcschr(id,'@');
	if (x==NULL) {
		_tagusage* t=GetTagUsage(id);
		UnicodeString ks(kw);
		t->wb=DecodeEWB(ks);
		t->ip=DecodeEIP(ks);
		return;
	}
	UChar* gi=mynew UChar [x-id+1];
	wcsncpy(gi,id,x-id);
	gi[x-id]=0;
	_attusage* a=GetAttUsage(gi,x+1,url);
	if (a==NULL) return; int i;
	for (i=0;i<IP_MAX;i++) {
		if (ipname[i]==kw) break;
	}
	if (i==IP_MAX) return;
	a->ip=(indexpol) i;
	a->target=tg;
}


void CSaraHeaderReader::SetIPSpecA(DOMElement* m,_attusage* a,const xairo_uchar* elt)
{
	if (a->ip==IP_CDATA) return;
	DOMElement* e=m_doc->createElement(s_xaira_item);
	e->setAttribute(s_type,s_indexPol);
	m->appendChild(e);
	DOMElement* e1=m_doc->createElement(s_attributePolicy);
	e1->setAttribute(s_ident,a->name.getTerminatedBuffer());
	e1->setAttribute(s_type,ipnname[a->ip].getTerminatedBuffer());
	if (!a->uri.isEmpty()) {
		e1->setAttribute(s_ns,a->uri.getTerminatedBuffer());
	}
	e->appendChild(e1);
	DOMElement* e4=m_doc->createElement(s_nameList);
	e1->appendChild(e4);
	DOMElement* e5=m_doc->createElement(s_gi);
	e4->appendChild(e5);
	DOMText* t=m_doc->createTextNode(elt);
	e5->appendChild(t);
	if (a->ip==IP_IDREF) {
		DOMElement* e2=m_doc->createElement(s_joinTo);
		e1->appendChild(e2);
		DOMElement* e2a=m_doc->createElement(s_gi);
		e2->appendChild(e2a);
		DOMText* t1=m_doc->createTextNode((const xairo_uchar*) a->target.getTerminatedBuffer());
		e2a->appendChild(t1);
	}
}

void CSaraHeaderReader::SetIPSpec()
{
	DOMElement* m=SetHeaderSection(HD_IPS);
	for (int i=0;i<m_nglobatts;i++) {
		_attusage* a=&m_globals[i];
		SetIPSpecA(m,a,s_asterisk);
	}
	for (int i=0;i<m_ntags;i++) {
		_tagusage *t=&m_tagusage[i];
		if (t->ip!= EIP_DEFAULT) {
			DOMElement* e=m_doc->createElement(s_xaira_item);
			e->setAttribute(s_type,s_indexPol);
			m->appendChild(e);
			DOMElement* e1=m_doc->createElement(s_elementPolicy);
			e1->setAttribute(s_ident,t->name.getTerminatedBuffer());
			e1->setAttribute(s_type,CodeEIP2(t->ip).getTerminatedBuffer());	
			e->appendChild(e1);
		}
		for (int j=0;j<t->natts;j++) {
			_attusage* a=&t->atts[j];
			SetIPSpecA(m,a,t->name.getTerminatedBuffer());
		}
	}
}


void CSaraHeaderReader::SetIPSpecOld()
{
	DOMElement* m=SetHeaderSection(HD_IPS);
	for (int i=0;i<m_nglobatts;i++) {
		_attusage* a=&m_globals[i];
		if (a->ip==IP_CDATA) continue;
		DOMElement* e=m_doc->createElement(s_xaira_item);
		m->appendChild(e);
		DOMElement* e1=m_doc->createElement(s_ident);
		e->appendChild(e1);
		UnicodeString gi("@");
		gi+=a->name;
		DOMText* t1=m_doc->createTextNode((const xairo_uchar*) gi.getTerminatedBuffer());
		e1->appendChild(t1);
		DOMElement* e2=m_doc->createElement(s_kw);
		e->appendChild(e2);
		DOMText* t2=m_doc->createTextNode((const xairo_uchar*) ipname[a->ip].getTerminatedBuffer());
		e2->appendChild(t2);
		if (!a->uri.isEmpty()) {
			DOMElement* e3=m_doc->createElement(s_ns);
			e->appendChild(e3);
			DOMText* t3=m_doc->createTextNode(a->uri.getTerminatedBuffer());
			e3->appendChild(t3);
		}
	}
	for (int i=0;i<m_ntags;i++) {
		_tagusage *t=&m_tagusage[i];
		if (t->ip!= EIP_DEFAULT || t->wb) {
			DOMElement* e=m_doc->createElement(s_xaira_item);
			m->appendChild(e);
			DOMElement* e1=m_doc->createElement(s_ident);
			e->appendChild(e1);
			UnicodeString gi(t->name);
			DOMText* t1=m_doc->createTextNode((const xairo_uchar*) gi.getTerminatedBuffer());
			e1->appendChild(t1);
			DOMElement* e2=m_doc->createElement(s_kw);
			e->appendChild(e2);
			UnicodeString sip=CodeEIP(t->ip);
			CodeEWB(t->wb,sip);
			DOMText* t2=m_doc->createTextNode((const xairo_uchar*) sip.getTerminatedBuffer());
			e2->appendChild(t2);
		}
		for (int j=0;j<t->natts;j++) {
			_attusage* a=&t->atts[j];
			if (a->ip==IP_CDATA) continue;
			DOMElement* e=m_doc->createElement(s_xaira_item);
			m->appendChild(e);
			DOMElement* e1=m_doc->createElement(s_ident);
			e->appendChild(e1);
			UnicodeString gi(t->name);
			gi+="@";
			gi+=a->name;
			DOMText* t1=m_doc->createTextNode((const xairo_uchar*) gi.getTerminatedBuffer());
			e1->appendChild(t1);
			DOMElement* e2=m_doc->createElement(s_kw);
			e->appendChild(e2);
			DOMText* t2=m_doc->createTextNode((const xairo_uchar*) ipname[a->ip].getTerminatedBuffer());
			e2->appendChild(t2);
			if (a->ip==IP_IDREF) {
				DOMElement* e3=m_doc->createElement(s_gi);  // was target
				e->appendChild(e3);
				DOMText* t3=m_doc->createTextNode((const xairo_uchar*) a->target.getTerminatedBuffer());
				e3->appendChild(t3);
			}
			if (!t->uri.isEmpty()) {
				DOMElement* e4=m_doc->createElement(s_ns);
				e->appendChild(e4);
				DOMText* t4=m_doc->createTextNode(t->uri.getTerminatedBuffer());
				e4->appendChild(t4);
			}
		}
	}
}


_tagusage* CSaraHeaderReader::GetTagUsage(UnicodeString gi)
{
		for (int i=0;i<m_ntags;i++) {
			_tagusage* t=&m_tagusage[i];
			if (gi==t->name) return t;
		}
		return NULL;
}

_tagusage* CSaraHeaderReader::GetTagUsage(UnicodeString gi,UnicodeString url)
{
		for (int i=0;i<m_ntags;i++) {
			_tagusage* t=&m_tagusage[i];
			if (gi==t->name && url==t->uri) return t;
		}
		return NULL;
}

_attusage* CSaraHeaderReader::GetAttUsage(UnicodeString gi,UnicodeString att)
{
	if (gi.isEmpty()||gi==s_asterisk) {
		for (int i=0;i<m_nglobatts;i++)
		{
			_attusage* a=&m_globals[i];
			if (att==a->name) return a; // check namespace here
		}
	} else {
		for (int i=0;i<m_ntags;i++) {
			_tagusage* t=&m_tagusage[i];
			if (gi!=t->name) continue;
			for (int j=0;j<t->natts;j++) {
				_attusage* a=&t->atts[j];
				if (att==a->name) return a;
			}
		}
	}
	return NULL;
}

_attusage* CSaraHeaderReader::GetAttUsage(UnicodeString gi,UnicodeString att,UnicodeString url)
{
	if (gi.isEmpty()||gi==s_asterisk) {
		for (int i=0;i<m_nglobatts;i++)
		{
			_attusage* a=&m_globals[i];
			if (att==a->name && url==a->uri) return a;
		}
	} else {
		for (int i=0;i<m_ntags;i++) {
			_tagusage* t=&m_tagusage[i];
			if (gi!=t->name || url != t->uri) continue;
			for (int j=0;j<t->natts;j++) {
				_attusage* a=&t->atts[j];
				if (att==a->name) return a;
			}
		}
	}
	return NULL;
}

// note that the bib file must be UTF-16 encoded

static UChar s_xarabibstart[]={'<','x','a','i','r','a',':','b','i','b','l','i','o','g','r','a','p','h','y',0};
static UChar s_xarabibend[]={'<','/','x','a','i','r','a',':','b','i','b','l','i','o','g','r','a','p','h','y','>',0};

bool CSaraHeaderReader::ReadBib(UnicodeString path)
{
	FILE* f=iopen(path,s_rb);
	if (f==NULL) return false;
	fseek(f,0,SEEK_END);
	int k=ftell(f)/2;
	fseek(f,0,SEEK_SET);
	printf("k=%d\n",k);
	xairo_uchar* b=mynew xairo_uchar[k+1];
	fread(b,2,k,f);
	b[k]=0;
	fclose(f);
	const xairo_uchar* p0=b; int t=0;
	while (p0<b+k) {
		const xairo_uchar* p=wcsstr((const xairo_uchar*) p0,s_xarabibstart);
		if (p==NULL) break;
		p=wcschr(p,'>');
		if (p==NULL) break;
		const xairo_uchar* q=wcsstr(p,s_xarabibend);
		if (q==NULL) break;
		t++;
		p0=q+wcslen(s_xarabibend);
	}
	printf("t=%d\n",t);
	m_bib=mynew _bib [t];
	p0=b; t=0;
	while (p0<b+k) {
		const xairo_uchar* p=wcsstr(p0,s_xarabibstart);
		if (p==NULL) break;
		p=wcschr(p,'>');
		if (p==NULL) break;
		m_bib[t].start=p+1-b;
		const xairo_uchar* q=wcsstr(p,s_xarabibend);
		if (q==NULL) break;
		m_bib[t].length=q-(b+m_bib[t].start);
		t++;
		p0=q+wcslen(s_xarabibend);
	}
	m_nbib=t;
	mydelete [] b;
	return true;
}


void CSaraHeaderReader::GetBibTagUsage()
{
	m_nbtags=0; 
	if (!m_bExists) return;
	DOMElement* root=GetHeaderSection(HD_BIBTAGS);
	DOMNodeList* ll=NULL; int k=0;
	if (root) {
		ll=root->getChildNodes();
		k=ll->getLength();
	}
	_tagusage* t0=mynew _tagusage [k];
	int j=0; int i;
	for (i=0;i<k;i++) {
		DOMNode* n0=ll->item(i);
		if (n0->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* tag=(DOMElement*) n0;
		if (wcscmp(tag->getTagName(),s_tagUsage) != 0) continue;
		t0[j].name=(const UChar*) tag->getAttribute(s_gi);
		t0[j].occurs=0;
		DOMElement* sub=(DOMElement*) tag->getFirstChild();
		t0[j].gloss="";
		if (sub != NULL && sub->getNodeType()== DOMNode::ELEMENT_NODE && 
				wcscmp(sub->getTagName(),s_desc)==0) {
			DOMText* gloss=(DOMText*) sub->getFirstChild();
			if (gloss != NULL) t0[j].gloss=(const UChar*) gloss->getData();
		}
		sub=(DOMElement*) sub->getNextSibling();
		GetAttUsage(sub, t0[j].natts,t0[j].atts);
		j++;
	}
	m_btagusage=mynew _tagusage[j];
	m_nbtags=j;
	for (i=0;i<j;i++) {
		m_btagusage[i]=t0[i];
	}
}



void CSaraHeaderReader::SetBibTagUsage()
{
	if (!m_bExists) return;
	DOMElement* e=SetHeaderSection(HD_BIBTAGS);
	return;
	DOMNodeList *l=m_doc->getElementsByTagName(s_btagsDecl);
	DOMNode* root=NULL;
	if (l->getLength()==0) {
		DOMNodeList *ll=m_doc->getElementsByTagName(s_fileDesc);
		if (ll->getLength()==0) return;  // should be fatal
		root=m_doc->createElement(s_btagsDecl);
		DOMNode* n=ll->item(0);
		n->appendChild(root);
	}
	else root=l->item(0); // should only be one but what the hell
// remove all its kids
	DOMNodeList* ll=root->getChildNodes();
	int k=ll->getLength(); int i;
	for (i=k-1;i>=0;i--) {
		DOMNode* n=ll->item(i);
		root->removeChild(n);
	}
	for (i=0;i<m_nbtags;i++) {
		if (m_btagusage[i].name[0]==0) continue;
		DOMElement* e=m_doc->createElement(s_tagUsage);
		e->setAttribute(s_gi,(const xairo_uchar*) m_btagusage[i].name.getTerminatedBuffer());
		root->appendChild(e);
		DOMElement* ee=m_doc->createElement(s_desc);
		e->appendChild(ee);
		DOMText* eee=m_doc->createTextNode((const xairo_uchar*) m_btagusage[i].gloss.getTerminatedBuffer());
		ee->appendChild(eee);
		DOMElement* au=m_doc->createElement(s_list);
		e->appendChild(au);
		au->setAttribute(s_type,s_attlist);
		int n=m_btagusage[i].natts;
		for (int j=0;j<n;j++) {
			_attusage* aus=&m_btagusage[i].atts[j];
			DOMElement* ex=m_doc->createElement(s_item);
			au->appendChild(ex);
			ex->setAttribute(s_n,(const xairo_uchar*) aus->name.getTerminatedBuffer());
			DOMElement* gl=m_doc->createElement(s_desc);
			ex->appendChild(gl);
			DOMText* t=m_doc->createTextNode((const xairo_uchar*) aus->gloss.getTerminatedBuffer());
			gl->appendChild(t);
		}
	}
}






FILE* iopen(UnicodeString s,const UChar* mode)
{
	return _wfopen((const xairo_uchar*) s.getTerminatedBuffer(),(const xairo_uchar*) mode);
}

UnicodeString CSaraHeaderReader::GetToken(UnicodeString& s)
{
	int k=s.length(); int i=0;
	while (i<k && s[i]<=' ') i++;
	if (i==k) return "";
	int j=i;
	while (j<k && s[j] > ' ') j++;
	UnicodeString tok;
	s.extract(i,j-i,tok);
	s.remove(0,j);
	return tok;
}

_addkey& _addkey::operator= (const _addkey& a)
{
	name=a.name;
	desc=a.desc;
	proc=a.proc;
	ngi=a.ngi;
	gi=mynew UnicodeString [ngi];
	for (int i=0;i<ngi;i++) {
		gi[i]=a.gi[i];
	}
	att=a.att;
	isLemma=a.isLemma;
	isPseudo=a.isPseudo;
	bCaseSensitive=a.bCaseSensitive;
	dproc=a.dproc;
	defval=a.defval;
	return *this;
}

_ls& _ls::operator= (const _ls& l)
{
	name=l.name;
	desc=l.desc;
	n=l.n;
	a=mynew UnicodeString [n];
	for (int i=0;i<n;i++) {
		a[i]=l.a[i];
	}
	return *this;
}

_region& _region::operator= (const _region& r)
{
	name=r.name;
	desc=r.desc;
	n=r.n;
	nak=r.nak;
	a=mynew UnicodeString [n];
	for (int i=0;i<n;i++) {
		a[i]=r.a[i];
	}
	return *this;
}

void CSaraHeaderReader::DelClass(_class* c)
{
	for (int i=0;i<c->n;i++) {
		DelClass(&c->l[i]);
	}
	if (c->n) mydelete [] c->l;
}

void CSaraHeaderReader::DelTaxonomies()
{
	for (int i=0;i<m_ntax;i++) {
		_taxonomy* t=&m_taxonomies[i];
		for (int j=0;j<t->n;j++) {
			DelClass(&t->l[j]);
		}
		if (t->n) mydelete [] t->l;
	}
	if (m_ntax) delete [] m_taxonomies;
}

void CSaraHeaderReader::DelTagUsage()
{
	if (m_nglobatts) DelAttUsage(m_nglobatts,m_globals);
	for (int i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		if (t->natts) DelAttUsage(t->natts,t->atts);
	}
	mydelete [] m_tagusage;
}

void CSaraHeaderReader::DelAttUsage(int n,_attusage* a)
{
	for (int i=0;i<n;i++) {
		_attusage* aa=&a[i];
		if (aa->cb) mydelete [] aa->cb;
	}
	mydelete [] a;
}
	

UnicodeString StringsToString(int n,UnicodeString* l)
{
	UnicodeString x;
	for (int i=0;i<n;i++) {
		x+=l[i];
		if (i<n-1) x+=",";
	}
	return x;
}


void StringToStrings(UnicodeString s,int& n,UnicodeString*& l)
{
	int k=-1; n=1;
	while ((k=s.indexOf((UChar) ',',k+1)) !=-1) n++;
	l=mynew UnicodeString [n];
	int k0=0;
	for (int i=0;i<n;i++) {
		k=s.indexOf((UChar) ',',k0);
		if (k==-1) k=s.length();
		l[i]=UnicodeString(s,k0,k-k0);
		k0=k+1;
	}
}

/*
void CSaraHeaderReader::DescTax(FILE* f,const UChar* name)
{
	GetTaxonomies();
	for (int i=0;i<m_ntax;i++) {
		_taxonomy* t=&m_taxonomies[i];
		fwprintf(f,L("att %s MULTIDREFS %s %s\r\n"),name,t->id.getTerminatedBuffer(),t->id.getTerminatedBuffer());
		fwprintf(f,L("menu 0 %s\r\n"),t->id.getTerminatedBuffer());
		_class* l=t->l;
		for (int j=0;j<t->n;j++) {
			_class* c=&l[j];
			fwprintf(f,L("item %s %s\r\n"),c->id.getTerminatedBuffer(),c->desc.getTerminatedBuffer());
		}
	}
}
*/
#if 0

void CSaraHeaderReader::MakeDSC(UnicodeString s)
{
	FILE* f=_wfopen(s.getTerminatedBuffer(),L("wt"));
//	fwprintf(f,L"\xfeff");
	fwprintf(f,L"ver %d\r\n",100+DBVER);
	fwprintf(f,L"# %s generated by Sara indextools\r\n",s);
	fwprintf(f,L"option noposindex\r\n",s);
	GetLabels();
	for (int i=0;i<m_nlabels;i++) {
		_lab* l=&m_labels[i];
		if (l->isSpec) {
			switch(l->e) {
			case SPE_LINE:
				fwprintf(f,L"label #LINE #AUTO\r\n");
				break;
			case SPE_FILE:
				fwprintf(f,L"label #FILE #AUTO\r\n");
				break;
			}
		} else if (l->isAuto) {
			fwprintf(f,L"label %s #AUTO\r\n",l->elt[0].getTerminatedBuffer());
		} else 
			fwprintf(f,L"label %s %s\r\n",l->elt[0].getTerminatedBuffer(),l->att.getTerminatedBuffer());
	}
	GetTagUsage();
	GetIPSpec();
	GetAddKeys();
	GetCodeBooks();
	GetTaxonomies();
	for (i=1;i<m_naddkeys;i++) {
		_addkey* a=&m_addkeys[i];
		if (a->isLemma) 
			fwprintf(f,L"lemmata %s %s\r\n",a->name.getTerminatedBuffer(),a->desc.getTerminatedBuffer());
		else 
			fwprintf(f,L"catt %s %d %s\r\n",a->name.getTerminatedBuffer(),i-1,a->att.getTerminatedBuffer());
		if (a->isPseudo) {
			for (int i=0;i<a->ngi;i++) {
				fwprintf(f,L"pos %s %s %s\r\n",a->name.getTerminatedBuffer(),
					a->gi[i].getTerminatedBuffer(),a->gi[i].getTerminatedBuffer());
			}
			if (a->dproc==_addkey::DP_USEVAL) {
				fwprintf(f,L"pos %s %s %s\r\n",a->name.getTerminatedBuffer(),
					a->defval.getTerminatedBuffer(),a->defval.getTerminatedBuffer());
			}
		}
	}
	for (int j=0;j<m_nglobatts;j++) {
		_attusage* a=&m_globals[j];
		switch (a->ip) {
			case IP_CONTENTA:
				fwprintf(f,L"att %s CATT 0\r\n",a->name.getTerminatedBuffer());
				break;
			case IP_ID:
				fwprintf(f,L"att %s ID 0\r\n",a->name.getTerminatedBuffer());
				break;
			case IP_IDREF:
				fwprintf(f,L"att %s IDREF %s\r\n",a->name.getTerminatedBuffer(),a->target.getTerminatedBuffer());
				break;
			default:
				fwprintf(f,L"att %s CDATA 0\r\n",a->name.getTerminatedBuffer());
		}
	}
	for (i=0;i<m_ntags;i++) {
		_tagusage* t=&m_tagusage[i];
		if (t->gloss.isEmpty())
			fwprintf(f,L"elt %s e b\r\n",t->name.getTerminatedBuffer());
		else 
			fwprintf(f,L"elt %s e b %s\r\n",t->name.getTerminatedBuffer(),t->gloss.getTerminatedBuffer());
		int k=t->natts;
		for (int j=0;j<k;j++) {
			_attusage* a=&t->atts[j];
			switch (a->ip) {
			case IP_CDATA:
				fwprintf(f,L"att %s CDATA 0\r\n",a->name.getTerminatedBuffer());
				if (a->bCB) {
					UnicodeString ca;
					for (int jj=0;jj<m_naddkeys;jj++) {
						_addkey* ak=&m_addkeys[jj];
						for (int jjj=0;jjj<ak->ngi;jjj++) {
							if (ak->gi[jjj]==t->name && ak->att==a->name)
								ca=ak->name.getTerminatedBuffer();
						}
					}
					UnicodeString t=a->gloss;
					if (t.isEmpty()) t=a->name;
					if (ca.isEmpty()) fwprintf(f,L"menu 0 %s\r\n",t.getTerminatedBuffer());
					for (int jj=0;jj<a->nc;jj++) {
						UnicodeString s=a->cb[jj].desc;
						if (s.isEmpty()) s=a->cb[jj].code;
						if (ca.isEmpty()) 
							fwprintf(f,L"item %s %s\r\n",a->cb[jj].code.getTerminatedBuffer(),s);
						else 
							fwprintf(f,L"pos %s %s %s\r\n",ca,a->cb[jj].code.getTerminatedBuffer(),s);
					}
				}
				break;
			case IP_ID:
				fwprintf(f,L"att %s ID 0\r\n",a->name.getTerminatedBuffer());
				break;
			case IP_IDREF:
				fwprintf(f,L"att %s IDREF %s\r\n",a->name.getTerminatedBuffer(),a->target.getTerminatedBuffer());
				break;
			case IP_NONE:
				fwprintf(f,L"att %s NULL 0\r\n",a->name.getTerminatedBuffer());
				break;
			case IP_TAX:
				DescTax(f,a->name.getTerminatedBuffer());
				break;
			}
		}

	}
	GetBibTagUsage();
	for (i=0;i<m_nbtags;i++) {
		_tagusage* t=&m_btagusage[i];
			fwprintf(f,L"eltb %s e b\r\n",t->name.getTerminatedBuffer());
		int k=t->natts;
		for (int j=0;j<k;j++) {
			_attusage* a=&t->atts[j];
				fwprintf(f,L"attb %s\r\n",a->name.getTerminatedBuffer());
		}
	}
	GetEntities();
	if (m_nents) {
		fwprintf(f,L"# char table\r\n");
		fwprintf(f,L"radix 16\r\n");
		for (int i=0;i<m_nents;i++) {
			_entity* e=&m_ents[i];
			fwprintf(f,L"char %s %04x\r\n",e->name,e->ch);
		}
	}
	GetRanges();
	for (int i=0;i<m_nuranges;i++) {
		fwprintf(f,L"charset %s\r\n",m_uranges[i].getTerminatedBuffer());
	}
	for (int i=0;i<m_nwtags;i++) {
		fwprintf(f,L"wtag %s\r\n",m_wtags[i].s.getTerminatedBuffer());
	}
/*	fwprintf(f,L"# configuration\n");
	GetXara();
	p=m_dsc.GetHeadPosition();
	while (p) {
		fprintf(f,"%s\n",m_dsc.GetNext(p));
	} */
	fclose(f);
//	((CMainFrame*) AfxGetMainWnd())->Log(L"save",L"Wrote "+s);
}

#endif

void CSaraHeaderReader::GetEntities()
{
/*
	m_ents.RemoveAll();
	SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
	parser->setFeature(L"http://xml.org/sax/features/validation",false);
	EntHandle handler;
	handler.m_ents=&m_ents;
	parser->setDeclarationHandler(&handler);
	parser->setEntityResolver(&handler);
	parser->setExitOnFirstFatalError(false);
	CString path=THEFILES->GetPath(1);
	parser->parse(path);
	delete parser;
*/
}

static UChar s_xara_encoding[]={'x','a','r','a',':','e','n','c','o','d','i','n','g',0};
static UChar s_xara_range[]={'x','a','r','a',':','r','a','n','g','e',0};

void CSaraHeaderReader::GetRanges()
{
	if (m_uranges) {
		delete [] m_uranges;
		m_uranges=NULL;
	}
	DOMElement* n0=GetUniqueNode(s_encodingDesc);
	if (n0==NULL) return;
	DOMElement* n=GetNamedNote(n0,s_xara_encoding);
	if (n==NULL) return;
	DOMNodeList* l=n->getChildNodes();
	int t=0;
	for (int i=0;i<(int) l->getLength();i++) {
		DOMNode* n=l->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) n;
		if (wcscmp(e->getTagName(),s_ident) != 0) continue;
		if (wcscmp(e->getAttribute(s_type),s_xara_range)!= 0) continue;
		t++;
	}
	m_nuranges=t;
	m_uranges=mynew UnicodeString [t];
	t=0;
	for (int i=0;i<(int) l->getLength();i++) {
		DOMNode* n=l->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* e=(DOMElement*) n;
		if (wcscmp(e->getTagName(),s_ident) != 0) continue;
		if (wcscmp(e->getAttribute(s_type),s_xara_range)!= 0) continue;
		m_uranges[t++]=(const UChar*) GetContent(e);
	}
}

void CSaraHeaderReader::CodeEWB(bool b,UnicodeString& s)
{
	if (b) s.insert(0,'*');
}

UnicodeString CSaraHeaderReader::CodeEIP(unsigned ip)
{
	char cc[4]="-+?";
	UnicodeString s;
	s+=UnicodeString((UChar) cc[EIP_THIS(ip)]);
	s+=(UChar) cc[EIP_CONTENT(ip)];
	s+=(UChar) cc[EIP_CHILDREN(ip)];
	return s;
}

UnicodeString CSaraHeaderReader::CodeEIP2(unsigned ip)
{
	bool bThis=(EIP_THIS(ip)>0);
	bool bContent=(EIP_CONTENT(ip)>0);
	bool bChildren=(EIP_CHILDREN(ip)>0);
	if (bThis) return "all";
	else if (bContent) return "content";
	else if (bChildren) return "markup";
	else return "none";
}

bool CSaraHeaderReader::DecodeEWB(UnicodeString& s) {
	bool b=false;
	if (s[0]=='*') {
		b=true;
		s.remove(0,1);
	}
	return b;
}

unsigned CSaraHeaderReader::DecodeEIP(UnicodeString s)
{
	int ip1, ip2, ip3;
	if (s.length() != 3) return EIP_DEFAULT;
	if (s[0]=='+') ip1=1;
	else if (s[0]=='-') ip1=0;
	else ip1=2;
	if (s[1]=='+') ip2=1;
	else if (s[1]=='-') ip2=0;
	else ip2=2;
	if (s[2]=='+') ip3=1;
	else if (s[2]=='-') ip3=0;
	else ip3=2;
	return EIP_MAKE(ip1,ip2,ip3);
}
	
unsigned CSaraHeaderReader::DecodeEIP2(UnicodeString s)
{
	if (s=="all")
		return EIP_MAKE(EIP_YES,EIP_YES,EIP_YES);
	else if (s=="markup")
		return EIP_MAKE(EIP_YES,EIP_NO,EIP_YES);
	else if (s=="content")
		return EIP_MAKE(EIP_NO,EIP_YES,EIP_NO);
	else if (s=="none")
		return EIP_MAKE(EIP_NO,EIP_NO,EIP_NO);
	else {
		HeaderError(HE_BAD_EIP,true,L("Unknown index policy %S"),s.getTerminatedBuffer());
		return EIP_DEFAULT;
	}

}


_valsource::_valsource()
{
	ngi=0;
	gi=NULL;
	bdefform=false;
	bcasefold=false;
	type=ident=ns=gloss=def=s_;
}

void CSaraHeaderReader::SetTagUsage()
{
	if (!m_bExists) return;
	DOMElement* e=SetHeaderSection(HD_TAGUSAGE);
	DOMElement* m=m_doc->createElement(s_xaira_item);
	m->setAttribute(s_type,s_element);
	m->setAttribute(s_ident,s_asterisk);
//	m->setAttribute(s_ns,L"");
	e->appendChild(m);
	DOMElement* al=m_doc->createElement(s_attList);
	m->appendChild(al);
	for (int j=0;j<m_nglobatts;j++) {
		_attusage* aus=&m_globals[j];
		SetAttInList2(aus,al);
	}		
	for (int i=0;i<m_ntags;i++) {
		if (m_tagusage[i].name.isEmpty()) continue;
		DOMElement* m=m_doc->createElement(s_xaira_item);
		m->setAttribute(s_type,s_element);
		m->setAttribute(s_ident,m_tagusage[i].name.getTerminatedBuffer());
		e->appendChild(m);
		for (int j=0;j<m_nglobatts;j++) {
			_attusage* aus=&m_globals[j];
			SetAttInList2(aus,m);
		}		
		if (!m_tagusage[i].uri.isEmpty()) m->setAttribute(s_ns,m_tagusage[i].uri.getTerminatedBuffer());
		if (!m_tagusage[i].gloss.isEmpty()) {
			DOMElement* g=m_doc->createElement(s_desc);
			m->appendChild(g);
			DOMText* t=m_doc->createTextNode(m_tagusage[i].gloss.getTerminatedBuffer());
			g->appendChild(t);
		}
		if (m_tagusage[i].natts==0) continue;
		DOMElement* al=m_doc->createElement(s_attList);
		m->appendChild(al);
		for (int j=0;j<m_tagusage[i].natts;j++) {
			_attusage* aus=&m_tagusage[i].atts[j];
			SetAttInList2(aus,al);
		}
	}
}

void CSaraHeaderReader::SetAttInList2(_attusage* aus,DOMElement* au)
{
	DOMElement* ex=m_doc->createElement(s_attDef);
	ex->setAttribute(s_ident,aus->name.getTerminatedBuffer());
	if (!aus->uri.isEmpty())
		ex->setAttribute(s_ns,aus->uri.getTerminatedBuffer());
	if (!aus->gloss.isEmpty()) {
		DOMElement* g=m_doc->createElement(s_desc);
		ex->appendChild(g);
		DOMText* t=m_doc->createTextNode(aus->gloss.getTerminatedBuffer());
		g->appendChild(t);
	}
	if (aus->bCB) {
		DOMElement* ev=m_doc->createElement(s_valList);
		SetCodeBook(aus,ev);
		ex->appendChild(ev);
	}
	au->appendChild(ex);
}

void CSaraHeaderReader::HeaderError(header_error he,bool bCanCont,const xairo_uchar* fmt,...)
{	
	IncErrCount();
	va_list list; 
	va_start(list,fmt);
	xairo_uchar buffer[1000];
	vswprintf(buffer,fmt,list);
	va_end(list);
	if (m_error_handler==NULL) return;
	bool bCont=m_error_handler(he,_wcsdup(buffer),m_lparam);
	if (!bCont || !bCanCont) throw new CHeaderError; 
}
/*

void CSaraHeaderReader::HeaderError(header_error he,const char* fmt,...)
{	
	IncErrCount();
	va_list list; 
	va_start(list,fmt);
	char buffer[1000];
	xairo_uchar wbuffer[1000];
	vsprintf(buffer,fmt,list);
	va_end(list);
	mbstowcs(wbuffer,buffer,1000);
	if (m_error_handler==NULL) return;
	bool bCont=m_error_handler(he,_wcsdup(wbuffer),m_lparam);
	if (!bCont) throw "Fatal error in header"; // weird - why a SAX error?

}
*/
