// SaraIndexContentHandler.cpp: implementation of the SaraIndexContentHandler class.
//
//////////////////////////////////////////////////////////////////////

#define _CRTDBG_MAP_ALLOC
#include "StdAfx.h"

#include "Corpus.h"
#include "SaraIndexContentHandler.h"
#include "TextReader.h"
#ifdef _WIN32
#include "..\xaira-tools\SaraHeaderReader.h"
#else
#include "SaraHeaderReader.h"
#endif

#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif

#undef THIS_FILE
static char THIS_FILE[] = __FILE__;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#define TALLOC 10000

SaraIndexContentHandler::SaraIndexContentHandler()
{
	m_loc=NULL;
//	m_locale=UnicodeString(loc);
	m_c=mynew xairo_uchar[TALLOC]; 
	m_c[0]=0;
	m_line=mynew int [TALLOC];
	m_col=mynew int [TALLOC];
	m_t=mynew xairo_uchar[TALLOC];
	m_alloc=TALLOC;
//	m_off=NULL;
//	m_labno[0]=m_labno[1]=0;
	m_ents=NULL;
	m_bCDATA=false;
	m_bWTag=false;
	m_bIndexMarkup=m_bIndexContent=true;
}

SaraIndexContentHandler::~SaraIndexContentHandler()
{
	mydelete [] m_c;
	mydelete [] m_t;
	mydelete [] m_line;
	mydelete [] m_col;
}

void SaraIndexContentHandler::SetOutput(CIndexer *out)
{
	m_indexer=out;
}

#define MAXNO 50





void SaraIndexContentHandler::startDocument()
{
	m_indexer->Index(it_start,UnicodeString(FILEN),0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(ENAME),m_indexer->m_name,0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(ENCODING),m_indexer->m_reader->m_enc,0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(PATH),m_indexer->m_relpath,0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(ATTACH),UnicodeString("/"),0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->m_bDoc=false;
	m_indexer->InitTax();
	TryFileLabel(m_indexer->m_name,m_indexer->m_if);
	m_indexer->IndexLabel(UnicodeString("unit"),m_indexer->GetReader()->m_nulllab,1,1,m_parent);
	m_indexer->IndexScope(1,1,m_parent);
	m_l1=0;
	m_lll=-1;
	Location();
	InitTasks();
	m_locale="-";
}

void SaraIndexContentHandler::endDocument()
{
	m_l0++;
	m_indexer->IndexLabel(UnicodeString("unit"),m_indexer->GetReader()->m_nulllab,m_l0,m_c0,m_parent);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);
	if (m_indexer->m_bDoc) {
		m_indexer->UndefTax(m_l0,m_c0,m_parent);
		m_indexer->Index(it_end,UnicodeString("document"),m_l0,m_c0,UnicodeString(XAIRA_URI),m_parent);
		m_indexer->Index(it_enddoc,UnicodeString(""),m_l0,m_c0,UnicodeString(""),m_parent);
	}
	m_indexer->Index(it_end,UnicodeString(FILEN),m_l0,m_c0,UnicodeString(XAIRA_URI),m_parent);

	Location();
}



void SaraIndexContentHandler::TryLang(UnicodeString el,const Attributes& attrs)
{
	int k=attrs.getLength(); int j;
	for (j=0;j<k;j++) {
		UnicodeString la("lang");
		if (la.compare(attrs.getLocalName(j))== 0) break;
	}
	if (j==k) return;
	int n=m_indexer->GetReader()->m_nlangs;
	_lang* l=NULL; int i;
	for (i=0;i<n;i++) {
		l=&m_indexer->GetReader()->m_langs[i];
		if (wcscmp((const xairo_uchar*) l->id,attrs.getValue(j))==0) break;
	}
	if (i == n) return;
	m_indexer->IndexLabel(UnicodeString("locale"),UnicodeString(l->locale),m_l0,m_c0,m_parent);
	m_indexer->SetLang(l->locale);
	Task(TSK_ENDLANG,0,0,m_locale);
	m_locale=l->locale;
}

static UChar tserr[]={'T','a','s','k',' ','s','t','a','c','k',' ','t','o','o',' ','d','e','e','p',',',
' ','s','w','i','t','c','h','e','d',' ','o','f','f','.',0};
static UChar nesterr[]={'N','e','s','t','e','d',' ','d','o','c','u','m','e','n','t',' ','l','a','b','e','l',' ','i','g','n','o','r','e','d',0};

void SaraIndexContentHandler::Task(taskop t,int i,int j,UnicodeString arg)
{
	if (m_ntasks >= MAXTASKS) {
		m_indexer->Error(m_l0,m_c0,tserr);
		return;
	}
	m_tasks[m_ntasks].t=t;
	m_tasks[m_ntasks].i=i;
	m_tasks[m_ntasks].j=j;
	m_tasks[m_ntasks].arg=arg;
	m_ntasks++;
}


void SaraIndexContentHandler::PushTasks()
{
	if (m_ntaskstack < MAXDEPTH) {
		for (int i=0;i<MAXTASKS;i++) {
			m_taskstack[m_ntaskstack*MAXTASKS+i]=m_tasks[i];
		}
	} else if (m_ntaskstack==MAXDEPTH) {
		m_indexer->Error(m_l0,m_c0,tserr);
	}
	m_ntaskstack++;
}

void SaraIndexContentHandler::InitTasks()
{
	m_ntaskstack=0;
}

void SaraIndexContentHandler::ClearTasks()
{
	m_ntasks=0;
	for (int i=0;i<MAXTASKS;i++) {
		m_tasks[i].t=TSK_NULL;
	}
}

void SaraIndexContentHandler::PopTasks()
{
	m_ntaskstack--;
	if (m_ntaskstack <MAXDEPTH) {
		for (int i=0;i<MAXTASKS;i++)
		{
			DoTask(&m_taskstack[m_ntaskstack*MAXTASKS+i]);
		}
	}
}

void SaraIndexContentHandler::DoTask(_task* t)
{
	switch (t->t) {
		case TSK_NULL:
			return;
		case TSK_ENDLANG:
			m_indexer->IndexLabel(UnicodeString("locale"),t->arg,m_l0,m_c0,m_parent);
			m_locale=UnicodeString(t->arg);
			m_indexer->SetLang(Ascii(t->arg));
			break;
		case TSK_ENDCA:
			{
			_addkey* a=&m_indexer->GetReader()->m_addkeys[t->i];
			a->cache=t->arg;
			a->bdef=t->j != 0;
			m_indexer->Regions(t->i,t->arg);
//			if (t->i==m_indexer->GetReader()->m_rak) m_indexer->m_dict.m_cr=m_indexer->GetReader()->Region(a->cache);

			}
			break;
		case TSK_ENDLAB:
			{
			_lab* l=&m_indexer->GetReader()->m_labels[t->i];
			if (!l->isMilestone) {
				l->cache=t->arg;
				l->bdef=t->j != 0;
			}
			}
			break;
		case TSK_ENDIP:
			{
				m_bIndexContent=t->i != 0;
				m_bIndexMarkup=t->j != 0;
			}
			break;
		case TSK_ENDPARENT:
			m_parent=t->i;
			break;


	}
}

static UChar s_pc_d[]={'%','d',0};

void SaraIndexContentHandler::TryExtraLabels(UnicodeString el,UnicodeString u,const Attributes& attrs)
{
	int n=m_indexer->GetReader()->m_nlabels;
	for (int i=2;i<n;i++) {
		_lab* l=&m_indexer->GetReader()->m_labels[i]; int j;
		if (l->ns != u) continue;
		for (j=0;j<l->ne;j++)
			if (l->elt[j]==el) break;
		if (j==l->ne) continue;
		if (l->bdef) return; // no nesting as yet
		UnicodeString sl;
		if (l->isAuto) {
			XMLCh k[MAXNO];
			swprintf(k,s_pc_d,l->seq++);
			sl=(const UChar*) k;
		} else {
			int k=attrs.getLength(); int j;
			for (j=0;j<k;j++) {
				if (l->att==(const UChar*) attrs.getLocalName(j)) break;
			}
			if (j < k)
				sl=(const UChar*) attrs.getValue(j);
		}
		Task(TSK_ENDLAB,i,l->bdef,l->cache);
		l->bdef=true;
		l->cache=sl;
	}
}

UnicodeString SaraIndexContentHandler::MakeUnique(UnicodeString s)
{
	return m_indexer->m_dm.CorpusTextName(s);
}

UnicodeString SaraIndexContentHandler::MakeLabel(int i)
{
	_lab* lab0=&m_indexer->GetReader()->m_labels[i];
	UnicodeString fmt=lab0->fmt;
	int n=m_indexer->GetReader()->m_nlabels;
	int x=0; int y; UnicodeString s;
	if (fmt.isEmpty()) if (lab0->bdef) return lab0->cache;
	else return "?";
	while (x < fmt.length()) {
		y=fmt.indexOf((UChar) '%',x);
		s.append(fmt,x,y-x);
		xairo_uchar c=fmt[y+1];
		int p=c-'0';
		if ( p >= 0 && p <= i) {
			_lab* l=&m_indexer->GetReader()->m_labels[p];
			if (l->bdef) s.append(l->cache);
			else s.append("?");
		} else if (c=='%') 
			s.append("%");
		else s.append(fmt,y,2);
		x=y+2;
	}
	return s;
}



void SaraIndexContentHandler::StructureLabel()
{
	for (int i=1;i<m_indexer->GetReader()->m_nlabels;i++) {
		_lab* l=&m_indexer->GetReader()->m_labels[i];
		if (l->bdef) return;
		if (l->isSpec) return; // no structure labels when lines are units
	}
//	m_indexer->IndexLabel(UnicodeString(L"unit"),m_indexer->GetReader()->m_nulllab,m_l0,m_c0);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);
}

static UChar s_note[]={'n','o','t','e',0};
static UChar dlfmt[]={'D','o','c','u','m','e','n','t',' ','l','a','b','e','l',' ','%','s',0};


void SaraIndexContentHandler::TryLabel(int i,UnicodeString el,UnicodeString u,const Attributes& attrs)
{
	int n=m_indexer->GetReader()->m_nlabels;
	if (i >= n) return;
	_lab* l=&m_indexer->GetReader()->m_labels[i];
	if (l->isSpec) return; 
	if (!l->bGlob) {
		int j;
//		if (l->ns != u ) return;
		for (j=0;j<l->ne;j++)
			if (l->elt[j]==el) break;
		if (j==l->ne && !l->bGlob) return;
	}
	else {
		int k=attrs.getLength(); int j;
		for (j=0;j<k;j++) {
			if (l->att==(const UChar*) attrs.getLocalName(j) &&
				l->ans==(const UChar*) attrs.getURI(j)) break;
		}
		if (j==k) return;
	}
	if (i==0 && l->bdef) {
		m_indexer->Error(m_l0,m_c0,nesterr);
		return;
	}
	if (l->bdef && !l->isMilestone) return; // was co for CSL
	if (i==0 && m_indexer->m_bDoc) {
		m_indexer->UndefTax(m_l0,m_c0,m_parent);
		m_indexer->Index(it_end,UnicodeString("document"),m_l0,m_c0,UnicodeString(XAIRA_URI),m_parent);
		m_indexer->Index(it_enddoc,UnicodeString(""),m_l0,m_c0,UnicodeString(""),m_parent);
	}
	if (i==0) {
		m_indexer->Index(it_start,UnicodeString("document"),m_l0,m_c0,UnicodeString(XAIRA_URI),m_parent);
		m_indexer->InitTax();
	}
// now evaluate the label
	UnicodeString sl;
	if (l->isAuto) {
		XMLCh k[MAXNO]; 
		swprintf(k,s_pc_d,l->seq++);
		sl=(const UChar*) k;
	} else {
		int k=attrs.getLength(); int j;
		for (j=0;j<k;j++) {
			if (l->att==(const UChar*) attrs.getLocalName(j)) break;
		}
		if (j<k) sl=(const UChar*) attrs.getValue(j);
		else sl="?";
	}
	Task(TSK_ENDLAB,i,l->bdef,l->cache);
	l->bdef=true;
	l->cache=sl;
// now emit at right level
	sl=MakeLabel(i);
	if (i==0) sl=MakeUnique(sl);
	m_indexer->IndexLabel(UnicodeString(i==0?"document":"unit"),sl,m_l0,m_c0,m_parent);
	if (l->isScope) m_indexer->IndexScope(m_l0,m_c0,m_parent);
	if (i==0) {
		m_indexer->Log(s_note,dlfmt,sl.getTerminatedBuffer());
		m_indexer->Bib(m_l0,m_c0,m_parent,sl);
		m_indexer->m_bDoc=true;
		for (int i=1;i<n;i++) {
			_lab* l=&m_indexer->GetReader()->m_labels[i];
			l->bdef=false;
			l->seq=0;
		}
		m_indexer->IndexLabel(UnicodeString("unit"),UnicodeString("none"),m_l0,m_c0,m_parent);
		m_indexer->IndexScope(m_l0,m_c0,m_parent);
	}

}


void SaraIndexContentHandler::TryFileLabel(UnicodeString name,int nf)
{
	int n=m_indexer->GetReader()->m_nlabels;
	if (nf>1) {
		if (n==0) return;
		_lab* l=&m_indexer->GetReader()->m_labels[0];
		if (!l->isSpec) return;
		if (l->e != SPE_FILE) return;
	}
	m_indexer->m_bDoc=true;
	m_indexer->Index(it_start,UnicodeString("document"),1,1,UnicodeString(XAIRA_URI),m_parent);
	UnicodeString fn;
	int xx=name.lastIndexOf((UChar) '.');
	if (xx==-1)
		fn=name;
	else
		fn.setTo(name,0,xx);
	m_indexer->IndexLabel(UnicodeString("document"),fn,1,1,m_parent);
	m_parent=0;
	m_indexer->Bib(1,1,m_parent,fn);
	for (int i=1;i<n;i++) {
		_lab* l=&m_indexer->GetReader()->m_labels[i];
		l->bdef=false;
		l->seq=0;
	}
	//	m_indexer->IndexLabel(L"unit",L"none",m_l0,m_c0);

}

void SaraIndexContentHandler::startPrefixMapping(const XMLCh* prefix,const XMLCh* uri)
{
	m_indexer->IndexSPM(UnicodeString((const UChar*) prefix),UnicodeString((const UChar*) uri),m_l0,m_c0,m_parent);
}

void SaraIndexContentHandler::endPrefixMapping(const XMLCh* prefix)
{
	m_indexer->IndexEPM(UnicodeString((const UChar*) prefix),m_l0,m_c0,m_parent);
}

void SaraIndexContentHandler::startElement
    (
        const   XMLCh* const    uri,
        const   XMLCh* const    localname,
        const   XMLCh* const    qname,
        const   Attributes&     attrs
	)
{
	bool bW=false;
	switch (m_indexer->GetReader()->TestWTag((const xairo_uchar*) localname)){
		case 0: // no w tags
			Commit(false);
			m_bWTag=true;
			break;
		case 1: // this is a w-tag
			Clear();
			bW=true;
			break;
		default: // there are w-tags but this isn't one
			if (!m_bWTag) Commit(false);
			break;
	}
	m_bFix=false;
	ClearTasks();
	for (int i=0;i<	m_indexer->GetReader()->m_nlabels;i++) {
		TryLabel(i,(const UChar*) localname,(const UChar*) uri,attrs);
	}
	if (!bW) StructureLabel(); bool bC=m_bIndexContent; bool bM=m_bIndexMarkup;
//	m_indexer->Index(it_start,UnicodeString(localname),m_l0,m_c0,UnicodeString(uri));
//	int seq=m_indexer->GetSeq();
	int ll=m_loc->getLineNumber();
	int cc=m_loc->getColumnNumber();
	m_indexer->Premap2(m_l0,m_c0);
	int seq=m_indexer->IndexStartTag(UnicodeString((const UChar*) localname),m_l0,m_c0,ll,cc,UnicodeString((const UChar*) uri),bC, bM,m_parent);
	Task(TSK_ENDPARENT,m_parent,0,"");
	if (seq !=-1) m_parent=seq;
	if (bC != m_bIndexContent || bM != m_bIndexMarkup) {
		Task(TSK_ENDIP,m_bIndexContent,m_bIndexMarkup,"");
		m_bIndexContent=bC;
		m_bIndexMarkup=bM;
	}
// Now set new IP


	int n=attrs.getLength();
	int rak=m_indexer->GetReader()->m_rak;
	for (int j=0;j<m_indexer->GetReader()->m_naddkeys;j++) {
		_addkey* a=&m_indexer->GetReader()->m_addkeys[j];
		a->bInd=false;
		if (!a->isPseudo) continue;
		for (int jj=0;jj<a->ngi;jj++) {
			if (a->gi[jj]==(const UChar*) localname) {
				Task(TSK_ENDCA,j,a->bdef,a->cache);
				a->cache=a->gi[jj];
				a->bdef=true;
				if (!a->isLemma && j != rak) m_indexer->IndexAddkey(a->name,a->cache,m_l0,m_c0,m_parent);
				else m_indexer->IndexNullAddkey(a->name,a->cache);
				m_indexer->Regions(j,a->cache);
			}
		}
	}
	for (int i=0;i<n;i++) {
		bool bIndexed=false;
		for (int j=0;j<m_indexer->GetReader()->m_naddkeys;j++) {
			_addkey* a=&m_indexer->GetReader()->m_addkeys[j];
			if (a->att==(const UChar*) attrs.getLocalName(i))
			{
				int jj=0;
				for (;jj<a->ngi;jj++) {
					if (a->gi[jj]==(const UChar*) localname) break;
				}
				if (a->ngi==0 || jj<a->ngi)
				{
					Task(TSK_ENDCA,j,a->bdef,a->cache);
					a->cache=(const UChar*) attrs.getValue(i);
					a->bdef=true;
					a->bInd=true;
					m_indexer->Regions(j,a->cache);
					if (!a->isLemma) m_indexer->IndexAddkey(a->name,a->cache,m_l0,m_c0,m_parent);
					else m_indexer->IndexNullAddkey(a->name,a->cache);
					bIndexed=true;
				}
			}
		}
//		if (bIndexed) continue; // comment out killed declarations
		if (!m_bIndexMarkup) continue;
		const XMLCh* u=attrs.getURI(i);
		if (u[0] || m_indexer->GetReader()->IsGlobal(localname,uri,attrs.getLocalName(i),attrs.getURI(i))) {
//		if (m_indexer->GetReader()->IsGlobal(localname,uri,attrs.getLocalName(i),attrs.getURI(i))) {
			m_indexer->IndexGlobalAttr(UnicodeString((const UChar*) attrs.getLocalName(i)),UnicodeString((const UChar*) attrs.getValue(i)),m_l0,m_c0,UnicodeString((const UChar*) u),m_parent);
		}
		else {
			m_indexer->IndexLocalAttr(UnicodeString((const UChar*) attrs.getLocalName(i)),UnicodeString((const UChar*) localname),UnicodeString((const UChar*) attrs.getValue(i)),m_l0,m_c0,UnicodeString((const UChar*) uri),m_parent);
		}
	}
// Deal with #undef values
	for (int j=0;j<m_indexer->GetReader()->m_naddkeys;j++) {
		_addkey* a=&m_indexer->GetReader()->m_addkeys[j];
		if (a->bInd) continue; int jj;
		for (jj=0;jj<a->ngi;jj++) {
			if (a->gi[jj]==(const UChar*) localname) break;
		}
		if (jj==a->ngi) continue;
		if (!a->isLemma) m_indexer->IndexAddkey(a->name,a->cache,m_l0,m_c0,m_parent);
	}
	TryLang((const UChar*) localname,attrs);
	PushTasks();
	Location();
}

void SaraIndexContentHandler::endElement
	(
		const XMLCh* const uri, 
		const XMLCh* const localname, 
		const XMLCh* const qname
		)
{
	switch (m_indexer->GetReader()->TestWTag(localname)){
		case 0: // no w tags
			Commit(m_indexer->GetReader()->TestWB(localname));
			m_bWTag=false;
			break;
		case 1:
			Commit0();
			break;
		case 2:
			if (!m_bWTag) Commit(m_indexer->GetReader()->TestWB(localname));
			break;
	}
	m_bFix=false;
//	Commit();
	PopTasks();
	int ll=m_loc->getLineNumber();
	int cc=m_loc->getColumnNumber();
//	m_indexer->Premap2(m_l0,m_c0);
	m_indexer->IndexEndTag(UnicodeString((const UChar*) localname),m_l0,m_c0,ll,cc,UnicodeString((const UChar*) uri),m_bIndexMarkup,m_parent);
	Location();
}


int sz(int c)
{
	if (c=='&') return 5;
	if (c=='<') return 4;
	return 1;
}

void SaraIndexContentHandler::ignorableWhitespace
    (
        const   XMLCh* const    chars
        , const XMLSize_t    length
		)
{
	characters(chars,length);
}

// redo this to use a large expandable buffer
void SaraIndexContentHandler::characters
    (
        const   XMLCh* const    chars
        , const XMLSize_t    length
		)
{
	
	int n0=wcslen(m_c);
	int n=n0+length;
	if (n+1>m_alloc) {
		XMLCh* c=m_c; xairo_uchar* t=m_t;
		int* line=m_line; int* col=m_col;
		while (m_alloc <= n) m_alloc+=TALLOC;
		m_c=mynew xairo_uchar [m_alloc];
		m_line=mynew int[m_alloc];
		m_col=mynew int [m_alloc];
		m_t=mynew xairo_uchar [m_alloc];
		wcscpy(m_c,c);
		for (int i=0;i<n0;i++) {
			m_line[i]=line[i];
			m_col[i]=col[i];
			assert(m_col[i]>0);
		}
		mydelete [] c;
		mydelete [] line;
		mydelete [] col;
		mydelete [] t;
	}
	wcsncat(m_c,chars,length);
	m_c[n]=0;

/* BUG
When Xerces encounters a character entity it flushes its content buffer but reports the position
as though it had read the & of the entity. If this is not the first piece of content we must
now correct the last bit for this. Note that this will not work in other Sax parsers. Note also 
that we must commit content at the end of an external entity for this to work. */
	if (n0 > 0 && m_bFix) {
		int ll=m_line[n0-1]; // was m_n0
		for (int i=m_n0+1;i<n0;i++) {
//		for (int i=n0-1;i>=m_n0+1;i--) {
//			if (m_line[i]==ll) m_col[i]--;
			m_col[i]--;
			if (m_col[i]==0) {
				m_line[i]--; 
				m_col[i]=m_indexer->m_reader->LineLength(m_line[i])+1;
				assert(m_col[i] > 0);
			}
//			m_col[i]=m_col[i-1]; m_line[i]=m_line[i-1];
		}
//		if (m_l0==ll) m_c0--; 
		m_c0--; 
		if (m_c0==0) {
			m_l0--;
			m_c0=m_indexer->m_reader->LineLength(m_l0)+1;
		}
		assert(m_c0 > 0);
	}
/*
End of bug fix */

//	for (i=0;i<n-n0;i++) {
//		m_line[i+n0]=m_l0;
//		m_col[i+n0]=m_c0;
//		if (m_c[i+n0]==10) {
//			m_l0++; m_c0=1;
//		} else m_c0++;
//	}
	m_line[n0]=m_l0;
	m_col[n0]=m_c0;
	assert(m_col[n0] > 0);
	Location();
	int l0=m_l0; int c0=m_c0;
	for (int i=n-n0-1;i>0;i--) {
		if (c0==1) {
			l0--; c0=m_indexer->m_reader->LineLength(l0)+1;
//			assert(m_c[i+n0]==10); false if later corrected above - darn, was a good assert!
		} else c0--;
		m_line[i+n0]=l0;
		m_col[i+n0]=c0; assert(c0 > 0);
	}
	m_n0=n0;
	m_bFix=true;
//	Location();
}

void SaraIndexContentHandler::Reset()
{
	Clear();
}
void SaraIndexContentHandler::Clear()
{
	m_c[0]=0; 
}




void SaraIndexContentHandler::Commit0()
{
	if (!m_bIndexContent) return;
	_addkey* a=&m_indexer->GetReader()->m_addkeys[0];
	UnicodeString u;
	if (a->proc==_addkey::AKP_FORM)
		u=(const UChar*) m_c;
	else {
		u=a->cache;
	}
//	if (u==NULL) return;
	UnicodeString uu=NormSpace(u);
	m_indexer->IndexWord((const xairo_uchar*) uu.getTerminatedBuffer(),m_line[0],m_col[0],m_l0,m_c0,m_parent);
	m_c[0]=0;
}

UnicodeString SaraIndexContentHandler::NormSpace(UnicodeString s)
{
	UnicodeString ss;
	bool bSpace=true; int x=0;
	for (int i=0;i<s.length();i++) {
		UChar c=s.charAt(i);
		if (u_isUWhiteSpace(c)|| c==' ') {
			if (bSpace) continue;
			else {
				ss.append((UChar) ' ');
				x++;
				bSpace=true;
			}
		} else {
			ss.append(c);
			x++;
			bSpace=false;
		}
	}
	if (bSpace && x>0) ss.truncate(x-1);
	return ss;
}


void SaraIndexContentHandler::Commit(bool bNow)
{
	int lc=wcslen(m_c);
	CTokeniser* t=m_indexer->GetTokeniser();
	m_line[lc]=m_l0; m_col[lc]=m_c0; assert(m_c0 > 0);
	UnicodeString tok(m_c);
	t->SetText(tok);
	int k=0; int kk=k; int nn=0;
	while (t->GetNext(k) && (k < lc || (k==lc && bNow))) {
		for (int i=0;i<k-kk;i++) {
			m_t[i]=m_c[i+kk];
		}
		m_t[k-kk]=0;
		if (m_bIndexContent) m_indexer->IndexWord(m_t,m_line[kk],m_col[kk],m_line[k],m_col[k],m_parent);
		kk=k;
	}
	for (int i=0;i<lc-kk;i++) {
		m_c[i]=m_c[kk+i];
		m_line[i]=m_line[kk+i];
		m_col[i]=m_col[kk+i];
		assert(m_col[i] > 0);
	}
	m_c[k-kk]=0;
	if (lc > kk) m_indexer->Premap(m_line[0],m_col[0]);
}

void SaraIndexContentHandler::error(const SAXParseException& exception)
{
	m_indexer->Error(exception);
	Location();
}

void SaraIndexContentHandler::fatalError(const SAXParseException& exception)
{
	m_indexer->Error(exception);
	Location();
}

void SaraIndexContentHandler::Location()
{
	if (m_loc==NULL) return;
	m_l0=m_loc->getLineNumber();
	m_c0=m_loc->getColumnNumber();
	if (m_bCDATA) m_c0-=3;
}

void SaraIndexContentHandler::processingInstruction
    (
        const   XMLCh* const    target
        , const XMLCh* const    data
		) 
{
	m_bFix=false;
	m_indexer->IndexPI(UnicodeString((const UChar*) target),UnicodeString((const UChar*) data),m_l0,m_c0,m_parent);
	Location();
}

void SaraIndexContentHandler::comment
    (
        const   XMLCh* const    target
        , XMLSize_t c
		) 
{
//	Commit();
//	Location();
}

void SaraIndexContentHandler::startCDATA()
{
	m_bCDATA=true;
	Commit(false);
//	Location();
}

void SaraIndexContentHandler::endCDATA()
{
	m_bCDATA=false;
	Commit(false);
	Location();
}

static UChar s_3pc[]={'%','d','-','%','d','-','%','d',0};

static UChar s_exterr[]={'E','x','t','e','r','n','a','l',' ','e','n','t','i','t','y',' ','%','s',' ','m','u','s','t',' ','b','e',' ',
's','p','e','c','i','f','i','e','d',' ','b','y',' ','r','e','l','a','t','i','v','e',' ','p','a','t','h',0};

void SaraIndexContentHandler::startEntity(const XMLCh* const name)
{
	if (ignore((const UChar*) name)) return;
	Commit(true);
	m_indexer->IndexLabel(UnicodeString("unit"),UnicodeString("none"),m_l0,m_c0,m_parent);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);
	int i0=m_indexer->m_if; int w0=m_indexer->m_nw;
	int o=m_indexer->m_reader->Offset(m_l0,m_c0);
	xairo_uchar at[1000];
	swprintf(at,s_3pc,i0,o+1,w0);
	UnicodeString path=resolve((const UChar*) name);
	if (path[0]==0 || path[0]=='/' || path[0]=='\\') {
		xairo_uchar w[1000];
		swprintf(w,s_exterr,name);
		m_indexer->Error(m_l0,m_c0,w);
	}
	m_indexer->m_dict.m_corpus->StartEntity((const UChar*) name,path);
	Location();
	m_indexer->IndexGlobalAttrVal(UnicodeString(ENAME),UnicodeString((const UChar*) name),0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(ENCODING),m_indexer->m_reader->m_enc,0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(PATH),m_indexer->m_relpath,0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexGlobalAttrVal(UnicodeString(ATTACH),UnicodeString((const UChar*) at),0,1,UnicodeString(XAIRA_URI),m_parent);
	m_indexer->IndexLabel(UnicodeString("unit"),UnicodeString("none"),m_l0,m_c0,m_parent);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);

}

void SaraIndexContentHandler::endEntity(const XMLCh* const name)
{
	if (ignore((const UChar*) name)) return;
	Commit(true);
	m_indexer->IndexLabel(UnicodeString("unit"),UnicodeString("none"),m_l0,m_c0,m_parent);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);
	m_indexer->m_dict.m_corpus->EndEntity();
	Location();
	m_indexer->IndexLabel(UnicodeString("unit"),UnicodeString("none"),m_l0,m_c0,m_parent);
	m_indexer->IndexScope(m_l0,m_c0,m_parent);
}

void SaraIndexContentHandler::internalEntityDecl(const XMLCh* const a, const XMLCh* const b)
{
	UnicodeString sa((const UChar*) a);
	UnicodeString sb((const UChar*) b);
	m_indexer->IndexInternalEntity(sa,sb);
}


void SaraIndexContentHandler::externalEntityDecl(const XMLCh* const name,const XMLCh* const pid,const XMLCh* const sid)
{
	_entdecl* e=mynew _entdecl;
	e->name=(const UChar*) name;
	e->sid=(const UChar*) sid;
	e->next=NULL;
	_entdecl** l=&m_ents;
	while (*l) l=&((*l)->next);
	*l=e;
}

UnicodeString SaraIndexContentHandler::resolve(UnicodeString name)
{
	_entdecl* e=m_ents;
	while (e) {
		if (name==e->name) return e->sid;
		e=e->next;
	}
	return "";
}

bool SaraIndexContentHandler::ignore(UnicodeString name)
{
	if (name=="[dtd]") return true;
	_entdecl* e=m_ents;
	while (e) {
		if (name==e->name) return false;
		e=e->next;
	}
	return true;
}


