// SaraIndexContentHandler.h: interface for the SaraIndexContentHandler class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SARAINDEXCONTENTHANDLER_H__236E7817_F6D8_47B6_B66E_271892F102CE__INCLUDED_)
#define AFX_SARAINDEXCONTENTHANDLER_H__236E7817_F6D8_47B6_B66E_271892F102CE__INCLUDED_

typedef unsigned short XMLChar;

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

using namespace xercesc;

#ifndef _WIN32
#include "./SaraDefHandler.h"
#else
#include "../xaira-tools/SaraDefHandler.h"
#endif

#include "Tokeniser.h"
#include "Indexer.h"

struct _entdecl {
	UnicodeString name;
	UnicodeString sid;
	_entdecl* next;
};

enum taskop {TSK_ENDLANG,TSK_ENDCA,TSK_ENDLAB,TSK_ENDIP,TSK_ENDPARENT,TSK_NULL};

struct _task {
	taskop t;
	int i,j;
	UnicodeString arg;
	_task():t(TSK_NULL){}
	_task(taskop t0,int i0,int j0,UnicodeString arg0):t(t0),i(i0),j(j0),arg(arg0){}
	virtual ~_task() {}
};


#define MAXTASKS 5
#define MAXDEPTH 100

class SaraIndexContentHandler : public CSaraDefHandler  
{
	friend class CIndexer;
public:
	SaraIndexContentHandler();
	virtual ~SaraIndexContentHandler();
    void characters
    (
        const   XMLCh* const    chars
        , const XMLSize_t    length
		);
	void endDocument();
    void endElement
	(
		const XMLCh* const uri, 
		const XMLCh* const localname, 
		const XMLCh* const qname
		);
    void ignorableWhitespace
    (
        const   XMLCh* const    chars
        , const XMLSize_t    length
		) ;
    void processingInstruction
    (
        const   XMLCh* const    target
        , const XMLCh* const    data
		);
    void setDocumentLocator(const Locator* const locator) {m_loc=locator;}
    void startDocument();
    void startElement
    (
        const   XMLCh* const    uri,
        const   XMLCh* const    localname,
        const   XMLCh* const    qname,
        const   Attributes&     attrs
		); 
	void startPrefixMapping
	( 
		const	XMLCh* const	prefix,
		const	XMLCh* const	uri
		); 
	void endPrefixMapping
	(
		const	XMLCh* const	prefix
		);
	void skippedEntity
	(
		const	XMLCh* const	name
		) {}
    void comment
    (
        const   XMLCh* const    chars
        , const XMLSize_t    length
		);
	virtual void startCDATA ();
	virtual void endCDATA();
	void startEntity(const XMLCh* const);
	void endEntity(const XMLCh* const);
	void externalEntityDecl(const XMLCh* const,const XMLCh* const,const XMLCh* const);
	void internalEntityDecl(const XMLCh* const,const XMLCh* const);
    void warning(const SAXParseException& exception) {}
    void error(const SAXParseException& exception); 
    void fatalError(const SAXParseException& exception); 
	void resetErrors() {}
	void SetOutput(CIndexer *out);
//	CTokeniser m_tok;
	XMLCh *m_c, *m_t;
	int *m_line, *m_col;
	int m_n0;
	int m_alloc;
	CIndexer *m_indexer;
	int m_parent;
	void Reset();
//	class CTextReader *m_reader;
private:
	int m_cr;
	void Commit(bool);
	void Commit0();
	void Clear();
	int m_l0, m_l1, m_c0;
	int m_lll;
	void Location();
	const Locator *m_loc;
	void TryLang(UnicodeString,const Attributes&);
	void TryLabel(int,UnicodeString,UnicodeString,const Attributes&);
	void TryExtraLabels(UnicodeString,UnicodeString,const Attributes&);
	void TryFileLabel(UnicodeString,int);
	void TryLineLabel();
	void StructureLabel();
	UnicodeString MakeLabel(int);
	UnicodeString MakeUnique(UnicodeString);
	//	int m_labno[2];
	UnicodeString resolve(UnicodeString);
	bool ignore(UnicodeString);
	_entdecl* m_ents;
	void InitTasks();
	void PushTasks();
	void PopTasks();
	void ClearTasks();
	void Task(taskop,int,int,UnicodeString);
	void DoTask(_task*);
	_task m_tasks[MAXTASKS];
	int m_ntasks;
	_task m_taskstack[MAXTASKS*MAXDEPTH];
	int m_ntaskstack;
	UnicodeString m_locale;
	UnicodeString NormSpace(UnicodeString);
	bool m_bCDATA;
	bool m_bWTag;
	bool m_bFix;
	bool m_bIndexMarkup, m_bIndexContent;

};

#endif // !defined(AFX_SARAINDEXCONTENTHANDLER_H__236E7817_F6D8_47B6_B66E_271892F102CE__INCLUDED_)
