// newindexer.cpp : Defines the entry point for the console application.
//

#include "StdAfx.h"

#ifdef _WIN32
#include <windows.h> // eek
#endif

#include "Corpus.h"
#include "MemoryErr.h"
#include "TextReader.h"

void console(const xairo_uchar* w);
void console(const char* err);
void Usage();
void FatalError(const char*,const char*);
void FatalError(const char* s,const xairo_uchar*);

#define PARAMFILE "corpus_parameters.xml"

int Process(const char* paramfile,bool bDump,bool bFreq,bool bHdr,bool bMem,int kLocs,int,char*,char*);

void term();

struct _term {};

int main(int argc, char* argv[])
{
	bool bErr=false;bool bDump=false; bool bFreq=false; bool bHdr=true; bool bCache=false;
	char cl[800]; char sBrk[500];
	cl[0]=0; sBrk[0]=0;
	int kLocs=MINLOCS;
	int kStrength=5;
	bool bMem=false;
	bool bValid=true;
	char paramfile[_MAX_PATH];
	strcpy(paramfile,PARAMFILE);
	for (int i=1;i<argc;i++) {
		strcat(cl,argv[i]);
		strcat(cl," ");
	}
	for (int i=1;i<argc;i++) {
		if (strcmp(argv[i],"-p")==0 && i+1 < argc) {
			strcpy(paramfile,argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-d")==0) {
			bDump=true;
		} else if (strcmp(argv[i],"-f")==0) {
			bFreq=true;
		} else if (strcmp(argv[i],"-e")==0) {
			CTextReader::DumpAll("MIME");
			return 0;
		} else if (strcmp(argv[i],"-m")==0) {
			bMem=true;
		} else if (strcmp(argv[i],"-l")==0 && i+1 < argc) {
			kLocs=atoi(argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-s")==0 && i+1 < argc) {
			kStrength=atoi(argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-b")==0 && i+1 < argc) {
			strcpy(sBrk,argv[i+1]);
			i++;
		}
		else
			bErr=true;
	}
	if (bErr) 
	{	
		Usage();
		FatalError("Command line error",cl);
		return 1;
	}
	int k=0;
	k=Process(paramfile,bDump,bFreq,bHdr,bMem,kLocs,kStrength,cl,sBrk);

	if (k) return k;
#ifdef _DEBUG
	HANDLE herr=CreateFile("dump.txt",GENERIC_WRITE, FILE_SHARE_WRITE, 
      NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	_CrtSetReportMode(_CRT_WARN,_CRTDBG_MODE_FILE);
	_CrtSetReportFile(_CRT_WARN,herr);
	_CrtMemState m;
	_CrtMemCheckpoint(&m);
	_CrtMemDumpStatistics(&m);
	_CrtDumpMemoryLeaks();
	CloseHandle(herr);
#endif
	return k;
}

void term()
{
	exit(100);
}

int Process(const char* paramfile,bool bDump,bool bFreq,bool bHdr,bool bMem,int kLocs,int kStrength,char* cl,char* sBrk)
{
    try
    {
         XMLPlatformUtils::Initialize();

    }
    catch (const XMLException& toCatch)
    {
		FatalError("Error during Xerces initialization",toCatch.getMessage());
         return 2;
    }
    CCorpus c(bDump,bFreq,bHdr,bMem,kLocs,kStrength,cl,sBrk);
	try {
		c.ReadParameters(paramfile);
	}
	catch (SAXException& e)
	{
		FatalError("Error processing parameter file",e.getMessage());
		return 3;
	}
	try {
		c.ReadFileList();

	}
	catch (SAXException& e)
	{
		FatalError("Error processing file list",e.getMessage());
		return 4;
	}
	try 
	{
		c.TestIndexDir();
	}
	catch (CDiskWriteErr*)
	{
		FatalError("Index directory does not exist or cannot be written.","");
		return 6;
	}
	c.IndexPhase1();

//	XMLPlatformUtils::Terminate();

    return c.Status();
}


void FatalError(const char* s,const char* p)
{
	FILE* f=fopen("fatal_err.txt","w");
	if (f==NULL) return;
	fprintf(f,"%s: %s",s,p);
	printf("%s: %s\n",s,p);
	fclose(f);
}

void FatalError(const char* s,const xairo_uchar* p)
{
	char* c=new char[wcslen(p)+1];
	wcstombs(c,p,wcslen(p)+1);
	FILE* f=fopen("fatal_err.txt","w");
	if (f==NULL) return;
	fprintf(f,"%s: %s",s,c);
	printf("%s: %s\n",s,c);
	fclose(f);
	delete [] c;
}
	
using namespace std;

void Usage()
{
	cout << "Usage: newindexer [-d] [-f] [-p params]\n";
	cout << "-d: dump index as text\n";
	cout << "-f: make frequency table\n";
	cout << "-l d: set loc buffer to hold d entries\n"; 
	cout << "-m: dump system memory parameters\n";
	cout << "-p params: use params as parameter file\n";
	cout << "-e: list known converters and exit\n";
	cout << "-s: run an old-style Sara server";
	cout << "\n";
}