// xaira_daemon.cpp : Defines the entry point for the console application.
//

#define _CRT_SECURE_NO_DEPRECATE
#define _CRT_NON_CONFORMING_SWPRINTFS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#ifndef _WIN32
#define _getpid getpid
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <sys/wait.h>
#include <pwd.h>
#include <fcntl.h>
#ifdef HAVE_ASM_BYTEORDER_H
#include <asm/byteorder.h>
#endif
#include <arpa/inet.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifndef HAVE_SOCKLEN_T
typedef int socklen_t;
#endif
typedef int PSOCKET;
#define INVALID_SOCKET -1
/*
struct in_addr
{
        union
        {
               struct { u_char s_b1, s_b2, s_b3, s_b4; };
               struct { u_short s_w1, s_w2; };
               u_long S_addr;
       } S_un;
};

struct  sockaddr_in
{
       short    sin_family;
       u_short  sin_port;
       struct   in_addr sin_addr;
       char     sin_zero;
};
*/

//#define s_addr S_un.S_addr
#define INADDR_ANY 0
#define SOCKET_ERROR -1
#define closesocket(x) close(x)
#define SEPCHAR '/'
#define _itoa itoa
#ifdef PATH_MAX
#define _MAX_PATH PATH_MAX
#else
#define _MAX_PATH 1024
#endif
#else
#include <process.h>
#include <winsock2.h>
#include <ws2tcpip.h>
typedef SOCKET PSOCKET;
typedef int socklen_t;
#define SEPCHAR '\\'
#endif

#include <unicode/unistr.h>

#include "../XairoHeader.h"
#include "Users.h"

#include "XairaServer.h"
#include "XairaXMLServer.h"

using namespace XAIRO_NAMESPACE;

#define LOGFILE "xaira_access_log.txt"
#ifndef _WIN32
void InitInt(PSOCKET,CXairaServer*);

#ifdef LOGPATH
#define LOGDIR LOGPATH
#else
#define LOGDIR "/tmp"
#endif
#ifndef DATAPATH
#error The symbol DATAPATH must be defined.
#endif
#else
#define LOGDIR "."
#endif

#define IPV6_SUPPORT

#define MAXTRY 3

void Usage();
void Session(PSOCKET);
void NewSession(PSOCKET,PSOCKET,const char*);
void SessionLog(const char*);
void SessionLogErr();
void OpenSessionLog();
void CloseSessionLog();
void detach();

void Message(UnicodeString a)
{
	char* s=new char[a.length()+1];
	a.extract(0,a.length(),s);
	printf("%s",s);
	delete [] s;
}

class MyBootLog : public CXairoIBootMonitor
{
public:
	void Log(const xairo_uchar* u);

};

void MyBootLog::Log(const xairo_uchar* u) 
{
	Message(u);
}



CXairaServer* server=NULL;
char logdir[_MAX_PATH];
char logpath[_MAX_PATH];
char pidpath[_MAX_PATH];
char datapath[_MAX_PATH];
char cname[_MAX_PATH];
bool bFork=true;
bool bLog=false;
bool bSara=false;
bool bSecure=false;
CUsers users;
int maxtry=MAXTRY;

#define MAXTIME 100

void Log(const char* s,...)
{
	va_list arg;
	va_start(arg,s);
	FILE* f=fopen(logpath,"a");
	if (f==NULL) return;
	char ts[MAXTIME];
	time_t t;
	time(&t);
	strftime(ts,MAXTIME,"%c: ",localtime(&t));
	fprintf(f,"%s",ts);
	vfprintf(f,s,arg);
	fprintf(f,"\n");
	va_end(arg);
	fclose(f);
}

#ifndef _WIN32
void catch_child(int sig)
{
   int   status;
   pid_t pid;
   
   if (sig == SIGCHLD) {
      while (   (pid = waitpid(-1, &status, WNOHANG)) == -1
             && (errno == EINTR))
         ;
      if (WIFSIGNALED(status))
         Log("Child pid=%d was killed with signal %d",
                pid, WTERMSIG(status));
      else if (WIFEXITED(status))
         Log("Child pid=%d exited with status %d",
                pid, WEXITSTATUS(status));
   }
}
#endif

int main(int argc, char* argv[])
{
	bool bErr=false;
	unsigned lport=7000;
	char paramfile[_MAX_PATH];
	paramfile[0]=0;
	strcpy(logdir,LOGDIR);
	const char* lp=getenv("XAIRA_LOGPATH");
	if (lp) strcpy(logdir,lp);
#ifdef DATAPATH
	strcpy(datapath,DATAPATH);
#endif
	const char* dpe=getenv("XAIRA_DATAPATH");
	if (dpe) strcpy(logdir,dpe);
	for (int i=1;i<argc;i++) {
		if (strcmp(argv[i],"-p")==0 && i+1 < argc) {
			strcpy(paramfile,argv[i+1]);
			i++;
		}
		else if (strcmp(argv[i],"-a")==0 && i+1 < argc) {
			strcpy(logdir,argv[i+1]);
			i++;
		}
		else if (strcmp(argv[i],"-d")==0 && i+1 < argc) {
			strcpy(datapath,argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-l")==0 && i+1 < argc) {
			lport=atoi(argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-f")==0) {
			bFork=false;
		} else if (strcmp(argv[i],"-g")==0) {
			bLog=true;
		} else if (strcmp(argv[i],"-u")==0) {
			bSecure=true;
		} else if (strcmp(argv[i],"-s")==0) {
			bSara=true;
		}
		else
			bErr=true;
	}
	if (bErr || paramfile[0]==0 || (bSara && bSecure)) 
	{	
		Usage();
		return 1;
	}
// initialise the corpus
	int ll=(int) strlen(logdir);
	if (ll > 0 && logdir[ll-1]!=SEPCHAR) {
		logdir[ll]=SEPCHAR;
		logdir[ll+1]=0;
	}
	ll=(int) strlen(datapath);
	if (ll > 0 && datapath[ll-1]==SEPCHAR) {
		datapath[ll-1]=0;
	}
	if (bSecure) {
		if (!users.Init(datapath,maxtry)) {
			printf("Cannot open user authentication file\n");
			return 2;
		}
	}
	strcpy(logpath,logdir);
	strcat(logpath,LOGFILE);
	strcpy(pidpath,logdir);
	strcat(pidpath,"xaira_daemon.pid");
	if (bSara) server=new CXairaSaraServer;
	else server=new CXairaXMLServer;
	if (!server->Init(paramfile,bLog,cname))
		return 3;
// (Windows only) Setup sockets
#ifdef _WIN32
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	wVersionRequested = MAKEWORD( 2, 2 );
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		return 2;
	}
#endif
// create the listening socket
	PSOCKET sock=socket(AF_INET, SOCK_STREAM, 0);
	if (sock==INVALID_SOCKET) {
      printf("Error opening socket\n");
      return 2;
    }
	struct sockaddr_in passive;
    passive.sin_family = AF_INET;
    passive.sin_addr.s_addr = INADDR_ANY;
    passive.sin_port = htons(lport);
    if (bind(sock, (struct sockaddr *)&passive,sizeof(struct sockaddr))==SOCKET_ERROR) {
		printf("Error binding socket [%d]", errno);
		return 2;
    }
	char* one="1";
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));
	listen(sock, 1);
#ifndef _WIN32
   uid_t uid = getuid();
   struct passwd *pw = getpwuid(uid);
	printf("Started server (API %d)\n",server->getAPIVersion());
	Log("%s started server (pid=%d)",pw->pw_name,getpid());
#else
	printf("Started server (API %d)\n",server->getAPIVersion());
	Log("Started server");
#endif
#ifndef _WIN32
    struct sigaction  saChild;
    saChild.sa_handler = catch_child;
    sigemptyset(&saChild.sa_mask);
    saChild.sa_flags = SA_NOCLDSTOP;
    sigaction(SIGCHLD, &saChild, 0);
 
	if (bFork) detach();
#endif
	FILE* fpPID;
	if ((fpPID = fopen(pidpath, "w")) != NULL) {
	      fprintf(fpPID, "%d\n", _getpid());
		fclose(fpPID);
	}
	for (;;) {
		PSOCKET m;
		sockaddr_in inaddr;
		int inal = sizeof(inaddr);
        while ((m = accept(sock, (struct sockaddr *)&inaddr, (socklen_t*) &inal)) == SOCKET_ERROR) {
//			if (m==WSAEINTR) exit (0);
        }
#ifdef IPV6_SUPORT
		char host[NI_MAXHOST];
		char svce[NI_MAXSERVICE];
		getnameinfo(&inaddr,inal,host,NI_MAXHOST,svce,NI_MAXSERVICE,0);
		NewSession(m,sock,host);
#else
		struct hostent *h;
		h = gethostbyaddr((char *) &inaddr.sin_addr, sizeof(inaddr.sin_addr), AF_INET);
		const char* un;
		if (h)
			un=h->h_name;
		else
			un=inet_ntoa(inaddr.sin_addr);
		NewSession(m,sock,un);
#endif
	}
	return 0;
}

bool Authenticate(const char* w)
{
	char name[100]; char pwd[100];
	int k=sscanf(w,"<?auth name='%99[^']' pwd='%99[^']'?>",name,pwd);
	if (k < 2) return false;
	return users.Lookup(name,pwd);
}

void Usage()
{
	printf("Usage:\n");
	printf("xairo_daemon -p params -l port -a logdir options\n");
	printf("      params     path to corpus parameter file\n");
	printf("      port       port to listen on\n");
	printf("      logdir     directory for log files\n");
	printf("	  options:\n");
	printf("	        -g create logfile\n");
	printf("	        -s start Sara server\n");
	printf("	        -u user authentication\n");
	printf("\n");
}

#ifndef _WIN32
void NewSession(PSOCKET m,PSOCKET s,const char* un)
{
	if (bFork) {
		int child=fork();
		if (child) {
			Log("Connection from %s (pid=%d)",un,child);
			closesocket(m);
			return;
		}
		else {
			closesocket(s);
			Session(m);
			Log("%s disconnected",un);
			exit(0);
		}
	} else {
		Session(m);
	}
}
#else
struct _xh {
	PSOCKET s1;
	PSOCKET s2;
	const char* un;
};

unsigned __stdcall SessionHelper(void* p)
{
	_xh* h=(_xh*) p;
//	closesocket(h->s2);
	Log("Connection from %s",h->un);
	Session(h->s1);
	Log("%s disconnected",h->un);
	return 0;
}


void NewSession(PSOCKET m,PSOCKET s,const char* un)
{
	unsigned id;
	_xh* h=new _xh;
	h->s1=m;
	h->s2=s;
	h->un=un;
	_beginthreadex(NULL,0,SessionHelper,(void*) h,0,&id);
}
#endif

void Session(PSOCKET m)
{
// start a new thread
	int bTrue=1; bool bOKUser=false; int ntry=maxtry-1;
	setsockopt(m, SOL_SOCKET, SO_KEEPALIVE, (char*)&bTrue, sizeof bTrue);
#ifndef _WIN32
	InitInt(m,server);
#endif

	char buffer[1000]; char* w; char* x;
	if (bLog) OpenSessionLog();
	for (;;) {  
		int n=0; x=NULL; bool bClosed=false;
		for (;;) {
      		int k = recv(m, buffer, 1000, 0); 
			if (k==0 || k==SOCKET_ERROR) {
				bClosed=true;
				if (bLog && k > 0) SessionLogErr();
				break;
			}
			char* xx=new char[n+k];
			if (x) memcpy(xx,x,n);
			memcpy(xx+n,buffer,k);
			if (x) delete [] x;
			x=xx;
			n+=k;
			w=x;
			int nw=n;
			if (w[nw-1]==0) break;
		}
		if (bClosed) break;
		if (bLog) SessionLog(w);
		const char* ans;
		if (strcmp(w,"<?xaira-client?>")==0 && !bSara)
		{
			char a[_MAX_PATH];
			const char* cn=server->getCorpusName();
			const char* url=users.getDownload();
			if (bSecure)
			sprintf(a,"<?xaira-server api='%d' name='%s' url='%s' authenticate='true'?>",server->getAPIVersion(),cn,url);
			else 
			sprintf(a,"<?xaira-server api='%d' name='%s' url='%s'?>",server->getAPIVersion(),cn,url);
			ans=a;
			bOKUser=!bSecure;
		}
		else if (bOKUser)
			ans=server->Exec(w);
		else if (ntry <= 0)
		{
			ans="<?auth status='term'?>"; 
			ntry--;
		}
		else {
			ntry--;
			bOKUser=Authenticate(w);
			ans=bOKUser?"<?auth status='ok'?>":"<?auth status='fail'?>";
		}
		if (bLog) SessionLog(ans);
		delete [] x;
		int k=send(m,ans,(int) strlen(ans)+1,0);
		if (k==SOCKET_ERROR) {
			if (bLog) SessionLogErr();
			break;
		}
		if (ntry < 0) break;
   }
	if (bLog) CloseSessionLog();
   closesocket(m);
}

#ifndef _WIN32
#define EOLS "\n"
#else
#define EOLS "\r\n"
#endif

#define SESSIONLOG "%sxaira_session%d.txt"

const char* SessionLogFile()
{
	static char p[_MAX_PATH];
#ifndef _WIN32
	int id=getpid();
#else
	int id=0;
#endif
#ifndef _WIN32
	snprintf(p,_MAX_PATH,SESSIONLOG,logdir,id);
#else
	_snprintf(p,_MAX_PATH,SESSIONLOG,logdir,id);
#endif
	return p;
}

void SessionLog(const char* s)
{
	FILE* f=fopen(SessionLogFile(),"ab");
	if (f==NULL) return;
	fwrite(s,strlen(s),1,f);
	fwrite(EOLS,sizeof(EOLS)-1,1,f);
	fclose(f);
}


void SessionLogErr()
{
#ifndef _WIN32
		int e=errno;
#else
		int e=WSAGetLastError();
#endif
	char x[1000];
	sprintf(x,"<socket_error>%d</socket_error>",e);
	FILE* f=fopen(SessionLogFile(),"ab");
	if (f==NULL) return;
	fwrite(x,strlen(x),1,f);
	fwrite(EOLS,sizeof(EOLS)-1,1,f);
	fclose(f);
}

unsigned char bom[3]={0xEF,0xBB,0xBF};


void OpenSessionLog()
{
	FILE* f=fopen(SessionLogFile(),"wb");
	if (f==NULL) return;
	fwrite(bom,3,1,f);
	fprintf(f,"<session_log>\n");
	fclose(f);
}

void CloseSessionLog()
{
	FILE* f=fopen(SessionLogFile(),"ab");
	if (f==NULL) return;
	fprintf(f,"</session_log>\n");
	fclose(f);
}

#ifndef _WIN32
void detach()
{
   pid_t pgrp;

   /*
    * if our parent is init (pid 1) we don't have a controlling
    * terminal
    */
   if (getppid() != 1) {
      switch(fork()) {
       case -1:
//         syslog(LOG_ERR, "fork() failed [%d]", errno);
         perror("fork() failed");
         exit(1);
         break;

       case 0:
#if !defined(NO_SETSID)
         if ((pgrp = setsid()) == -1) {
 //           syslog(LOG_ERR, "setsid() failed [%d]", errno);
            perror("setsid() failed");
            exit(1);
         }
#else
         if ((pgrp = setpgrp(getpid(), 0)) == -1) {
 //           syslog(LOG_ERR, "setpgrp() failed [%d]", errno);
            perror("setpgrp");
            exit(1);
         }
#endif
         break;

       default:
         exit(0);
      }
   }
}     

static CXairaServer* g_x;
static PSOCKET g_s;

void sig_urgent(int);

void InitInt(PSOCKET s,CXairaServer* x)
{
	g_x=x; g_s=s;
	struct sigaction sa;
    sa.sa_handler = sig_urgent;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGURG, &sa, NULL);
    fcntl(s, F_SETOWN, getpid());

}

void sig_urgent(int sig)
{
   char      buffer[100];
   recv(g_s, buffer, sizeof(buffer), 0);
   if (bLog) SessionLog("<?int?>");
   g_x->Interrupt();
}

#endif