// xaira_test.cpp : Defines the entry point for the console application.
//
#include "stdafx.h"


#ifndef _WIN32
typedef unsigned short xairo_uchar;
#include "../indexer/tinylib.h"
#else


#endif

#ifndef _WIN32
#ifndef DATAPATH
#error The symbol DATAPATH must be defined
#endif
#define SEPCHAR '\\'
#else
#define SEPCHAR '/'
#endif

#include "../XairoHeader.h"

using namespace XAIRO_NAMESPACE;

#include "XairaTestHarness.h"

void Usage()
{
	printf("Usage:\n");
	printf("xaira_test -p path -s script\n");
	printf("     where path is the path to the corpus parameter file.\n");
	printf("xaira_test -u host [-l port] -s script\n");
	printf("     where host is the server host address and port the port (default 7000).\n\n");
}

void Message(UnicodeString a)
{
	char* s=new char[a.length()+1];
	a.extract(0,a.length(),s);
	printf("%s",s);
	delete [] s;
}


class MyBootLog : public CXairoIBootMonitor
{
	void Log(const xairo_uchar* u) {Message(u);}
};

char datapath[_MAX_PATH];

int main(int argc, char* argv[])
{
	bool bErr=false; bool bCalib=false;
	unsigned lport=7000;
	char paramfile[_MAX_PATH];
	char scriptfile[_MAX_PATH];
	paramfile[0]=0;
	char url[_MAX_PATH];
	url[0]=0;
#ifdef DATAPATH
	strcpy(datapath,DATAPATH);
#endif
	char* dpe=getenv("XAIRA_DATAPATH");
	if (dpe) {
		if (dpe[0]==SEPCHAR)
			strcpy(datapath,dpe);
		else {
			_getcwd(datapath,_MAX_PATH);
			strcat(datapath,"/");
			strcat(datapath,dpe);
		}
	}
	for (int i=1;i<argc;i++) {
		if (strcmp(argv[i],"-p")==0 && i+1 < argc) {
			strcpy(paramfile,argv[i+1]);
			i++;
		}
		else if (strcmp(argv[i],"-s")==0 && i+1 < argc) {
			strcpy(scriptfile,argv[i+1]);
			i++;
		}
		else if (strcmp(argv[i],"-u")==0 && i+1 < argc) {
			strcpy(url,argv[i+1]);
			i++;
		} else if (strcmp(argv[i],"-l")==0 && i+1 < argc) {
			lport=atoi(argv[i+1]);
			i++;
		}
		 else if (strcmp(argv[i],"-c")==0) {
			bCalib=true;
		}
		else
			bErr=true;
	}
	if (bErr || paramfile[0]==0 && url[0]==0 || paramfile[0]!=0 && url[0]!=0 ) 
	{	
		Usage();
		return 1001;
	}
	CXairoInterface* s=NULL;
	MyBootLog b;
	if (paramfile[0]) {
		CXairoIParameters xp;
		xp.parameter_file=paramfile;
		xp.bm=&b;
		xp.inth=NULL;
#ifdef _WIN32
		char tp[_MAX_PATH];
		GetTempPathA(_MAX_PATH,tp);
		xp.temppath=tp;
		char dp[_MAX_PATH];
		GetModuleFileNameA(NULL,dp,_MAX_PATH);
		char* x=strrchr(dp,'\\');
		*x=0;
		xp.datapath=dp;
#else
		xp.temppath="/tmp";
		xp.datapath=datapath;
#endif
	s=CXairoImplementation::getServer(&xp);
	} else {
		s=CXairoProxyImplementation::getRemoteServer(url,lport,&b);
	}
	if (s==NULL) {
		printf("Cannot create server\n");
		return 1002;
	}
	CXairaTestHarness th;
	try {
		th.exec(scriptfile,s,bCalib);
	}
	catch (...) {
		printf("Uncaught exception\n");
		return 1003;
	}
	printf("Normal termination\n");
	return th.status();
}

